/*
 * Driver for FPGA Management Engine which implements all FPGA platform
 * level management features.
 *
 * Copyright 2016 Intel Corporation, Inc.
 *
 * Authors:
 *   Kang Luwei <luwei.kang@intel.com>
 *   Xiao Guangrong <guangrong.xiao@linux.intel.com>
 *   Joseph Grecco <joe.grecco@intel.com>
 *   Enno Luebbers <enno.luebbers@intel.com>
 *   Tim Whisonant <tim.whisonant@intel.com>
 *   Ananda Ravuri <ananda.ravuri@intel.com>
 *   Mitchel, Henry <henry.mitchel@intel.com>
 *
 * This work is licensed under the terms of the GNU GPL version 2. See
 * the COPYING file in the top-level directory.
 *
 */

#include <linux/avmmi-bmc.h>
#include <linux/bitfield-intel-pac.h>
#include <linux/clk-provider.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/stddef.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/uaccess.h>
#include <linux/intel-fpga.h>
#include <linux/fpga/fpga-mgr-4-12.h>
#include <linux/iopoll-4-12.h>
#include <linux/i2c/altera.h>
#include <linux/spi/altera.h>
#include <linux/mtd/altera-asmip2.h>
#include <linux/version.h>

#include "pac-hssi.h"
#include "feature-dev.h"
#include "fme.h"

#define PWR_THRESHOLD_MAX       0x7F

#define FME_DEV_ATTR(_name, _filename, _mode, _show, _store)	\
struct device_attribute dev_attr_##_name =			\
	__ATTR(_filename, _mode, _show, _store)

static char *fec_mode = "rs";
module_param(fec_mode, charp, 0444);
MODULE_PARM_DESC(child, "FEC mode");

static ssize_t revision_show(struct device *dev, struct device_attribute *attr,
			     char *buf)
{
	struct feature_fme_header *fme_hdr
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_HEADER);
	struct feature_header header;

	header.csr = readq(&fme_hdr->header);

	return scnprintf(buf, PAGE_SIZE, "%d\n", header.revision);
}

static DEVICE_ATTR_RO(revision);

static ssize_t ports_num_show(struct device *dev, struct device_attribute *attr,
			char *buf)
{
	struct feature_fme_header *fme_hdr
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_HEADER);
	struct feature_fme_capability fme_capability;

	fme_capability.csr = readq(&fme_hdr->capability);

	return scnprintf(buf, PAGE_SIZE, "%d\n", fme_capability.num_ports);
}

static DEVICE_ATTR_RO(ports_num);

static ssize_t cache_size_show(struct device *dev,
			struct device_attribute *attr, char *buf)
{
	struct feature_fme_header *fme_hdr
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_HEADER);
	struct feature_fme_capability fme_capability;

	fme_capability.csr = readq(&fme_hdr->capability);

	return scnprintf(buf, PAGE_SIZE, "%d\n", fme_capability.cache_size);
}

static DEVICE_ATTR_RO(cache_size);

static ssize_t version_show(struct device *dev, struct device_attribute *attr,
			char *buf)
{
	struct feature_fme_header *fme_hdr
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_HEADER);
	struct feature_fme_capability fme_capability;

	fme_capability.csr = readq(&fme_hdr->capability);

	return scnprintf(buf, PAGE_SIZE, "%d\n", fme_capability.fabric_verid);
}

static DEVICE_ATTR_RO(version);

static ssize_t socket_id_show(struct device *dev,
			struct device_attribute *attr, char *buf)
{
	struct feature_fme_header *fme_hdr
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_HEADER);
	struct feature_fme_capability fme_capability;

	fme_capability.csr = readq(&fme_hdr->capability);

	return scnprintf(buf, PAGE_SIZE, "%d\n", fme_capability.socket_id);
}

static DEVICE_ATTR_RO(socket_id);

static ssize_t bitstream_id_show(struct device *dev,
			struct device_attribute *attr, char *buf)
{
	struct feature_fme_header *fme_hdr
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_HEADER);
	u64 bitstream_id = readq(&fme_hdr->bitstream_id);

	return scnprintf(buf, PAGE_SIZE, "0x%llx\n", bitstream_id);
}

static DEVICE_ATTR_RO(bitstream_id);

static ssize_t bitstream_metadata_show(struct device *dev,
			struct device_attribute *attr, char *buf)
{
	struct feature_fme_header *fme_hdr
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_HEADER);
	u64 bitstream_md = readq(&fme_hdr->bitstream_md);

	return scnprintf(buf, PAGE_SIZE, "0x%llx\n", bitstream_md);
}

static DEVICE_ATTR_RO(bitstream_metadata);

static const struct attribute *fme_hdr_attrs[] = {
	&dev_attr_revision.attr,
	&dev_attr_ports_num.attr,
	&dev_attr_cache_size.attr,
	&dev_attr_version.attr,
	&dev_attr_socket_id.attr,
	&dev_attr_bitstream_id.attr,
	&dev_attr_bitstream_metadata.attr,
	NULL,
};

static int fme_hdr_init(struct platform_device *pdev, struct feature *feature)
{
	int ret;
	struct feature_fme_header *fme_hdr = feature->ioaddr;

	dev_dbg(&pdev->dev, "FME HDR Init.\n");
	dev_dbg(&pdev->dev, "FME cap %llx.\n", fme_hdr->capability.csr);

	ret = sysfs_create_files(&pdev->dev.kobj, fme_hdr_attrs);
	if (ret)
		return ret;

	return 0;
}

static void fme_hdr_uinit(struct platform_device *pdev, struct feature *feature)
{
	dev_dbg(&pdev->dev, "FME HDR UInit.\n");
	sysfs_remove_files(&pdev->dev.kobj, fme_hdr_attrs);
}

static struct feature_ops fme_hdr_ops = {
	.init = fme_hdr_init,
	.uinit = fme_hdr_uinit,
};

static ssize_t thermal_revision_show(struct device *dev,
				 struct device_attribute *attr, char *buf)
{
	struct feature_fme_thermal *fme_thermal
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_THERMAL_MGMT);
	struct feature_header header;

	header.csr = readq(&fme_thermal->header);

	return scnprintf(buf, PAGE_SIZE, "%d\n", header.revision);
}

static FME_DEV_ATTR(thermal_revision, revision, 0444,
		    thermal_revision_show, NULL);

static ssize_t thermal_threshold1_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_thermal *fme_thermal
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_THERMAL_MGMT);
	struct feature_fme_tmp_threshold tmp_threshold;

	tmp_threshold.csr = readq(&fme_thermal->threshold);

	return scnprintf(buf, PAGE_SIZE, "%d\n", tmp_threshold.tmp_thshold1);
}

static ssize_t thermal_threshold1_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct feature_platform_data *pdata = dev_get_platdata(dev);
	struct feature_fme_thermal *fme_thermal
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_THERMAL_MGMT);
	struct feature_fme_header *fme_hdr
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_HEADER);
	struct feature_fme_tmp_threshold tmp_threshold;
	struct feature_fme_capability fme_capability;
	int err;
	u8 tmp_threshold1;

	mutex_lock(&pdata->lock);
	tmp_threshold.csr = readq(&fme_thermal->threshold);

	err = kstrtou8(buf, 0, &tmp_threshold1);
	if (err) {
		mutex_unlock(&pdata->lock);
		return err;
	}

	fme_capability.csr = readq(&fme_hdr->capability);

	if (fme_capability.lock_bit == 1) {
		mutex_unlock(&pdata->lock);
		return -EBUSY;
	} else if (tmp_threshold1 > 100) {
		mutex_unlock(&pdata->lock);
		return -EINVAL;
	} else if (tmp_threshold1 == 0) {
		tmp_threshold.tmp_thshold1_enable = 0;
		tmp_threshold.tmp_thshold1 = tmp_threshold1;
	} else {
		tmp_threshold.tmp_thshold1_enable = 1;
		tmp_threshold.tmp_thshold1 = tmp_threshold1;
	}

	writeq(tmp_threshold.csr, &fme_thermal->threshold);
	mutex_unlock(&pdata->lock);

	return count;
}

static DEVICE_ATTR(threshold1, 0644,
	thermal_threshold1_show, thermal_threshold1_store);

static ssize_t thermal_threshold2_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_thermal *fme_thermal
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_THERMAL_MGMT);
	struct feature_fme_tmp_threshold tmp_threshold;

	tmp_threshold.csr = readq(&fme_thermal->threshold);

	return scnprintf(buf, PAGE_SIZE, "%d\n", tmp_threshold.tmp_thshold2);
}

static ssize_t thermal_threshold2_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct feature_platform_data *pdata = dev_get_platdata(dev);
	struct feature_fme_thermal *fme_thermal
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_THERMAL_MGMT);
	struct feature_fme_header *fme_hdr
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_HEADER);
	struct feature_fme_tmp_threshold tmp_threshold;
	struct feature_fme_capability fme_capability;
	int err;
	u8 tmp_threshold2;

	mutex_lock(&pdata->lock);
	tmp_threshold.csr = readq(&fme_thermal->threshold);

	err = kstrtou8(buf, 0, &tmp_threshold2);
	if (err) {
		mutex_unlock(&pdata->lock);
		return err;
	}

	fme_capability.csr = readq(&fme_hdr->capability);

	if (fme_capability.lock_bit == 1) {
		mutex_unlock(&pdata->lock);
		return -EBUSY;
	} else if (tmp_threshold2 > 100) {
		mutex_unlock(&pdata->lock);
		return -EINVAL;
	} else if (tmp_threshold2 == 0) {
		tmp_threshold.tmp_thshold2_enable = 0;
		tmp_threshold.tmp_thshold2 = tmp_threshold2;
	} else {
		tmp_threshold.tmp_thshold2_enable = 1;
		tmp_threshold.tmp_thshold2 = tmp_threshold2;
	}

	writeq(tmp_threshold.csr, &fme_thermal->threshold);
	mutex_unlock(&pdata->lock);

	return count;
}

static DEVICE_ATTR(threshold2, 0644,
	thermal_threshold2_show, thermal_threshold2_store);

static ssize_t thermal_threshold_trip_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_thermal *fme_thermal
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_THERMAL_MGMT);
	struct feature_fme_tmp_threshold tmp_threshold;

	tmp_threshold.csr = readq(&fme_thermal->threshold);

	return scnprintf(buf, PAGE_SIZE, "%d\n",
				tmp_threshold.therm_trip_thshold);
}

static DEVICE_ATTR(threshold_trip, 0444, thermal_threshold_trip_show, NULL);

static ssize_t thermal_threshold1_reached_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_thermal *fme_thermal
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_THERMAL_MGMT);
	struct feature_fme_tmp_threshold tmp_threshold;

	tmp_threshold.csr = readq(&fme_thermal->threshold);

	return scnprintf(buf, PAGE_SIZE, "%d\n",
				tmp_threshold.thshold1_status);
}

static DEVICE_ATTR(threshold1_reached, 0444,
	thermal_threshold1_reached_show, NULL);

static ssize_t thermal_threshold2_reached_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_thermal *fme_thermal
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_THERMAL_MGMT);
	struct feature_fme_tmp_threshold tmp_threshold;

	tmp_threshold.csr = readq(&fme_thermal->threshold);

	return scnprintf(buf, PAGE_SIZE, "%d\n",
				tmp_threshold.thshold2_status);
}

static DEVICE_ATTR(threshold2_reached, 0444,
	thermal_threshold2_reached_show, NULL);

static ssize_t thermal_threshold1_policy_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_thermal *fme_thermal
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_THERMAL_MGMT);
	struct feature_fme_tmp_threshold tmp_threshold;

	tmp_threshold.csr = readq(&fme_thermal->threshold);

	return scnprintf(buf, PAGE_SIZE, "%d\n",
				tmp_threshold.thshold_policy);
}

static ssize_t thermal_threshold1_policy_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct feature_platform_data *pdata = dev_get_platdata(dev);
	struct feature_fme_thermal *fme_thermal
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_THERMAL_MGMT);
	struct feature_fme_tmp_threshold tmp_threshold;
	int err;
	u8 thshold_policy;

	mutex_lock(&pdata->lock);
	tmp_threshold.csr = readq(&fme_thermal->threshold);

	err = kstrtou8(buf, 0, &thshold_policy);
	if (err) {
		mutex_unlock(&pdata->lock);
		return err;
	}

	if (thshold_policy == 0)
		tmp_threshold.thshold_policy = 0;
	else if (thshold_policy == 1)
		tmp_threshold.thshold_policy = 1;
	else {
		mutex_unlock(&pdata->lock);
		return -EINVAL;
	}

	writeq(tmp_threshold.csr, &fme_thermal->threshold);
	mutex_unlock(&pdata->lock);

	return count;
}

static DEVICE_ATTR(threshold1_policy, 0644,
	thermal_threshold1_policy_show, thermal_threshold1_policy_store);

static ssize_t thermal_temperature_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_thermal *fme_thermal
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_THERMAL_MGMT);
	struct feature_fme_temp_rdsensor_fmt1 temp_rdsensor_fmt1;

	temp_rdsensor_fmt1.csr = readq(&fme_thermal->rdsensor_fm1);

	return scnprintf(buf, PAGE_SIZE, "%d\n",
				temp_rdsensor_fmt1.fpga_temp);
}

static DEVICE_ATTR(temperature, 0444, thermal_temperature_show, NULL);

static struct attribute *thermal_mgmt_attrs[] = {
	&dev_attr_thermal_revision.attr,
	&dev_attr_temperature.attr,
	NULL,
};

static struct attribute_group thermal_mgmt_attr_group = {
	.name   = "thermal_mgmt",
	.attrs	= thermal_mgmt_attrs,
};

static struct attribute *thermal_threshold_attrs[] = {
	&dev_attr_threshold1.attr,
	&dev_attr_threshold2.attr,
	&dev_attr_threshold_trip.attr,
	&dev_attr_threshold1_reached.attr,
	&dev_attr_threshold2_reached.attr,
	&dev_attr_threshold1_policy.attr,
	NULL,
};

static struct attribute_group thermal_threshold_attr_group = {
	.name   = "thermal_mgmt",
	.attrs	= thermal_threshold_attrs,
};

static int thermal_mgmt_init(struct platform_device *pdev,
			     struct feature *feature)
{
	struct feature_fme_thermal *fme_thermal;
	struct feature_fme_tmp_threshold_cap thermal_cap;
	int ret;

	fme_thermal = get_feature_ioaddr_by_id(&pdev->dev,
					       FME_FEATURE_ID_THERMAL_MGMT);
	thermal_cap.csr = readq(&fme_thermal->threshold_cap);

	ret = sysfs_create_group(&pdev->dev.kobj, &thermal_mgmt_attr_group);
	if (ret)
		return ret;

	if (thermal_cap.tmp_thshold_disabled)
		return 0;

	ret = sysfs_merge_group(&pdev->dev.kobj, &thermal_threshold_attr_group);
	if (ret) {
		sysfs_remove_group(&pdev->dev.kobj, &thermal_mgmt_attr_group);
		return ret;
	}

	return 0;
}

static void thermal_mgmt_uinit(struct platform_device *pdev,
				struct feature *feature)
{
	sysfs_unmerge_group(&pdev->dev.kobj, &thermal_threshold_attr_group);
	sysfs_remove_group(&pdev->dev.kobj, &thermal_mgmt_attr_group);
}

static struct feature_ops thermal_mgmt_ops = {
	.init = thermal_mgmt_init,
	.uinit = thermal_mgmt_uinit,
};

static ssize_t pwr_revision_show(struct device *dev,
				 struct device_attribute *attr, char *buf)
{
	struct feature_fme_power *fme_power
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_POWER_MGMT);
	struct feature_header header;

	header.csr = readq(&fme_power->header);

	return scnprintf(buf, PAGE_SIZE, "%d\n", header.revision);
}

static FME_DEV_ATTR(pwr_revision, revision, 0444, pwr_revision_show, NULL);

static ssize_t consumed_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_power *fme_power
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_POWER_MGMT);
	struct feature_fme_pm_status pm_status;

	pm_status.csr = readq(&fme_power->status);

	return scnprintf(buf, PAGE_SIZE, "0x%x\n", pm_status.pwr_consumed);
}

static DEVICE_ATTR_RO(consumed);

static ssize_t pwr_threshold1_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_power *fme_power
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_POWER_MGMT);
	struct feature_fme_pm_ap_threshold pm_ap_threshold;

	pm_ap_threshold.csr = readq(&fme_power->threshold);

	return scnprintf(buf, PAGE_SIZE, "0x%x\n", pm_ap_threshold.threshold1);
}

static ssize_t pwr_threshold1_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct feature_platform_data *pdata = dev_get_platdata(dev);
	struct feature_fme_power *fme_power
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_POWER_MGMT);
	struct feature_fme_pm_ap_threshold pm_ap_threshold;
	u8 threshold;
	int err;

	mutex_lock(&pdata->lock);
	pm_ap_threshold.csr = readq(&fme_power->threshold);

	err = kstrtou8(buf, 0, &threshold);
	if (err) {
		mutex_unlock(&pdata->lock);
		return err;
	}

	if (threshold <= PWR_THRESHOLD_MAX)
		pm_ap_threshold.threshold1 = threshold;
	else {
		mutex_unlock(&pdata->lock);
		return -EINVAL;
	}

	writeq(pm_ap_threshold.csr, &fme_power->threshold);
	mutex_unlock(&pdata->lock);

	return count;
}

static FME_DEV_ATTR(pwr_threshold1, threshold1, 0644, pwr_threshold1_show,
		    pwr_threshold1_store);

static ssize_t pwr_threshold2_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_power *fme_power
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_POWER_MGMT);
	struct feature_fme_pm_ap_threshold pm_ap_threshold;

	pm_ap_threshold.csr = readq(&fme_power->threshold);

	return scnprintf(buf, PAGE_SIZE, "0x%x\n",
				pm_ap_threshold.threshold2);
}

static ssize_t pwr_threshold2_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct feature_platform_data *pdata = dev_get_platdata(dev);
	struct feature_fme_power *fme_power
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_POWER_MGMT);
	struct feature_fme_pm_ap_threshold pm_ap_threshold;
	u8 threshold;
	int err;

	mutex_lock(&pdata->lock);
	pm_ap_threshold.csr = readq(&fme_power->threshold);

	err = kstrtou8(buf, 0, &threshold);
	if (err) {
		mutex_unlock(&pdata->lock);
		return err;
	}

	if (threshold <= PWR_THRESHOLD_MAX)
		pm_ap_threshold.threshold2 = threshold;
	else {
		mutex_unlock(&pdata->lock);
		return -EINVAL;
	}

	writeq(pm_ap_threshold.csr, &fme_power->threshold);
	mutex_unlock(&pdata->lock);

	return count;
}

static FME_DEV_ATTR(pwr_threshold2, threshold2, 0644, pwr_threshold2_show,
		    pwr_threshold2_store);

static ssize_t threshold1_status_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_power *fme_power
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_POWER_MGMT);
	struct feature_fme_pm_ap_threshold pm_ap_threshold;

	pm_ap_threshold.csr = readq(&fme_power->threshold);

	return scnprintf(buf, PAGE_SIZE, "%u\n",
				pm_ap_threshold.threshold1_status);
}

static DEVICE_ATTR_RO(threshold1_status);

static ssize_t threshold2_status_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_power *fme_power
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_POWER_MGMT);
	struct feature_fme_pm_ap_threshold pm_ap_threshold;

	pm_ap_threshold.csr = readq(&fme_power->threshold);

	return scnprintf(buf, PAGE_SIZE, "%u\n",
				pm_ap_threshold.threshold2_status);
}
static DEVICE_ATTR_RO(threshold2_status);

static ssize_t rtl_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_power *fme_power
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_POWER_MGMT);
	struct feature_fme_pm_status pm_status;

	pm_status.csr = readq(&fme_power->status);

	return scnprintf(buf, PAGE_SIZE, "%u\n",
				pm_status.fpga_latency_report);
}

static DEVICE_ATTR_RO(rtl);

static ssize_t xeon_limit_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_power *fme_power
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_POWER_MGMT);
	struct feature_fme_pm_xeon_limit xeon_limit;

	xeon_limit.csr = readq(&fme_power->xeon_limit);

	if (!xeon_limit.enable)
		xeon_limit.pwr_limit = 0;

	return scnprintf(buf, PAGE_SIZE, "%u\n", xeon_limit.pwr_limit);
}
static DEVICE_ATTR_RO(xeon_limit);

static ssize_t fpga_limit_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct feature_fme_power *fme_power
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_POWER_MGMT);
	struct feature_fme_pm_fpga_limit fpga_limit;

	fpga_limit.csr = readq(&fme_power->fpga_limit);

	if (!fpga_limit.enable)
		fpga_limit.pwr_limit = 0;

	return scnprintf(buf, PAGE_SIZE, "%u\n", fpga_limit.pwr_limit);
}
static DEVICE_ATTR_RO(fpga_limit);

static struct attribute *power_mgmt_attrs[] = {
	&dev_attr_pwr_revision.attr,
	&dev_attr_consumed.attr,
	&dev_attr_pwr_threshold1.attr,
	&dev_attr_pwr_threshold2.attr,
	&dev_attr_threshold1_status.attr,
	&dev_attr_threshold2_status.attr,
	&dev_attr_xeon_limit.attr,
	&dev_attr_fpga_limit.attr,
	&dev_attr_rtl.attr,
	NULL,
};

static struct attribute_group power_mgmt_attr_group = {
	.attrs	= power_mgmt_attrs,
	.name	= "power_mgmt",
};

static int power_mgmt_init(struct platform_device *pdev,
				struct feature *feature)
{
	int ret;

	ret = sysfs_create_group(&pdev->dev.kobj, &power_mgmt_attr_group);
	if (ret)
		return ret;

	return 0;
}

static void power_mgmt_uinit(struct platform_device *pdev,
				struct feature *feature)
{
	sysfs_remove_group(&pdev->dev.kobj, &power_mgmt_attr_group);
}

static struct feature_ops power_mgmt_ops = {
	.init = power_mgmt_init,
	.uinit = power_mgmt_uinit,
};

static int hssi_mgmt_init(struct platform_device *pdev, struct feature *feature)
{
	dev_dbg(&pdev->dev, "FME HSSI Init.\n");
	return 0;
}

static void hssi_mgmt_uinit(struct platform_device *pdev,
				struct feature *feature)
{
	dev_dbg(&pdev->dev, "FME HSSI UInit.\n");
}

static struct feature_ops hssi_mgmt_ops = {
	.init = hssi_mgmt_init,
	.uinit = hssi_mgmt_uinit,
};

#define FLASH_CAP_OFT 8

static int qspi_flash_init(struct platform_device *pdev,
			   struct feature *feature)
{
	struct altera_asmip2_plat_data qdata;
	struct platform_device *subdev;

	dev_dbg(&pdev->dev, "FME QSPI FLASH Init\n");
	dev_dbg(&pdev->dev, "index %d flash cap 0x%llx\n",
		feature->resource_index,
		(unsigned long long)readq(feature->ioaddr + FLASH_CAP_OFT));

	memset(&qdata, 0, sizeof(qdata));
	qdata.csr_base = feature->ioaddr + FLASH_CAP_OFT;
	qdata.num_chip_sel = 1;

	subdev = feature_create_subdev(pdev, ALTERA_ASMIP2_DRV_NAME,
				       &qdata, sizeof(qdata));
	if (IS_ERR(subdev))
		return PTR_ERR(subdev);

	feature_set_priv(feature, subdev);
	return 0;
}

static void qspi_flash_uinit(struct platform_device *pdev,
			     struct feature *feature)
{
	struct platform_device *subdev = feature_get_priv(feature);

	dev_dbg(&pdev->dev, "FME QSPI FLASH UInit\n");

	feature_destroy_subdev(subdev);
	feature_set_priv(feature, NULL);
}

static struct feature_ops qspi_flash_ops = {
	.init = qspi_flash_init,
	.uinit = qspi_flash_uinit,
};

static ssize_t emif_active_show(struct device *dev,
				 struct device_attribute *attr, char *buf)
{
	struct feature_fme_emif *fme_emif
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_EMIF_MGMT);
	struct feature_fme_emif_status emif_status;

	emif_status.csr = readq(&fme_emif->status);

	return scnprintf(buf, PAGE_SIZE, "0x%x\n", emif_status.active);
}

static FME_DEV_ATTR(emif_active, active, 0444, emif_active_show, NULL);

static ssize_t emif_cal_fail_show(struct device *dev,
				 struct device_attribute *attr, char *buf)
{
	struct feature_fme_emif *fme_emif
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_EMIF_MGMT);
	struct feature_fme_emif_status emif_status;

	emif_status.csr = readq(&fme_emif->status);

	return scnprintf(buf, PAGE_SIZE, "0x%x\n", emif_status.cal_failure);
}

static FME_DEV_ATTR(emif_cal_fail, calibration_failure, 0444,
			emif_cal_fail_show, NULL);

static ssize_t emif_complete_show(struct device *dev,
				 struct device_attribute *attr, char *buf)
{
	struct feature_fme_emif *fme_emif
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_EMIF_MGMT);
	struct feature_fme_emif_status emif_status;

	emif_status.csr = readq(&fme_emif->status);

	return scnprintf(buf, PAGE_SIZE, "0x%x\n", emif_status.complete);
}

static FME_DEV_ATTR(emif_complete, complete, 0444, emif_complete_show, NULL);

static ssize_t emif_clear_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct feature_fme_emif *fme_emif
		= get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_EMIF_MGMT);
	struct feature_fme_emif_control emif_control = { { 0 } };
	u8 clear_bits;
	int err;

	err = kstrtou8(buf, 0, &clear_bits);
	if (err)
		return err;

	if (clear_bits & ~EMIF_CLEAR_MASK)
		return -EINVAL;

	emif_control.clear = clear_bits;
	writeq(emif_control.csr, &fme_emif->control);

	return count;
}

static FME_DEV_ATTR(emif_clear, clear, 0200, NULL, emif_clear_store);

static struct attribute *emif_attrs[] = {
	&dev_attr_emif_active.attr,
	&dev_attr_emif_cal_fail.attr,
	&dev_attr_emif_complete.attr,
	&dev_attr_emif_clear.attr,
	NULL,
};

static struct attribute_group emif_attr_group = {
	.name	= "emif_mgmt",
	.attrs	= emif_attrs,
};

static int emif_init(struct platform_device *pdev,
				struct feature *feature)
{
	return sysfs_create_group(&pdev->dev.kobj, &emif_attr_group);
}

static void emif_uinit(struct platform_device *pdev,
				struct feature *feature)
{
	sysfs_remove_group(&pdev->dev.kobj, &emif_attr_group);
}

static struct feature_ops emif_ops = {
	.init = emif_init,
	.uinit = emif_uinit,
};

static int pac_hssi_mgmt_init(struct platform_device *pdev,
			      struct feature *feature)
{
	struct pac_hssi_plat_data hdata;
	struct platform_device *subdev;
	struct feature_header header;

	dev_dbg(&pdev->dev, "FME PAC HSSI MGMT Init\n");

	header.csr = readq(feature->ioaddr);

	memset(&hdata, 0, sizeof(hdata));
	hdata.csr_base = feature->ioaddr + sizeof(struct feature_header);
	hdata.revision = header.revision;

	subdev = feature_create_subdev(pdev, PAC_HSSI_DRV_NAME, &hdata,
				       sizeof(hdata));
	if (IS_ERR(subdev))
		return PTR_ERR(subdev);

	feature_set_priv(feature, subdev);
	return 0;
}

static void pac_hssi_mgmt_uinit(struct platform_device *pdev,
				struct feature *feature)
{
	struct platform_device *subdev = feature_get_priv(feature);

	dev_dbg(&pdev->dev, "FME PAC HSSI MGMT UInit\n");

	feature_destroy_subdev(subdev);
	feature_set_priv(feature, NULL);
}

static struct feature_ops pac_hssi_mgmt_ops = {
	.init = pac_hssi_mgmt_init,
	.uinit = pac_hssi_mgmt_uinit,
};

static int s10_sdm_init(struct platform_device *pdev, struct feature *feature)
{
	dev_dbg(&pdev->dev, "FME S10 SDM Init\n");
	return 0;
}

static void s10_sdm_uinit(struct platform_device *pdev, struct feature *feature)
{
	dev_dbg(&pdev->dev, "FME S10 SDM UInit\n");
}

static struct feature_ops s10_sdm_mgmt_ops = {
	.init = s10_sdm_init,
	.uinit = s10_sdm_uinit,
};

static int avmmi_bmc_init(struct platform_device *pdev,
			  struct feature *feature)
{
	struct avmmi_plat_data qdata;
	struct platform_device *subdev;

	dev_dbg(&pdev->dev, "FME AVMMI_BMC Init\n");

	memset(&qdata, 0, sizeof(qdata));

	qdata.csr_base = feature->ioaddr;

	subdev = feature_create_subdev(pdev, AVMMI_BMC_DRV_NAME,
				       &qdata, sizeof(qdata));
	if (IS_ERR(subdev))
		return PTR_ERR(subdev);

	feature_set_priv(feature, subdev);
	return 0;
}

static void avmmi_bmc_uinit(struct platform_device *pdev,
			    struct feature *feature)
{
	struct platform_device *subdev = feature_get_priv(feature);

	dev_dbg(&pdev->dev, "FME AVMMI_BMC UInit\n");

	feature_destroy_subdev(subdev);
	feature_set_priv(feature, NULL);
}

static struct feature_ops avmmi_bmc_ops = {
	.init = avmmi_bmc_init,
	.uinit = avmmi_bmc_uinit,
};

#define INDIRECT_ADDR		0x0
#define INDIRECT_WR		BIT_ULL(8)
#define INDIRECT_RD		BIT_ULL(9)
#define INDIRECT_RD_DATA	0x8
#define INDIRECT_DATA_MASK	GENMASK_ULL(31, 0)
#define INDIRECT_DEBUG		BIT_ULL(32)
#define INDIRECT_WR_DATA	0x10

/* max10 register addr is 32 bits */
#define REG_SIZE		4UL

/* max10 register value is 32 bits */
#define VAL_SIZE		4UL

static int do_reg_write(void *context, unsigned int reg,
			unsigned int *value, u16 count)
{
	void __iomem *base = context;
	int i;

	for (i = 0; i < count; i++) {
		writeq(*value++, base + INDIRECT_WR_DATA);
		writeq((reg >> 2) | INDIRECT_WR, base + INDIRECT_ADDR);
	}

	return 0;
}

static int indirect_gather_write(void *context,
				 const void *reg_buf, size_t reg_len,
				 const void *val_buf, size_t val_len)
{
	u32 reg = *(u32 *)reg_buf;
	u16 count = val_len / VAL_SIZE;

	//printk("%s: reg=0x%x, val=0x%x\n", __func__, reg, *(u32 *)val_buf);

	if (reg_len != REG_SIZE || !val_len || !IS_ALIGNED(val_len, VAL_SIZE))
		return -EINVAL;

	return do_reg_write(context, reg, (u32 *)val_buf, count);
}

static int indirect_write(void *context, const void *data, size_t count)
{
	
	if (count < REG_SIZE + VAL_SIZE)
		return -EINVAL;

	return indirect_gather_write(context, data, REG_SIZE,
				     data + REG_SIZE, count - REG_SIZE);
}

static int do_reg_read(void *context, unsigned int reg,
		       unsigned int *value, u16 count)
{
	void __iomem *base = context;
	int i;
	u64 v;

	for (i = 0; i < count; i++) {
		writeq((reg >> 2) | INDIRECT_RD, base + INDIRECT_ADDR);
		v = readq(base + INDIRECT_RD_DATA);

		/*
		 * FIXME: Read one more time to avoid HW timing issue. This is
		 * a short term workaround solution, and must be removed once
		 * hardware fixing is done.
		 */
		v = readq(base + INDIRECT_RD_DATA);

		*value++ = v & INDIRECT_DATA_MASK;
	}

	return 0;
}

static int indirect_read(void *context, const void *reg_buf, size_t reg_len,
		  void *val_buf, size_t val_len)
{
	u32 reg = *(u32 *)reg_buf;
	u16 count = val_len / VAL_SIZE;

	//printk("%s: reg=0x%x, count=%d\n", __func__, reg, count);

	BUG_ON(reg_len != 4);
	BUG_ON(val_len != 4);
	BUG_ON(!reg_buf);
	BUG_ON(!val_buf);

	if (reg_len != REG_SIZE || !val_len)
		return -EINVAL;

	return do_reg_read(context, reg, val_buf, count);
}
static const struct regmap_bus indirect_regbus = {
	.gather_write = indirect_gather_write,
	.write = indirect_write,
	.read = indirect_read,
	.reg_format_endian_default = REGMAP_ENDIAN_NATIVE,
	.val_format_endian_default = REGMAP_ENDIAN_NATIVE,
};

#define SPI_CORE_PARAMETER	0x8
#define SHIFT_MODE		BIT_ULL(1)
#define SHIFT_MODE_MSB		0
#define SHIFT_MODE_LSB		1
#define DATA_WIDTH		GENMASK_ULL(7, 2)
#define NUM_CHIPSELECT		GENMASK_ULL(13, 8)
#define CLK_POLARITY		BIT_ULL(14)
#define CLK_PHASE		BIT_ULL(15)
#define PERIPHERAL_ID		GENMASK_ULL(47, 32)
#define SPI_CLK                 GENMASK_ULL(31, 22)
#define SPI_INDIRECT_ACC_OFST	0x10

static int max10_spi_init(struct platform_device *pdev, struct feature *feature)
{
	struct altera_spi_platform_data pdata;
	void __iomem *base = feature->ioaddr;
	struct platform_device *subdev;
	struct spi_board_info *info;
	u64 v;

	dev_dbg(&pdev->dev, "FME MAX10 SPI Init\n");

	v = readq(base + SPI_CORE_PARAMETER);

	info = devm_kzalloc(&pdev->dev, sizeof(*info), GFP_KERNEL);
	if (!info)
		return -ENOMEM;

	snprintf(info->modalias, sizeof(info->modalias), "intel-max10");
	info->chip_select = 0;

	memset(&pdata, 0, sizeof(pdata));
	pdata.mode_bits = SPI_CS_HIGH;
	if (FIELD_GET(CLK_POLARITY, v))
		pdata.mode_bits |= SPI_CPOL;
	if (FIELD_GET(CLK_PHASE, v))
		pdata.mode_bits |= SPI_CPHA;

	pdata.num_chipselect = FIELD_GET(NUM_CHIPSELECT, v);
	pdata.bits_per_word_mask =
		SPI_BPW_RANGE_MASK(1, FIELD_GET(DATA_WIDTH, v));
	
#if LINUX_VERSION_CODE == KERNEL_VERSION(3,10,0)
	info->max_speed_hz = FIELD_GET(SPI_CLK, v) *1000*1000;
#endif

	pdata.num_devices = 1;
	pdata.devices = info;
	pdata.regmap_ctx = base + SPI_INDIRECT_ACC_OFST;
	pdata.regbus = indirect_regbus;

	subdev = feature_create_subdev(pdev, ALTERA_SPI_DRV_NAME, &pdata,
				       sizeof(pdata));
	if (IS_ERR(subdev))
		return PTR_ERR(subdev);

	dev_dbg(&pdev->dev, "%s cs %d bpm 0x%x mode 0x%x\n", __func__,
		pdata.num_chipselect, pdata.bits_per_word_mask,
		pdata.mode_bits);

	feature_set_priv(feature, subdev);
	return 0;
}

static void max10_spi_uinit(struct platform_device *pdev,
			    struct feature *feature)
{
	struct platform_device *subdev = feature_get_priv(feature);

	dev_dbg(&pdev->dev, "FME MAX10 SPI UInit\n");

	feature_destroy_subdev(subdev);
	feature_set_priv(feature, NULL);
}

static struct feature_ops max10_spi_ops = {
	.init = max10_spi_init,
	.uinit = max10_spi_uinit,
};

#define I2C_CORE_PARAMETER	0x8
#define FIFO_DEPTH		GENMASK_ULL(8, 0)
#define IN_CLK			GENMASK_ULL(19, 10)
#define BUS_MAX_FREQ		GENMASK_ULL(23, 20)
#define I2C_DEV_ID		GENMASK_ULL(55, 24)
#define I2C_INDIRECT_ACC_OFST	0x10

struct mac_rom_i2c_priv {
	struct platform_device *subdev;
	struct i2c_board_info info;
	int clk_id;
};

static int mac_rom_i2c_init(struct platform_device *pdev,
			    struct feature *feature)
{
	void __iomem *base = feature->ioaddr;
	struct altr_i2c_platform_data pdata;
	struct device *dev = &pdev->dev;
	struct mac_rom_i2c_priv *priv;
	const char *clk_name;
	int ret;
	u64 v;

	dev_dbg(dev, "FME MAC ROM I2C Init\n");

	priv = devm_kzalloc(dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	v = readq(base + I2C_CORE_PARAMETER);
	if (FIELD_GET(I2C_DEV_ID, v) != 0xEE011)
		return -ENODEV;

	priv->clk_id = alloc_fpga_id(FEATURE_ID, dev);
	if (priv->clk_id < 0)
		return priv->clk_id;

	clk_name = devm_kasprintf(dev, GFP_KERNEL, "i2c-altera-clk.%d",
				  priv->clk_id);
	if (!clk_name) {
		ret = -ENODEV;
		goto exit_free_id;
	}

	dev_dbg(dev, "%s clk_name %s\n", __func__, clk_name);

	snprintf(priv->info.type, sizeof(priv->info.type), "24c512");
	priv->info.addr = 0x51;

	memset(&pdata, 0, sizeof(pdata));
	pdata.num_devices = 1;
	pdata.devices = &priv->info;
	pdata.fifo_size = FIELD_GET(FIFO_DEPTH, v);
	pdata.bus_clk_rate = FIELD_GET(BUS_MAX_FREQ, v) ? 400000 : 100000;
	/* FIXME: some bug on 3.10 kernel, we cannt to get clk from clk_get_rate() */
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,6,0)
	pdata.i2c_clk = clk_register_fixed_rate(&pdev->dev, clk_name, NULL, 0,
						FIELD_GET(IN_CLK, v) * 1000000);
	if (IS_ERR(pdata.i2c_clk)) {
		ret = PTR_ERR(pdata.i2c_clk);
		goto exit_free_id;
	}
#endif /* LINUX_VERSION_CODE > KERNEL_VERSION(4,6,0) */

	pdata.regmap_ctx = base + I2C_INDIRECT_ACC_OFST;
	pdata.regbus = indirect_regbus;

	priv->subdev = feature_create_subdev(pdev, ALTERA_I2C_DRV_NAME, &pdata,
					     sizeof(pdata));
	if (IS_ERR(priv->subdev)) {
		ret = PTR_ERR(priv->subdev);
		goto exit_unregister_clk;
	}

	feature_set_priv(feature, priv);
	return 0;

exit_unregister_clk:
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,6,0)
	clk_unregister_fixed_rate(pdata.i2c_clk);
#endif /* LINUX_VERSION_CODE > KERNEL_VERSION(4,6,0) */
exit_free_id:
	free_fpga_id(FEATURE_ID, priv->clk_id);
	return ret;
}

static void mac_rom_i2c_uinit(struct platform_device *pdev,
			      struct feature *feature)
{
	struct mac_rom_i2c_priv *priv = feature_get_priv(feature);
	struct altr_i2c_platform_data *pdata;

	dev_dbg(&pdev->dev, "FME MAC ROM I2C UInit\n");

	pdata = dev_get_platdata(&priv->subdev->dev);
	feature_destroy_subdev(priv->subdev);
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,6,0)
	clk_unregister_fixed_rate(pdata->i2c_clk);
#endif /* LINUX_VERSION_CODE > KERNEL_VERSION(4,6,0) */
	free_fpga_id(FEATURE_ID, priv->clk_id);
	feature_set_priv(feature, NULL);
}

static struct feature_ops mac_rom_i2c_ops = {
	.init = mac_rom_i2c_init,
	.uinit = mac_rom_i2c_uinit,
};

/* same as SPI_CORE_PARAMETER */
#define NIOS_SPI_PARAM		0x8

#define NIOS_SPI_CTRL		0x10
#define CTRL_CMD		GENMASK_ULL(63, 62)
#define CMD_NOP			0
#define CMD_RD			1
#define CMD_WR			2
#define CTRL_ADDR		GENMASK_ULL(44, 32)
#define CTRL_WR_DATA		GENMASK_ULL(31, 0)

#define NIOS_SPI_STAT		0x18
#define STAT_RW_VAL		BIT_ULL(32)
#define STAT_RD_DATA		GENMASK_ULL(31, 0)

#define NIOS_INIT		0x1000
#define REQ_FEC_MODE		GENMASK(23, 8)
#define FEC_MODE_NO		0x0
#define FEC_MODE_KR		0x5555
#define FEC_MODE_RS		0xaaaa
#define NIOS_INIT_START		BIT(1)
#define NIOS_INIT_DONE		BIT(0)
#define NIOS_VERSION		0x1004
#define NIOS_VERSION_MAJOR	GENMASK(31, 28)
#define NIOS_VERSION_MINOR	GENMASK(27, 24)
#define NIOS_VERSION_PATCH	GENMASK(23, 20)
#define PKVL_A_MODE_STS		0x1020
#define PKVL_B_MODE_STS		0x1024

/*
 * HW team estimates that 1000 loops should be roughly enough, we extend
 * to 10000 loops for safety.
 */
#define NIOS_SPI_WAIT_TIMEOUT	10000

static int nios_indirect_reg_write(void *context, unsigned int reg,
				   unsigned int val)
{
	void __iomem *base = context;
	u64 v = 0;
	int loops = NIOS_SPI_WAIT_TIMEOUT;

	v |= FIELD_PREP(CTRL_CMD, CMD_WR);
	v |= FIELD_PREP(CTRL_ADDR, reg);
	v |= FIELD_PREP(CTRL_WR_DATA, val);
	writeq(v, base + NIOS_SPI_CTRL);

	v = readq(base + NIOS_SPI_STAT);
	while (!(v & STAT_RW_VAL) && --loops) {
		cpu_relax();
		v = readq(base + NIOS_SPI_STAT);
	}

	return loops ? 0 : -ETIMEDOUT;
}

static int nios_indirect_reg_read(void *context, unsigned int reg,
				  unsigned int *val)
{
	void __iomem *base = context;
	u64 v = 0;
	int loops = NIOS_SPI_WAIT_TIMEOUT;

	v |= FIELD_PREP(CTRL_CMD, CMD_RD);
	v |= FIELD_PREP(CTRL_ADDR, reg);
	writeq(v, base + NIOS_SPI_CTRL);

	v = readq(base + NIOS_SPI_STAT);
	while (!(v & STAT_RW_VAL) && --loops) {
		cpu_relax();
		v = readq(base + NIOS_SPI_STAT);
	}

	*val = FIELD_GET(STAT_RD_DATA, v);

	return loops ? 0 : -ETIMEDOUT;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,16,0)
static int nios_indirect_gather_write(void *context,
                                     const void *reg, size_t reg_len,
                                     const void *val, size_t val_len)
{
        return -ENOTSUPP;
}

static int nios_indirect_read(void *context, const void *addr, size_t reg_size,
                void *val, size_t val_size)
{
        BUG_ON(!addr);
        BUG_ON(!val);
        BUG_ON(reg_size != 4);
        BUG_ON(val_size != 4);

        return nios_indirect_reg_read(context, *(u32 *)addr, val);
}

static int nios_indirect_write(void *context, const void *data, size_t count)
{
	unsigned int reg, val;

        BUG_ON(!data);

	reg = *(unsigned int *)data;
	val = *((unsigned int *)(data+4));

        if (WARN_ONCE(count < 4, "Invalid register access"))
                return -EINVAL;

        return nios_indirect_reg_write(context, reg,
                val);
}
#endif

static const struct regmap_bus nios_indirect_regbus = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,16,0)
	.reg_write = nios_indirect_reg_write,
	.reg_read = nios_indirect_reg_read,
#else
        .gather_write = nios_indirect_gather_write,
        .write = nios_indirect_write,
        .read = nios_indirect_read,
        .reg_format_endian_default = REGMAP_ENDIAN_NATIVE,
        .val_format_endian_default = REGMAP_ENDIAN_NATIVE,
#endif
};

static ssize_t nios_version_show(struct device *dev,
				 struct device_attribute *attr, char *buf)
{
	void __iomem *base;
	unsigned int val;
	int ret;

	base = get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_NIOS_SPI);

	ret = nios_indirect_reg_read(base, NIOS_VERSION, &val);
	if (ret)
		return ret;

	return sprintf(buf, "%x.%x.%x\n",
		       (u8)FIELD_GET(NIOS_VERSION_MAJOR, val),
		       (u8)FIELD_GET(NIOS_VERSION_MINOR, val),
		       (u8)FIELD_GET(NIOS_VERSION_PATCH, val));
}
static DEVICE_ATTR_RO(nios_version);

static ssize_t fec_mode_show(struct device *dev,
			     struct device_attribute *attr, char *buf)
{
	void __iomem *base;
	unsigned int val;
	int ret;

	base = get_feature_ioaddr_by_id(dev, FME_FEATURE_ID_NIOS_SPI);

	ret = nios_indirect_reg_read(base, NIOS_INIT, &val);
	if (ret)
		return ret;

	switch (FIELD_GET(REQ_FEC_MODE, val)) {
	case FEC_MODE_NO:
		return sprintf(buf, "no\n");
	case FEC_MODE_KR:
		return sprintf(buf, "kr\n");
	case FEC_MODE_RS:
		return sprintf(buf, "rs\n");
	}

	return -EFAULT;
}
static DEVICE_ATTR_RO(fec_mode);

static struct attribute *nios_spi_attrs[] = {
	&dev_attr_nios_version.attr,
	&dev_attr_fec_mode.attr,
	NULL,
};

static struct attribute_group nios_spi_attr_group = {
	.name = "nios_spi",
	.attrs = nios_spi_attrs,
};

static int init_error_detected(struct device *dev, void __iomem *base)
{
	unsigned int val;

	if (nios_indirect_reg_read(base, PKVL_A_MODE_STS, &val))
		return true;

	dev_dbg(dev, "PKVL A Mode Status 0x%x\n", val);

	if (val >= 0x100)
		return true;

	if (nios_indirect_reg_read(base, PKVL_B_MODE_STS, &val))
		return true;

	dev_dbg(dev, "PKVL B Mode Status 0x%x\n", val);

	if (val >= 0x100)
		return true;

	return false;
}

static int nios_spi_init_done_check(struct device *dev, void __iomem *base)
{
	unsigned long poll_timeout;
	unsigned int val;
	int ret;

	/*
	 * this SPI is shared by NIOS core inside FPGA, NIOS will use this SPI
	 * master to do some one time initialization after power up, and then
	 * release the control to OS. driver needs to poll on INIT_DONE to
	 * see when driver could take the control.
	 *
	 * Please note that after 3.x.x version, INIT_START is introduced, so
	 * driver needs to trigger START firstly and then check INIT_DONE.
	 */

	ret = nios_indirect_reg_read(base, NIOS_VERSION, &val);
	if (ret)
		return ret;

	/*
	 * If NIOS version register is totally uninitialized(== 0x0), then the
	 * NIOS firmware is missing. So host could take control of SPI master
	 * safely, but initialization work for NIOS is not done. This is an
	 * issue of FPGA image. We didn't error out because we need SPI master
	 * to reprogram a new image.
	 */
	if (val == 0) {
		dev_err(dev, "NIOS version reg = 0x%x, skip INIT_DONE check, but PKVL may be uninitialized\n",
			val);
		return 0;
	}

	if (FIELD_GET(NIOS_VERSION_MAJOR, val) >= 3) {
		/* read NIOS_INIT to check if PKVL INIT done or not */
		ret = nios_indirect_reg_read(base, NIOS_INIT, &val);
		if (ret)
			return ret;

		/* check if PKVLs are initialized already */
		if (val & NIOS_INIT_DONE || val & NIOS_INIT_START)
			goto nios_init_done;

		/* configure FEC mode per module param */
		val = NIOS_INIT_START;

		/* FEC mode will be ignored by hardware in 10G mode */
		if (!strcmp(fec_mode, "no"))
			val |= FIELD_PREP(REQ_FEC_MODE, FEC_MODE_NO);
		else if (!strcmp(fec_mode, "kr"))
			val |= FIELD_PREP(REQ_FEC_MODE, FEC_MODE_KR);
		else if (!strcmp(fec_mode, "rs"))
			val |= FIELD_PREP(REQ_FEC_MODE, FEC_MODE_RS);
		else
			return -EINVAL;

		ret = nios_indirect_reg_write(base, NIOS_INIT, val);
		if (ret)
			return ret;
	}

nios_init_done:
	/* polls on NIOS_INIT_DONE */
	poll_timeout = jiffies + msecs_to_jiffies(10000);
	while (1) {
		ret = nios_indirect_reg_read(base, NIOS_INIT, &val);
		if (ret)
			return ret;

		if (val & NIOS_INIT_DONE)
			break;

		if (!time_after(jiffies, poll_timeout)) {
			msleep(100);
			continue;
		}

		dev_err(dev, "NIOS_INIT_DONE timedout\n");
		nios_indirect_reg_read(base, PKVL_A_MODE_STS, &val);
		dev_err(dev, "PKVL_A_MODE_STS %x\n", val);
		nios_indirect_reg_read(base, PKVL_B_MODE_STS, &val);
		dev_err(dev, "PKVL_B_MODE_STS %x\n", val);
		return -ETIMEDOUT;
	}

	dev_dbg(dev, "INIT_DONE detected!\n");

	/*
	 * after INIT_DONE is detected, it still need to check if any ERR
	 * detected.
	 */
	if (init_error_detected(dev, base))
		dev_err(dev, "INIT_DONE, but error found during init\n");

	return 0;
}

static int nios_spi_init(struct platform_device *pdev, struct feature *feature)
{
	struct altera_spi_platform_data pdata;
	void __iomem *base = feature->ioaddr;
	struct platform_device *subdev;
	struct spi_board_info *info;
	int ret;
	u64 v;

	dev_dbg(&pdev->dev, "FME NIOS SPI Init\n");

	ret = nios_spi_init_done_check(&pdev->dev, base);
	if (ret)
		return ret;

	info = devm_kzalloc(&pdev->dev, sizeof(*info), GFP_KERNEL);
	if (!info)
		return -ENOMEM;

	snprintf(info->modalias, sizeof(info->modalias), "intel-max10");
	info->chip_select = 0;

	v = readq(base + NIOS_SPI_PARAM);

	memset(&pdata, 0, sizeof(pdata));
	pdata.mode_bits = SPI_CS_HIGH;
	if (FIELD_GET(CLK_POLARITY, v))
		pdata.mode_bits |= SPI_CPOL;
	if (FIELD_GET(CLK_PHASE, v))
		pdata.mode_bits |= SPI_CPHA;

	pdata.num_chipselect = FIELD_GET(NUM_CHIPSELECT, v);
	pdata.bits_per_word_mask =
		SPI_BPW_RANGE_MASK(1, FIELD_GET(DATA_WIDTH, v));

#if LINUX_VERSION_CODE == KERNEL_VERSION(3,10,0)
	info->max_speed_hz = FIELD_GET(SPI_CLK, v) *1000*1000;
#endif

	pdata.num_devices = 1;
	pdata.devices = info;
	pdata.regmap_ctx = base;
	pdata.regbus = nios_indirect_regbus;

	ret = sysfs_create_group(&pdev->dev.kobj, &nios_spi_attr_group);
	if (ret)
		return ret;

	subdev = feature_create_subdev(pdev, ALTERA_SPI_DRV_NAME, &pdata,
				       sizeof(pdata));
	if (IS_ERR(subdev)) {
		sysfs_remove_group(&pdev->dev.kobj, &nios_spi_attr_group);
		return PTR_ERR(subdev);
	}

	dev_dbg(&pdev->dev, "%s cs %d bpm 0x%x mode 0x%x\n", __func__,
		pdata.num_chipselect, pdata.bits_per_word_mask,
		pdata.mode_bits);

	feature_set_priv(feature, subdev);
	return 0;
}

static void nios_spi_uinit(struct platform_device *pdev,
			   struct feature *feature)
{
	struct platform_device *subdev = feature_get_priv(feature);

	dev_dbg(&pdev->dev, "FME NIOS SPI UInit\n");

	feature_destroy_subdev(subdev);
	feature_set_priv(feature, NULL);
	sysfs_remove_group(&pdev->dev.kobj, &nios_spi_attr_group);
}

static struct feature_ops nios_spi_ops = {
	.init = nios_spi_init,
	.uinit = nios_spi_uinit,
};

/* FIXME: wrong private feature id used in current VC hardware */
static struct feature_driver fme_feature_drvs[] = {
	{FEATURE_DRV(FME_FEATURE_ID_HEADER, &fme_hdr_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_THERMAL_MGMT, &thermal_mgmt_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_POWER_MGMT, &power_mgmt_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_GLOBAL_ERR, &global_error_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_PR_MGMT, &pr_mgmt_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_GLOBAL_IPERF, &global_iperf_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_HSSI_ETH, &hssi_mgmt_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_GLOBAL_DPERF, &global_dperf_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_QSPI_FLASH, &qspi_flash_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_EMIF_MGMT, &emif_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_PAC_HSSI_ETH, &pac_hssi_mgmt_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_S10_SDM_MB, &s10_sdm_mgmt_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_AVMMI_BMC, &avmmi_bmc_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_NIOS_SPI, &nios_spi_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_MAX10_SPI, &max10_spi_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_MAC_ROM_I2C, &mac_rom_i2c_ops),},
	{FEATURE_DRV(FME_FEATURE_ID_ETH_GROUP, &eth_group_ops),},
	{0, 0,}
};

static long fme_ioctl_check_extension(struct feature_platform_data *pdata,
				     unsigned long arg)
{
	/* No extension support for now */
	return 0;
}

static long
fme_ioctl_get_info(struct feature_platform_data *pdata, void __user *arg)
{
	struct fpga_fme_info info;
	struct fpga_fme *fme;
	unsigned long minsz;

	minsz = offsetofend(struct fpga_fme_info, capability);

	if (copy_from_user(&info, arg, minsz))
		return -EFAULT;

	if (info.argsz < minsz)
		return -EINVAL;

	mutex_lock(&pdata->lock);
	fme = fpga_pdata_get_private(pdata);
	info.flags = 0;
	info.capability = fme->capability;
	mutex_unlock(&pdata->lock);

	if (copy_to_user(arg, &info, sizeof(info)))
		return -EFAULT;

	return 0;
}

static int fme_ioctl_config_port(struct feature_platform_data *pdata,
				 u32 port_id, u32 flags, bool is_release)
{
	struct platform_device *fme_pdev = pdata->dev;
	struct feature_fme_header *fme_hdr;
	struct feature_fme_capability capability;

	if (flags)
		return -EINVAL;

	fme_hdr = get_feature_ioaddr_by_id(
		&fme_pdev->dev, FME_FEATURE_ID_HEADER);
	capability.csr = readq(&fme_hdr->capability);

	if (port_id >= capability.num_ports)
		return -EINVAL;

	return pdata->config_port(fme_pdev, port_id, is_release);
}

static long fme_ioctl_release_port(struct feature_platform_data *pdata,
				   void __user *arg)
{
	struct fpga_fme_port_release release;
	unsigned long minsz;

	minsz = offsetofend(struct fpga_fme_port_release, port_id);

	if (copy_from_user(&release, arg, minsz))
		return -EFAULT;

	if (release.argsz < minsz)
		return -EINVAL;

	return fme_ioctl_config_port(pdata, release.port_id,
				     release.flags, true);
}

static long fme_ioctl_assign_port(struct feature_platform_data *pdata,
				  void __user *arg)
{
	struct fpga_fme_port_assign assign;
	unsigned long minsz;

	minsz = offsetofend(struct fpga_fme_port_assign, port_id);

	if (copy_from_user(&assign, arg, minsz))
		return -EFAULT;

	if (assign.argsz < minsz)
		return -EINVAL;

	return fme_ioctl_config_port(pdata, assign.port_id,
				     assign.flags, false);
}

static int fme_open(struct inode *inode, struct file *filp)
{
	struct platform_device *fdev = fpga_inode_to_feature_dev(inode);
	struct feature_platform_data *pdata = dev_get_platdata(&fdev->dev);
	int ret;

	if (WARN_ON(!pdata))
		return -ENODEV;

	if (filp->f_flags & O_EXCL)
		ret = feature_dev_use_excl_begin(pdata);
	else
		ret = feature_dev_use_begin(pdata);

	if (ret)
		return ret;

	dev_dbg(&fdev->dev, "Device File Opened %d Times\n", pdata->open_count);
	filp->private_data = pdata;
	return 0;
}

static int fme_release(struct inode *inode, struct file *filp)
{
	struct feature_platform_data *pdata = filp->private_data;
	struct platform_device *pdev = pdata->dev;
	struct feature *feature;

	dev_dbg(&pdev->dev, "Device File Release\n");
	mutex_lock(&pdata->lock);
	__feature_dev_use_end(pdata);

	if (!pdata->open_count) {
		fpga_dev_for_each_feature(pdata, feature)
			fpga_msix_set_block(feature, 0, feature->ctx_num, NULL);
	}
	mutex_unlock(&pdata->lock);

	return 0;
}

static long fme_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct feature_platform_data *pdata = filp->private_data;
	struct platform_device *pdev = pdata->dev;
	struct feature *f;
	long ret;

	dev_dbg(&pdev->dev, "%s cmd 0x%x\n", __func__, cmd);

	switch (cmd) {
	case FPGA_GET_API_VERSION:
		return FPGA_API_VERSION;
	case FPGA_CHECK_EXTENSION:
		return fme_ioctl_check_extension(pdata, arg);
	case FPGA_FME_GET_INFO:
		return fme_ioctl_get_info(pdata, (void __user *)arg);
	case FPGA_FME_PORT_RELEASE:
		return fme_ioctl_release_port(pdata, (void __user *)arg);
	case FPGA_FME_PORT_ASSIGN:
		return fme_ioctl_assign_port(pdata, (void __user *)arg);
	default:
		/*
		 * Let sub-feature's ioctl function to handle the cmd
		 * Sub-feature's ioctl returns -ENODEV when cmd is not
		 * handled in this sub feature, and returns 0 and other
		 * error code if cmd is handled.
		 */
		fpga_dev_for_each_feature(pdata, f) {
			if (f->ops && f->ops->ioctl) {
				ret = f->ops->ioctl(pdev, f, cmd, arg);
				if (ret == -ENODEV)
					continue;
				else
					return ret;
			}
		}
	}

	return -EINVAL;
}

static const struct file_operations fme_fops = {
	.owner		= THIS_MODULE,
	.open		= fme_open,
	.release	= fme_release,
	.unlocked_ioctl = fme_ioctl,
};

static int fme_dev_init(struct platform_device *pdev)
{
	struct feature_platform_data *pdata = dev_get_platdata(&pdev->dev);
	struct fpga_fme *fme;

	fme = devm_kzalloc(&pdev->dev, sizeof(*fme), GFP_KERNEL);
	if (!fme)
		return -ENOMEM;

	fme->pdata = pdata;

	mutex_lock(&pdata->lock);
	fpga_pdata_set_private(pdata, fme);
	mutex_unlock(&pdata->lock);
	return 0;
}

static void fme_dev_destroy(struct platform_device *pdev)
{
	struct feature_platform_data *pdata = dev_get_platdata(&pdev->dev);
	struct fpga_fme *fme;

	mutex_lock(&pdata->lock);
	fme = fpga_pdata_get_private(pdata);
	fpga_pdata_set_private(pdata, NULL);
	mutex_unlock(&pdata->lock);

	devm_kfree(&pdev->dev, fme);
}

static int fme_probe(struct platform_device *pdev)
{
	int ret;

	ret = fme_dev_init(pdev);
	if (ret)
		goto exit;

	ret = fpga_dev_feature_init(pdev, fme_feature_drvs);
	if (ret)
		goto dev_destroy;

	ret = fpga_register_dev_ops(pdev, &fme_fops, THIS_MODULE);
	if (ret)
		goto feature_uinit;

	return 0;

feature_uinit:
	fpga_dev_feature_uinit(pdev);
dev_destroy:
	fme_dev_destroy(pdev);
exit:
	return ret;
}

static int fme_remove(struct platform_device *pdev)
{
	fpga_dev_feature_uinit(pdev);
	fpga_unregister_dev_ops(pdev);
	fme_dev_destroy(pdev);
	return 0;
}

static struct platform_driver fme_driver = {
	.driver	= {
		.name    = FPGA_FEATURE_DEV_FME,
	},
	.probe   = fme_probe,
	.remove  = fme_remove,
};

module_platform_driver(fme_driver);

MODULE_DESCRIPTION("FPGA Management Engine driver");
MODULE_AUTHOR("Intel Corporation");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS("platform:intel-fpga-fme");
