// SPDX-License-Identifier: GPL-2.0
/*
 * Intel Security Manager for FPGA
 *
 * Copyright (C) 2019 Intel Corporation, Inc.
 */

#include <linux/fpga/ifpga-sec-mgr.h>
#include <linux/idr.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/fs.h>
#include <uapi/linux/ifpga-sec-mgr.h>

#include <linux/version.h>

static DEFINE_IDA(ifpga_sec_mgr_ida);
static struct class *ifpga_sec_mgr_class;
static dev_t ifpga_sec_devt;

/**
 * ifpga_sec_mgr_create - create and initialize a IFPGA security manager struct
 *
 * @dev: create ifpga security manager device from pdev
 * @name: ifpga security manager name
 * @iops: pointer to structure of ifpga security manager ops
 * @priv: ifpga security manager private data
 *
 * The caller of this function is responsible for freeing the struct with
 * ifpga_sec_mgr_free().
 *
 * Return: pointer to struct ifpga_sec_mgr or NULL
 */
struct ifpga_sec_mgr *ifpga_sec_mgr_create(struct device *dev, const char *name,
					   const struct ifpga_sec_mgr_ops *iops,
					   void *priv)
{
	struct ifpga_sec_mgr *imgr;
	int id, ret;

	if (!iops || !iops->start || !iops->write_blk || !iops->data_sent ||
	    !iops->check_complete || !iops->cancel) {
		dev_err(dev, "Attempt to register without ifpga_sec_mgr_ops\n");
		return NULL;
	}

	if (!name || !strlen(name)) {
		dev_err(dev, "Attempt to register with no name!\n");
		return NULL;
	}

	imgr = kzalloc(sizeof(*imgr), GFP_KERNEL);
	if (!imgr)
		return NULL;

	id = ida_simple_get(&ifpga_sec_mgr_ida, 0, 0, GFP_KERNEL);
	if (id < 0) {
		ret = id;
		goto exit_kfree;
	}

	imgr->name = name;
	imgr->iops = iops;
	imgr->priv = priv;

	device_initialize(&imgr->dev);
	imgr->dev.class = ifpga_sec_mgr_class;
	imgr->dev.parent = dev;
	imgr->dev.id = id;
	imgr->dev.devt = MKDEV(MAJOR(ifpga_sec_devt), id);

	ret = dev_set_name(&imgr->dev, "ifpga_sec%d", id);
	if (ret)
		goto exit_remove_ida;

	mutex_init(&imgr->lock);
	return imgr;

exit_remove_ida:
	ida_simple_remove(&ifpga_sec_mgr_ida, id);
exit_kfree:
	kfree(imgr);

	return NULL;
}
EXPORT_SYMBOL_GPL(ifpga_sec_mgr_create);

/**
 * ifpga_sec_mgr_free - free a security mgr created with ifpga_sec_mgr_create()
 * @imgr: ifpga security manager struct
 */
void ifpga_sec_mgr_free(struct ifpga_sec_mgr *imgr)
{
	mutex_destroy(&imgr->lock);
	ida_simple_remove(&ifpga_sec_mgr_ida, imgr->dev.id);
	kfree(imgr);
}
EXPORT_SYMBOL_GPL(ifpga_sec_mgr_free);

static int ifpga_sec_mgr_open(struct inode *inode, struct file *filp)
{
	struct ifpga_sec_mgr *imgr = container_of(inode->i_cdev,
						  struct ifpga_sec_mgr, cdev);
	if (test_and_set_bit(0, &imgr->opened))
		return -EBUSY;

	filp->private_data = imgr;

	return 0;
}

#define	SIZE_4K		0x1000U
static int ifpga_sec_mgr_ioctl_write_blk(struct ifpga_sec_mgr *imgr,
					 unsigned long arg)
{
	struct ifpga_secure_write wb;
	unsigned long minsz;
	char *buf;
	int ret;

	minsz = offsetofend(struct ifpga_secure_write, buf);
	if (copy_from_user(&wb, (void __user *)arg, minsz))
		return -EFAULT;

	if (wb.argsz < minsz || wb.flags)
		return -EINVAL;

	/* Enforce 32-bit alignment on the write data */
	if ((wb.offset & 0x3) || (wb.size & 0x3))
		return -EINVAL;

	if (wb.size > SIZE_4K)
		return -ENOMEM;
 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0)
	if (!access_ok(u64_to_user_ptr(wb.buf), wb.size))
		return -EFAULT;
#else
	if (!access_ok(VERIFY_READ, u64_to_user_ptr(wb.buf), wb.size))
		return -EFAULT;
#endif

	buf = kzalloc(wb.size, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	if (copy_from_user(buf, u64_to_user_ptr(wb.buf), wb.size)) {
		ret = -EFAULT;
		goto done;
	}

	ret = imgr->iops->write_blk(imgr, &wb, buf);
done:
	kfree(buf);
	return ret;
}

static long ifpga_sec_mgr_ioctl(struct file *filp, unsigned int cmd,
				unsigned long arg)
{
	struct ifpga_sec_mgr *imgr = filp->private_data;
	int ret = -ENOTTY;

	mutex_lock(&imgr->lock);

	switch (cmd) {
	case IFPGA_SECURE_UPDATE_START:
		ret = imgr->iops->start(imgr);
		break;
	case IFPGA_SECURE_UPDATE_WRITE_BLK:
		ret = ifpga_sec_mgr_ioctl_write_blk(imgr, arg);
		break;
	case IFPGA_SECURE_UPDATE_DATA_SENT:
		ret = imgr->iops->data_sent(imgr);
		break;
	case IFPGA_SECURE_UPDATE_CHECK_COMPLETE:
		ret = imgr->iops->check_complete(imgr);
		break;
	case IFPGA_SECURE_UPDATE_CANCEL:
		ret = imgr->iops->cancel(imgr);
		break;
	}

	mutex_unlock(&imgr->lock);

	return ret;
}

static int ifpga_sec_mgr_release(struct inode *inode, struct file *filp)
{
	struct ifpga_sec_mgr *imgr = filp->private_data;

	mutex_lock(&imgr->lock);
	imgr->iops->cancel(imgr);
	mutex_unlock(&imgr->lock);
	clear_bit(0, &imgr->opened);

	return 0;
}

static const struct file_operations ifpga_sec_mgr_fops = {
	.owner = THIS_MODULE,
	.open = ifpga_sec_mgr_open,
	.release = ifpga_sec_mgr_release,
	.unlocked_ioctl = ifpga_sec_mgr_ioctl,
};

/**
 * ifpga_sec_mgr_register - register a IFPGA security manager
 *
 * @mgr: ifpga security manager struct
 *
 * Return: 0 on success, negative error code otherwise.
 */
int ifpga_sec_mgr_register(struct ifpga_sec_mgr *imgr, struct module *owner)
{
	const struct ifpga_sec_mgr_ops *iops = imgr->iops;
	int ret;

	/* start, write_blk, data_sent and check_complete
	 * must be implemented.
	 */
	if (!iops || !iops->start || !iops->write_blk ||
	    !iops->data_sent || !iops->check_complete)
		return -EINVAL;

	ret = device_add(&imgr->dev);
	if (ret)
		return ret;

	cdev_init(&imgr->cdev, &ifpga_sec_mgr_fops);
	imgr->cdev.owner = owner;
	imgr->cdev.kobj.parent = &imgr->dev.kobj;
	ret = cdev_add(&imgr->cdev, imgr->dev.devt, 1);
	if (ret)
		goto exit_device_del;

	dev_info(&imgr->dev, "%s registered\n", imgr->name);

	return 0;

exit_device_del:
	device_del(&imgr->dev);
	return ret;
}
EXPORT_SYMBOL_GPL(ifpga_sec_mgr_register);

/**
 * ifpga_sec_mgr_unregister - unregister a IFPGA security manager
 *
 * @mgr: fpga manager struct
 *
 * This function is intended for use in a IFPGA security manager
 * driver's remove() function.
 */
void ifpga_sec_mgr_unregister(struct ifpga_sec_mgr *imgr)
{
	dev_info(&imgr->dev, "%s %s\n", __func__, imgr->name);

	cdev_del(&imgr->cdev);
	device_unregister(&imgr->dev);
}
EXPORT_SYMBOL_GPL(ifpga_sec_mgr_unregister);

#define to_ifpga_sec_mgr(d) container_of(d, struct ifpga_sec_mgr, dev)

static void ifpga_sec_mgr_dev_release(struct device *dev)
{
	struct ifpga_sec_mgr *imgr = to_ifpga_sec_mgr(dev);

	mutex_destroy(&imgr->lock);
	ida_simple_remove(&ifpga_sec_mgr_ida, imgr->dev.id);
	kfree(imgr);
}

static int __init ifpga_sec_mgr_class_init(void)
{
	int ret;

	pr_info("Intel FPGA Security Manager\n");

	ifpga_sec_mgr_class = class_create(THIS_MODULE, "ifpga_sec_mgr");
	if (IS_ERR(ifpga_sec_mgr_class))
		return PTR_ERR(ifpga_sec_mgr_class);

	ret = alloc_chrdev_region(&ifpga_sec_devt, 0, MINORMASK,
				  "ifpga_sec_mgr");
	if (ret)
		goto exit_destroy_class;

	ifpga_sec_mgr_class->dev_release = ifpga_sec_mgr_dev_release;

	return 0;

exit_destroy_class:
	class_destroy(ifpga_sec_mgr_class);
	return ret;
}

static void __exit ifpga_sec_mgr_class_exit(void)
{
	unregister_chrdev_region(ifpga_sec_devt, MINORMASK);
	class_destroy(ifpga_sec_mgr_class);
	ida_destroy(&ifpga_sec_mgr_ida);
}

MODULE_DESCRIPTION("Intel FPGA Security Manager Driver");
MODULE_LICENSE("GPL v2");

subsys_initcall(ifpga_sec_mgr_class_init);
module_exit(ifpga_sec_mgr_class_exit)
