// SPDX-License-Identifier: GPL-2.0
/*
 * Driver for HSSI Configurable Ethernet support for Intel
 * Programmable Acceleration Cards with multiple QSFP interfaces.
 *
 * This is the version 2 implementation of HSSI, and is linked
 * into the same driver module with version 1. The primary difference
 * is that version 2 supports multiple QSFP interfaces and performs
 * functions that were previously performed by the NIOS processor.
 *
 * Copyright 2019 Intel Corporation, Inc.
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,4,0)
#include <linux/io-64-nonatomic-lo-hi.h>
#endif
#include <linux/iopoll-4-12.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include "pac-hssi.h"

#define HSSI_XCVR_CMD_CLEAR	0
#define HSSI_XCVR_CMD_READ	1
#define HSSI_XCVR_CMD_WRITE	2

#define HSSI_RESET_STAT		(GENMASK_ULL(3, 0)   | GENMASK_ULL(19, 16) | \
				 GENMASK_ULL(35, 32) | GENMASK_ULL(51, 48))
#define	HSSI_RESET_TIMEOUT	(GENMASK_ULL(5, 4)   | GENMASK_ULL(21, 20) | \
				 GENMASK_ULL(37, 36) | GENMASK_ULL(53, 52))

#define PER_QSFP_CHAN_CNT	4

/* Analog preemphasis tuning parameters */
#define PRE_TAP_ADDR			0x107
#define PRE_TAP_MAGNITUDE_MASK		GENMASK(4, 0)
#define PRE_TAP_MAX			15
#define PRE_TAP_POLARITY		BIT(5)	/* 1 = negative polarity */

#define POST_TAP_ADDR			0x105
#define POST_TAP_MAGNITUDE_MASK		GENMASK(4, 0)
#define POST_TAP_MAX			24
#define POST_TAP_POLARITY		BIT(6)	/* 1 = negative polarity */

#define VOD_COMP_ADDR			0x109
#define VOD_MASK			GENMASK(4, 0)
#define VOD_MIN				17
#define VOD_MAX				31

#define COMPENSATION_FLAG		BIT(5)	/* 1 = ON; 0 = OFF */

/* HSSI Capability - Read-Only */
struct hssi_capability {
	union {
		u64 csr;
		struct {
			u8  data_rate_avail_1g:1;   /* 1 = Enabled */
			u8  data_rate_avail_10g:1;
			u8  data_rate_avail_25g:1;
			u8  data_rate_avail_40g:1;
			u8  data_rate_avail_50g:1;
			u8  data_rate_avail_100g:1;
			u8  data_rate_avail_200g:1;
			u8  data_rate_avail_400g:1;
			u8  contains_pcs_1g:1;      /* 0 = PMA; 1 = PMA+PCS */
			u8  contains_pcs_10g:1;
			u8  contains_pcs_25g:1;
			u8  contains_pcs_40g:1;
			u8  contains_pcs_50g:1;
			u8  contains_pcs_100g:1;
			u8  contains_pcs_200g:1;
			u8  contains_pcs_400g:1;
			u8  contains_fec_1g:1;	    /* 1 = Enabled */
			u8  contains_fec_10g:1;
			u8  contains_fec_25g:1;
			u8  contains_fec_40g:1;
			u8  contains_fec_50g:1;
			u8  contains_fec_100g:1;
			u8  contains_fec_200g:1;
			u8  contains_fec_400g:1;
			u8  data_rate_switch:1;	    /* 1 = Enabled */
			u8  link_training:1;	    /* 1 = Enabled */
			u8  auto_negotiation:1;	    /* 1 = Enabled */
			u8  contains_mac:1;	    /* 1 = Enabled */
			u8  rsvd2:4;
			u8  num_qsfp_interfaces;    /* QSFP count */
			u32 rsvd3:24;
		};
	};
};

/* XCVR reconfig cmd and addr  */
struct hssi_rcfg_cmd_qsfp {
	union {
		u64 csr;
		struct {
			u8 command:2;		/* 0=Clr; 1=Read; 2=Write */
			u8 rsvd1:2;		/* Set to zero */
			u8 select_xcvr:2;       /* Select XCVR 0 - 3 */
			u16 rsvd2:10;		/* Set to zero */
			u16 xcvr_rcfg_addr:11;  /* PHY Xcvr or PLL address */
			u8 rsvd3:5;		/* Set to zero */
			u32 xcvr_rcfg_ack:1;	/* Reconfiguration ack */
			u32 rsvd4:31;		/* Set to zero */
		};
	};
};

/* XCVR reconfig read and write data  */
struct hssi_rcfg_data_qsfp {
	union {
		u64 csr;
		struct {
			u32 xcvr_rcfg_rdata;	/* RO: PHY or PLL rd data */
			u32 xcvr_rcfg_wdata;	/* RW: PHY or PLL wrt data */
		};
	};
};

/* HSSI control registers for QSFP interfaces */
struct hssi_ctrl_qsfp {
	union {
		u64 csr;
		struct {
			u8  data_rate_sel_1g:1;	    /* 1 = Selected */
			u8  data_rate_sel_10g:1;
			u8  data_rate_sel_25g:1;
			u8  data_rate_sel_40g:1;
			u8  data_rate_sel_50g:1;
			u8  data_rate_sel_100g:1;
			u8  data_rate_sel_200g:1;
			u8  data_rate_sel_400g:1;
			u8  global_reset:1;	  /* 1 = Active */
			u8  reconfig_reset:1;
			u8  chan0_reset:1;
			u8  chan1_reset:1;
			u8  chan2_reset:1;
			u8  chan3_reset:1;
			u8  select_atx_pll:1;	  /* 0 = 10G, 1 = 25G */
			u8  select_tx_core_clk:1; /* 0 = PHY, 1 = IOPLL */
			u8  select_rx_core_clk:1; /* 0 = PHY, 1 = IOPLL */
			u64 rsvd:47;
		};
	};
};

/* Per-QSFP-channel status (16 bits in size) */
struct hssi_stat_qsfp_chan {
	u8 tx_analog_rst_stat:1;
	u8 tx_dig_rst_stat:1;
	u8 rx_analog_rst_stat:1;
	u8 rx_dig_rst_stat:1;
	u8 tx_dig_rst_timeout:1;
	u8 rx_dig_rst_timeout:1;
	u8 tx_fifo_ready:1;
	u8 rx_fifo_ready:1;
	u8 tx_xfer_ready:1;
	u8 rx_xfer_ready:1;
	u8 tx_cal_busy:1;
	u8 rx_cal_busy:1;
	u8 rx_locked_to_data:1;
	u8 rx_locked_to_ref:1;
	u8 tx_ready:1;
	u8 rx_ready:1;
};

/* HSSI status registers for QSFP interfaces - Read Only */
struct hssi_stat_qsfp {
	union {
		u64 csr;
		struct hssi_stat_qsfp_chan chan[4];	/* 16 bits each */
	};
};

struct hssi_rcfg_qsfp {
	struct hssi_rcfg_cmd_qsfp cmd;
	struct hssi_rcfg_data_qsfp data;
	struct hssi_ctrl_qsfp ctrl;
	struct hssi_stat_qsfp stat;
};

/* FME HSSI VERSION 2 FEATURE */
struct hssi2_eth_regs {
	struct hssi_capability hssi_capability;
	struct hssi_rcfg_qsfp qsfp[];
};

/* Configuration parameters */
struct xcvr_config {
	u16 addr; /* Addresses are 11 bits */
	u8  mask;
	u8  data;
};

/* Configuration table for 10G */
static const struct xcvr_config config_10g[] = {
	{ 0x070, 0x0E, 0x00 },
	{ 0x086, 0x30, 0x30 },
	{ 0x087, 0x04, 0x04 },
	{ 0x088, 0x18, 0x00 },
	{ 0x089, 0x20, 0x20 },
	{ 0x08D, 0x04, 0x00 },
	{ 0x090, 0x02, 0x00 },
	{ 0x096, 0x8E, 0x80 },
	{ 0x0A0, 0x18, 0x00 },
	{ 0x0A2, 0xB8, 0xB8 },
	{ 0x0AC, 0x02, 0x02 },
	{ 0x0B0, 0x02, 0x00 },
	{ 0x110, 0x3F, 0x1A },
	{ 0x13F, 0x0F, 0x06 },
	{ 0x140, 0x07, 0x03 },
	{ 0x300, 0xE0, 0x00 },
	{ 0x301, 0xFF, 0x0E },
	{ 0x302, 0xDF, 0x02 },
	{ 0x303, 0xFF, 0x6A },
	{ 0x306, 0x30, 0x00 },
	{ 0x310, 0x07, 0x04 },
	{ 0x313, 0x3F, 0x04 },
	{ 0x314, 0x3F, 0x19 },
	{ 0x315, 0x07, 0x02 },
	{ 0x318, 0x02, 0x00 },
	{ 0x320, 0x07, 0x04 },
	{ 0x323, 0xC0, 0x00 }
};

/* Configuration table for 40G */
static const struct xcvr_config config_40g[] = {
	{ 0x070, 0x0E, 0x0E },
	{ 0x086, 0x30, 0x00 },
	{ 0x087, 0x04, 0x00 },
	{ 0x088, 0x18, 0x08 },
	{ 0x089, 0x20, 0x00 },
	{ 0x08D, 0x04, 0x04 },
	{ 0x090, 0x02, 0x02 },
	{ 0x096, 0x8E, 0x0E },
	{ 0x0A0, 0x18, 0x08 },
	{ 0x0A2, 0xB8, 0x00 },
	{ 0x0AC, 0x02, 0x00 },
	{ 0x0B0, 0x02, 0x02 },
	{ 0x110, 0x3F, 0x36 },
	{ 0x13F, 0x0F, 0x07 },
	{ 0x140, 0x07, 0x06 },
	{ 0x300, 0xE0, 0xA0 },
	{ 0x301, 0xFF, 0x5E },
	{ 0x302, 0xDF, 0xC6 },
	{ 0x303, 0xFF, 0xFA },
	{ 0x306, 0x30, 0x10 },
	{ 0x310, 0x07, 0x02 },
	{ 0x313, 0x3F, 0x0D },
	{ 0x314, 0x3F, 0x33 },
	{ 0x315, 0x07, 0x05 },
	{ 0x318, 0x02, 0x02 },
	{ 0x320, 0x07, 0x02 },
	{ 0x323, 0xC0, 0xC0 },
};

static int hssi_await_ack(struct pac_hssi *hssi, int qsfp_num)
{
	struct hssi2_eth_regs *hssi_regs = hssi->csr_base;
	struct hssi_rcfg_cmd_qsfp hssi_xcvr_cmd = { { 0 } };
	int err;

	/* Poll for the expected state of acknowlege bit */
	err = readq_poll_timeout(&hssi_regs->qsfp[qsfp_num].cmd.csr,
				 hssi_xcvr_cmd.csr,
				 hssi_xcvr_cmd.xcvr_rcfg_ack == 1,
				 HSSI_WRITE_POLL_INVL_US,
				 HSSI_WRITE_POLL_TIMEOUT_US);
	if (err) {
		dev_err(hssi->dev, "timeout, HSSI ack not received\n");
		return err;
	}

	hssi_xcvr_cmd.command = HSSI_XCVR_CMD_CLEAR; /* Clear ACK state */
	writeq(hssi_xcvr_cmd.csr, &hssi_regs->qsfp[qsfp_num].cmd);

	return 0;
}

struct qsfp_object {
	int index;
	struct kobject kobj;
	const struct attribute_group **attr_groups;
	struct list_head node;
	struct list_head cobj_list;
	void *priv;
};

struct chan_object {
	int index;
	int qsfp_index;
	struct kobject kobj;
	const struct attribute_group **attr_groups;
	struct list_head node;
	void *priv;
};

struct chan_obj_attribute {
	struct attribute attr;
	ssize_t (*show)(struct chan_object *cobj, char *buf);
	ssize_t (*store)(struct chan_object *cobj,
			 const char *buf, size_t n);
};

#define CHAN_OBJ_ATTR_RW(_name)				\
	struct chan_obj_attribute chan_obj_attr_##_name = __ATTR_RW(_name)

static int hssi_xcvr_read(struct pac_hssi *hssi, u8 qsfp_num, u8 chan_num,
			  u16 addr, u32 *data)
{
	struct hssi2_eth_regs *hssi_regs = hssi->csr_base;
	struct hssi_rcfg_cmd_qsfp hssi_xcvr_cmd = { { 0 } };
	struct hssi_rcfg_data_qsfp hssi_xcvr_data;
	int err;

	/* Read the desired address */
	hssi_xcvr_cmd.command = HSSI_XCVR_CMD_READ;
	hssi_xcvr_cmd.select_xcvr = chan_num;
	hssi_xcvr_cmd.xcvr_rcfg_addr = addr;
	writeq(hssi_xcvr_cmd.csr, &hssi_regs->qsfp[qsfp_num].cmd);

	/* Poll for ack */
	err = hssi_await_ack(hssi, qsfp_num);
	if (err)
		return err;

	hssi_xcvr_data.csr = readq(&hssi_regs->qsfp[qsfp_num].data);
	*data = hssi_xcvr_data.xcvr_rcfg_rdata;
	return 0;
}

static int hssi_xcvr_write(struct pac_hssi *hssi, u8 qsfp_num, u8 chan_num,
			   u16 addr, u32 data)
{
	struct hssi2_eth_regs *hssi_regs = hssi->csr_base;
	struct hssi_rcfg_cmd_qsfp hssi_xcvr_cmd = { { 0 } };
	struct hssi_rcfg_data_qsfp hssi_xcvr_data = { { 0 } };

	/* Set up the write data */
	hssi_xcvr_data.xcvr_rcfg_wdata = data;
	writeq(hssi_xcvr_data.csr, &hssi_regs->qsfp[qsfp_num].data);

	/* Trigger the write */
	hssi_xcvr_cmd.command = HSSI_XCVR_CMD_WRITE;
	hssi_xcvr_cmd.select_xcvr = chan_num;
	hssi_xcvr_cmd.xcvr_rcfg_addr = addr;
	writeq(hssi_xcvr_cmd.csr, &hssi_regs->qsfp[qsfp_num].cmd);

	/* Poll for ack */
	return hssi_await_ack(hssi, qsfp_num);
}

static int hssi_xcvr_rmw(struct pac_hssi *hssi, u8 qsfp_num, u8 chan_num,
			 u16 addr, u32 mask, u32 data)
{
	int err;
	u32 value;

	err = hssi_xcvr_read(hssi, qsfp_num, chan_num, addr, &value);
	if (err)
		return err;

	value &= ~mask;
	value |= (data & mask);

	return hssi_xcvr_write(hssi, qsfp_num, chan_num, addr, value);
}

static ssize_t tx_pre_tap_store(struct chan_object *cobj, const char *buf,
				size_t count)
{
	struct pac_hssi *hssi = cobj->priv;
	const char *p = buf;
	u8 polarity = 0;
	u8 magnitude;
	int err;

	if ((buf[0] == '+') || (buf[0] == '-')) {
		if (buf[0] == '-')
			polarity = PRE_TAP_POLARITY;
		p++;
	}

	err = kstrtou8(p, 0, &magnitude);
	if (err)
		return err;

	if (magnitude > PRE_TAP_MAX) {
		dev_err(hssi->dev, "Max pre-tap is %d\n", PRE_TAP_MAX);
		return -EINVAL;
	}

	mutex_lock(&hssi->bus_mutex);
	err = hssi_xcvr_rmw(hssi, cobj->qsfp_index, cobj->index,
			    PRE_TAP_ADDR,
			    PRE_TAP_POLARITY | PRE_TAP_MAGNITUDE_MASK,
			    polarity | magnitude);
	mutex_unlock(&hssi->bus_mutex);

	return err ? err : count;
}

static ssize_t tx_pre_tap_show(struct chan_object *cobj, char *buf)
{
	struct pac_hssi *hssi = cobj->priv;
	char polarity = '\0';
	u8 magnitude;
	u32 pre_tap;
	int err;

	mutex_lock(&hssi->bus_mutex);
	err = hssi_xcvr_read(hssi, cobj->qsfp_index, cobj->index,
			     PRE_TAP_ADDR, &pre_tap);
	mutex_unlock(&hssi->bus_mutex);

	if (err)
		return err;

	magnitude = pre_tap & PRE_TAP_MAGNITUDE_MASK;
	if (magnitude)
		polarity = pre_tap & PRE_TAP_POLARITY ? '-' : '+';

	return scnprintf(buf, PAGE_SIZE, "%c%u\n", polarity, magnitude);
}
static CHAN_OBJ_ATTR_RW(tx_pre_tap);

static ssize_t tx_post_tap_store(struct chan_object *cobj, const char *buf,
				 size_t count)
{
	struct pac_hssi *hssi = cobj->priv;
	const char *p = buf;
	u8 polarity = 0;
	u8 magnitude;
	int err;

	if ((buf[0] == '+') || (buf[0] == '-')) {
		if (buf[0] == '-')
			polarity = POST_TAP_POLARITY;
		p++;
	}

	err = kstrtou8(p, 0, &magnitude);
	if (err)
		return err;

	if (magnitude > POST_TAP_MAX) {
		dev_err(hssi->dev, "Max post-tap is %d\n", POST_TAP_MAX);
		return -EINVAL;
	}

	mutex_lock(&hssi->bus_mutex);
	err = hssi_xcvr_rmw(hssi, cobj->qsfp_index, cobj->index,
			    POST_TAP_ADDR,
			    POST_TAP_POLARITY | POST_TAP_MAGNITUDE_MASK,
			    polarity | magnitude);
	mutex_unlock(&hssi->bus_mutex);

	return err ? err : count;
}

static ssize_t tx_post_tap_show(struct chan_object *cobj, char *buf)
{
	struct pac_hssi *hssi = cobj->priv;
	char polarity = '\0';
	u8 magnitude;
	u32 post_tap;
	int err;

	mutex_lock(&hssi->bus_mutex);
	err = hssi_xcvr_read(hssi, cobj->qsfp_index, cobj->index,
			     POST_TAP_ADDR, &post_tap);
	mutex_unlock(&hssi->bus_mutex);

	if (err)
		return err;

	magnitude = post_tap & POST_TAP_MAGNITUDE_MASK;
	if (magnitude)
		polarity = post_tap & POST_TAP_POLARITY ? '-' : '+';

	return scnprintf(buf, PAGE_SIZE, "%c%u\n", polarity, magnitude);
}
static CHAN_OBJ_ATTR_RW(tx_post_tap);

static ssize_t tx_vod_store(struct chan_object *cobj, const char *buf,
			    size_t count)
{
	struct pac_hssi *hssi = cobj->priv;
	int err;
	u8 vod;

	err = kstrtou8(buf, 0, &vod);
	if (err)
		return err;

	if (vod > VOD_MAX) {
		dev_err(hssi->dev, "Max VOD is %d\n", VOD_MAX);
		return -EINVAL;
	} else if (vod < VOD_MIN) {
		dev_err(hssi->dev, "MIN VOD is %d\n", VOD_MIN);
		return -EINVAL;
	}

	mutex_lock(&hssi->bus_mutex);
	err = hssi_xcvr_rmw(hssi, cobj->qsfp_index, cobj->index,
			    VOD_COMP_ADDR, VOD_MASK, vod);
	mutex_unlock(&hssi->bus_mutex);

	return err ? err : count;
}

static ssize_t tx_vod_show(struct chan_object *cobj, char *buf)
{
	struct pac_hssi *hssi = cobj->priv;
	u32 vod;
	int err;

	mutex_lock(&hssi->bus_mutex);
	err = hssi_xcvr_read(hssi, cobj->qsfp_index, cobj->index,
			     VOD_COMP_ADDR, &vod);
	mutex_unlock(&hssi->bus_mutex);

	return err ? err : scnprintf(buf, PAGE_SIZE, "%lu\n", vod & VOD_MASK);
}
static CHAN_OBJ_ATTR_RW(tx_vod);

static ssize_t tx_comp_store(struct chan_object *cobj, const char *buf,
			     size_t count)
{
	struct pac_hssi *hssi = cobj->priv;
	int err;
	u8 compensation;

	err = kstrtou8(buf, 0, &compensation);
	if (err)
		return err;

	if (compensation != 0 && compensation != 1) {
		dev_err(hssi->dev, "Compensation must be 1 or 0");
		return -EINVAL;
	}

	mutex_lock(&hssi->bus_mutex);
	err = hssi_xcvr_rmw(hssi, cobj->qsfp_index, cobj->index,
			    VOD_COMP_ADDR, COMPENSATION_FLAG,
			    compensation ? COMPENSATION_FLAG : 0);
	mutex_unlock(&hssi->bus_mutex);

	return err ? err : count;
}

static ssize_t tx_comp_show(struct chan_object *cobj, char *buf)
{
	struct pac_hssi *hssi = cobj->priv;
	u32 compensation;
	int err;

	mutex_lock(&hssi->bus_mutex);
	err = hssi_xcvr_read(hssi, cobj->qsfp_index, cobj->index,
			     VOD_COMP_ADDR, &compensation);
	mutex_unlock(&hssi->bus_mutex);

	return err ? err : scnprintf(buf, PAGE_SIZE, "%u\n",
			 compensation & COMPENSATION_FLAG ? 1 : 0);
}
static CHAN_OBJ_ATTR_RW(tx_comp);

static struct attribute *chan_attrs[] = {
	&chan_obj_attr_tx_pre_tap.attr,
	&chan_obj_attr_tx_post_tap.attr,
	&chan_obj_attr_tx_vod.attr,
	&chan_obj_attr_tx_comp.attr,
	NULL,
};

static struct attribute_group chan_attr_group = {
	.attrs = chan_attrs,
};

static const struct attribute_group *chan_obj_attr_groups[] = {
	&chan_attr_group,
	NULL,
};

#define to_chan_obj_attr(_attr)					\
		container_of(_attr, struct chan_obj_attribute, attr)
#define to_chan_obj(_kobj) \
		container_of(_kobj, struct chan_object, kobj)

static ssize_t chan_obj_attr_show(struct kobject *kobj,
				  struct attribute *attr, char *buf)
{
	struct chan_obj_attribute *chan_attr = to_chan_obj_attr(attr);
	struct chan_object *cobj = to_chan_obj(kobj);
	ssize_t ret = -EIO;

	if (chan_attr->show)
		ret = chan_attr->show(cobj, buf);
	return ret;
};

static ssize_t chan_obj_attr_store(struct kobject *kobj,
				   struct attribute *attr, const char *buf,
				   size_t n)
{
	struct chan_obj_attribute *chan_attr = to_chan_obj_attr(attr);
	struct chan_object *cobj = to_chan_obj(kobj);
	ssize_t ret = -EIO;

	if (chan_attr->store)
		ret = chan_attr->store(cobj, buf, n);
	return ret;
}

static const struct sysfs_ops chan_obj_sysfs_ops = {
	.show = chan_obj_attr_show,
	.store = chan_obj_attr_store,
};

static void chan_obj_release(struct kobject *kobj)
{
	kfree(to_chan_obj(kobj));
}

static struct kobj_type chan_obj_ktype = {
	.sysfs_ops = &chan_obj_sysfs_ops,
	.release = chan_obj_release,
};

static struct chan_object *
create_chan_object(struct qsfp_object *qobj, int index,
		   const struct attribute_group **groups, const char *name)
{
	struct chan_object *cobj;
	int ret;

	cobj = kzalloc(sizeof(*cobj), GFP_KERNEL);
	if (!cobj)
		return ERR_PTR(-ENOMEM);

	cobj->index = index;
	cobj->qsfp_index = qobj->index;
	cobj->priv = qobj->priv;
	cobj->attr_groups = groups;
	INIT_LIST_HEAD(&cobj->node);

	ret = kobject_init_and_add(&cobj->kobj, &chan_obj_ktype, &qobj->kobj,
				   "%s%d", name, index);
	if (ret)
		goto put_exit;

	if (cobj->attr_groups) {
		ret = sysfs_create_groups(&cobj->kobj, cobj->attr_groups);
		if (ret)
			goto put_exit;
	}

	return cobj;

put_exit:
	kobject_put(&cobj->kobj);
	return ERR_PTR(ret);
}

struct qsfp_obj_attribute {
	struct attribute attr;
	ssize_t (*show)(struct qsfp_object *qobj, char *buf);
	ssize_t (*store)(struct qsfp_object *qobj,
			 const char *buf, size_t n);
};

#define QSFP_OBJ_ATTR_RO(_name)				\
	struct qsfp_obj_attribute qsfp_obj_attr_##_name = __ATTR_RO(_name)
#define QSFP_OBJ_ATTR_RW(_name)				\
	struct qsfp_obj_attribute qsfp_obj_attr_##_name = __ATTR_RW(_name)
#define QSFP_OBJ_ATTR_WO(_name)				\
	struct qsfp_obj_attribute qsfp_obj_attr_##_name = __ATTR_WO(_name)

static int create_chan_objects(struct qsfp_object *qobj)
{
	struct chan_object *cobj;
	u8 i;

	for (i = 0; i < PER_QSFP_CHAN_CNT; i++) {
		cobj = create_chan_object(qobj, i, chan_obj_attr_groups,
					  "chan");
		if (IS_ERR(cobj))
			return PTR_ERR(cobj);
		else if (cobj)
			list_add(&cobj->node, &qobj->cobj_list);
	}
	return 0;
}

static void remove_chan_objects(struct qsfp_object *qobj)
{
	struct chan_object *cobj, *cobj_tmp;

	list_for_each_entry_safe(cobj, cobj_tmp, &qobj->cobj_list, node) {
		list_del(&cobj->node);
		kobject_put(&cobj->kobj);
	}
}

static ssize_t ctrl_show(struct qsfp_object *qobj, char *buf)
{
	struct pac_hssi *hssi = qobj->priv;
	struct hssi2_eth_regs *hssi_regs = hssi->csr_base;
	u64 csr;

	mutex_lock(&hssi->bus_mutex);
	csr = readq(&hssi_regs->qsfp[qobj->index].ctrl);
	mutex_unlock(&hssi->bus_mutex);

	return scnprintf(buf, PAGE_SIZE, "0x%016llx\n", csr);
}

static ssize_t ctrl_store(struct qsfp_object *qobj, const char *buf,
			  size_t count)
{
	struct pac_hssi *hssi = qobj->priv;
	struct hssi2_eth_regs *hssi_regs = hssi->csr_base;
	u64 csr;
	int err;

	err = kstrtou64(buf, 0, &csr);
	if (err)
		return err;

	mutex_lock(&hssi->bus_mutex);
	writeq(csr, &hssi_regs->qsfp[qobj->index].ctrl);
	mutex_unlock(&hssi->bus_mutex);

	return count;
}
static QSFP_OBJ_ATTR_RW(ctrl);

static ssize_t stat_show(struct qsfp_object *qobj, char *buf)
{
	struct pac_hssi *hssi = qobj->priv;
	struct hssi2_eth_regs *hssi_regs = hssi->csr_base;
	u64 csr;

	mutex_lock(&hssi->bus_mutex);
	csr = readq(&hssi_regs->qsfp[qobj->index].stat);
	mutex_unlock(&hssi->bus_mutex);

	return scnprintf(buf, PAGE_SIZE, "0x%016llx\n", csr);
}
static QSFP_OBJ_ATTR_RO(stat);

static struct attribute *qsfp_attrs[] = {
	&qsfp_obj_attr_ctrl.attr,
	&qsfp_obj_attr_stat.attr,
	NULL,
};

static struct attribute_group qsfp_attr_group = {
	.attrs = qsfp_attrs,
};

static const struct attribute_group *qsfp_obj_attr_groups[] = {
	&qsfp_attr_group,
	NULL,
};

#define to_qsfp_obj_attr(_attr)					\
		container_of(_attr, struct qsfp_obj_attribute, attr)
#define to_qsfp_obj(_kobj) \
		container_of(_kobj, struct qsfp_object, kobj)

static ssize_t qsfp_obj_attr_show(struct kobject *kobj,
				  struct attribute *attr, char *buf)
{
	struct qsfp_obj_attribute *qsfp_attr = to_qsfp_obj_attr(attr);
	struct qsfp_object *qobj = to_qsfp_obj(kobj);
	ssize_t ret = -EIO;

	if (qsfp_attr->show)
		ret = qsfp_attr->show(qobj, buf);
	return ret;
};

static ssize_t qsfp_obj_attr_store(struct kobject *kobj,
				   struct attribute *attr, const char *buf,
				   size_t n)
{
	struct qsfp_obj_attribute *qsfp_attr = to_qsfp_obj_attr(attr);
	struct qsfp_object *qobj = to_qsfp_obj(kobj);
	ssize_t ret = -EIO;

	if (qsfp_attr->store)
		ret = qsfp_attr->store(qobj, buf, n);
	return ret;
}

static const struct sysfs_ops qsfp_obj_sysfs_ops = {
	.show = qsfp_obj_attr_show,
	.store = qsfp_obj_attr_store,
};

static void qsfp_obj_release(struct kobject *kobj)
{
	kfree(to_qsfp_obj(kobj));
}

static struct kobj_type qsfp_obj_ktype = {
	.sysfs_ops = &qsfp_obj_sysfs_ops,
	.release = qsfp_obj_release,
};

static struct qsfp_object *
create_qsfp_object(struct kobject *parent, int index,
		   const struct attribute_group **groups, const char *name,
		   void *priv)
{
	struct qsfp_object *qobj;
	int ret;

	qobj = kzalloc(sizeof(*qobj), GFP_KERNEL);
	if (!qobj)
		return ERR_PTR(-ENOMEM);

	qobj->index = index;
	qobj->priv = priv;
	qobj->attr_groups = groups;
	INIT_LIST_HEAD(&qobj->node);
	INIT_LIST_HEAD(&qobj->cobj_list);

	ret = kobject_init_and_add(&qobj->kobj, &qsfp_obj_ktype, parent,
				   "%s%d", name, index);
	if (ret)
		goto put_exit;

	if (qobj->attr_groups) {
		ret = sysfs_create_groups(&qobj->kobj, qobj->attr_groups);
		if (ret)
			goto put_exit;
	}

	ret = create_chan_objects(qobj);
	if (ret)
		goto put_exit;

	return qobj;

put_exit:
	kobject_put(&qobj->kobj);
	return ERR_PTR(ret);
}

static int create_qsfp_objects(struct kobject *parent, struct pac_hssi *hssi)
{
	struct hssi2_eth_regs *hssi_regs = hssi->csr_base;
	struct hssi_capability hssi_cap;
	struct qsfp_object *qobj;
	u8 i;

	hssi_cap.csr = readq(&hssi_regs->hssi_capability);
	for (i = 0; i < hssi_cap.num_qsfp_interfaces; i++) {
		qobj = create_qsfp_object(parent, i, qsfp_obj_attr_groups,
					  "qsfp", hssi);
		if (IS_ERR(qobj))
			return PTR_ERR(qobj);
		else if (qobj)
			list_add(&qobj->node, &hssi->qobj_list);
	}

	return 0;
}

static void remove_qsfp_objects(struct pac_hssi *hssi)
{
	struct qsfp_object *qobj, *qobj_tmp;

	list_for_each_entry_safe(qobj, qobj_tmp, &hssi->qobj_list, node) {
		remove_chan_objects(qobj);
		list_del(&qobj->node);
		kobject_put(&qobj->kobj);
	}
}

static ssize_t capability_show(struct device *dev,
			       struct device_attribute *attr, char *buf)
{
	struct pac_hssi *hssi = dev_get_drvdata(dev);
	struct hssi2_eth_regs *hssi_regs = hssi->csr_base;
	u64 csr;

	mutex_lock(&hssi->bus_mutex);
	csr = readq(&hssi_regs->hssi_capability);
	mutex_unlock(&hssi->bus_mutex);

	return scnprintf(buf, PAGE_SIZE, "0x%016llx\n", csr);
}
static DEVICE_ATTR_RO(capability);

static const struct attribute *hssi2_mgmt_attrs[] = {
	&dev_attr_capability.attr,
	NULL,
};

int hssi2_create_sysfs(struct device *dev)
{
	struct pac_hssi *hssi = dev_get_drvdata(dev);
	int ret;

	ret = create_qsfp_objects(&dev->kobj, hssi);
	if (ret)
		return ret;

	ret = sysfs_create_files(&dev->kobj, hssi2_mgmt_attrs);
	if (ret)
		remove_qsfp_objects(hssi);

	return ret;
}

void hssi2_remove_sysfs(struct device *dev)
{
	struct pac_hssi *hssi = dev_get_drvdata(dev);

	sysfs_remove_files(&dev->kobj, hssi2_mgmt_attrs);
	remove_qsfp_objects(hssi);
}
