/*
 * Driver for the PCIe device which locates between CPU and accelerated
 * function units(AFUs) and allows them to communicate with each other.
 *
 * Copyright 2016 Intel Corporation, Inc.
 *
 * Authors:
 *   Zhang Yi <Yi.Z.Zhang@intel.com>
 *   Xiao Guangrong <guangrong.xiao@linux.intel.com>
 *   Joseph Grecco <joe.grecco@intel.com>
 *   Enno Luebbers <enno.luebbers@intel.com>
 *   Tim Whisonant <tim.whisonant@intel.com>
 *   Ananda Ravuri <ananda.ravuri@intel.com>
 *   Mitchel, Henry <henry.mitchel@intel.com>
 *
 * This work is licensed under the terms of the GNU GPL version 2. See
 * the COPYING file in the top-level directory.
 *
 */

#include <linux/pci.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/stddef.h>
#include <linux/errno.h>
#include <linux/aer.h>
#include <linux/uuid.h>
#include <linux/kdev_t.h>
#include <linux/mfd/core.h>
#include <linux/mtd/altera-asmip2.h>

#include <linux/version.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0) || \
     (defined(RHEL_RELEASE_CODE) && \
        RHEL_RELEASE_CODE >= 0x0705)
#include <linux/sched/signal.h>
#else
#include <linux/sched.h>
#include <linux/idr.h>
#endif

#include "pac-hssi.h"
#include "pac-iopll.h"
#include "feature-dev.h"

#define DRV_VERSION	"0.14.0"
#define DRV_NAME	"intel-fpga-pci"

static DEFINE_MUTEX(fpga_id_mutex);

/* it is protected by fpga_id_mutex */
static struct idr fpga_ids[FPGA_ID_MAX];

enum cci_drv_state {
	CCI_DRV_STATE_RUNNING=0,
	CCI_DRV_STATE_PORTS_STOPPED=1,
};

struct cci_drvdata {
	int device_id;
	struct device *fme_dev;
	enum cci_drv_state drv_state;
	struct mutex lock;
	struct list_head port_dev_list;
	/* number of released ports which can be configured as VF  */
	int released_port_num;

	struct list_head regions;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,8,0) || \
     (defined(RHEL_RELEASE_CODE) && \
        RHEL_RELEASE_CODE >= 0x0705)
#else
	struct msix_entry *msix_entries;
#endif
};

struct cci_pci_region {
	int bar;
	void __iomem *ioaddr;

	struct list_head node;
};

static void fpga_ids_init(void)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(fpga_ids); i++)
		idr_init(fpga_ids + i);
}

static void fpga_ids_destroy(void)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(fpga_ids); i++)
		idr_destroy(fpga_ids + i);
}

int alloc_fpga_id(enum fpga_id_type type, struct device *dev)
{
	int id;

	WARN_ON(type >= FPGA_ID_MAX);
	mutex_lock(&fpga_id_mutex);
	id = idr_alloc(fpga_ids + type, dev, 0, 0, GFP_KERNEL);
	mutex_unlock(&fpga_id_mutex);
	return id;
}
EXPORT_SYMBOL_GPL(alloc_fpga_id);

void free_fpga_id(enum fpga_id_type type, int id)
{
	WARN_ON(type >= FPGA_ID_MAX);
	mutex_lock(&fpga_id_mutex);
	idr_remove(fpga_ids + type, id);
	mutex_unlock(&fpga_id_mutex);
}
EXPORT_SYMBOL_GPL(free_fpga_id);

static void cci_pci_add_port_dev(struct pci_dev *pdev,
				 struct platform_device *port_dev)
{
	struct cci_drvdata *drvdata = dev_get_drvdata(&pdev->dev);
	struct feature_platform_data *pdata = dev_get_platdata(&port_dev->dev);

	mutex_lock(&drvdata->lock);
	list_add(&pdata->node, &drvdata->port_dev_list);
	get_device(&pdata->dev->dev);
	mutex_unlock(&drvdata->lock);
}

static void cci_pci_remove_port_devs(struct pci_dev *pdev)
{
	struct cci_drvdata *drvdata = dev_get_drvdata(&pdev->dev);
	struct feature_platform_data *pdata, *ptmp;

	mutex_lock(&drvdata->lock);
	list_for_each_entry_safe(pdata, ptmp, &drvdata->port_dev_list, node) {
		struct platform_device *port_dev = pdata->dev;

		/* the port should be unregistered first. */
		WARN_ON(device_is_registered(&port_dev->dev));
		list_del(&pdata->node);
		free_fpga_id(PORT_ID, port_dev->id);
		put_device(&port_dev->dev);
	}
	mutex_unlock(&drvdata->lock);
}

static struct platform_device *cci_pci_lookup_port_by_id(struct pci_dev *pdev,
							 int port_id)
{
	struct cci_drvdata *drvdata = dev_get_drvdata(&pdev->dev);
	struct feature_platform_data *pdata;

	list_for_each_entry(pdata, &drvdata->port_dev_list, node)
		if (fpga_port_check_id(pdata->dev, &port_id))
			return pdata->dev;

	return NULL;
}

/* info collection during feature dev build. */
struct build_feature_devs_info {
	struct pci_dev *pdev;

	/*
	 * PCI BAR mapping info. Parsing feature list starts from
	 * BAR 0 and switch to different BARs to parse Port
	 */
	void __iomem *ioaddr;
	void __iomem *ioend;
	int current_bar;

	/* points to FME header where the port offset is figured out. */
	void __iomem *pfme_hdr;

	/* the container device for all feature devices */
	struct device *parent_dev;

	/* current feature device */
	struct platform_device *feature_dev;

	struct list_head sub_features;
	int feature_num;
};

struct feature_info {
	u64 fid;
	struct resource mmio_res;
	void __iomem *ioaddr;
	struct list_head node;
	unsigned int vec_start;
	unsigned int vec_cnt;
};

static void cci_pci_release_regions(struct pci_dev *pdev)
{
	struct cci_drvdata *drvdata = dev_get_drvdata(&pdev->dev);
	struct cci_pci_region *tmp, *region;

	list_for_each_entry_safe(region, tmp, &drvdata->regions, node) {
		list_del(&region->node);
		if (region->ioaddr)
			pci_iounmap(pdev, region->ioaddr);
		devm_kfree(&pdev->dev, region);
	}
}

static void __iomem *cci_pci_ioremap_bar(struct pci_dev *pdev, int bar)
{
	struct cci_drvdata *drvdata = dev_get_drvdata(&pdev->dev);
	struct cci_pci_region *region;

	list_for_each_entry(region, &drvdata->regions, node)
		if (region->bar == bar) {
			dev_dbg(&pdev->dev, "BAR %d region exists\n", bar);
			return region->ioaddr;
		}

	region = devm_kzalloc(&pdev->dev, sizeof(*region), GFP_KERNEL);
	if (!region)
		return NULL;

	region->bar = bar;
	region->ioaddr = pci_ioremap_bar(pdev, bar);
	if (!region->ioaddr) {
		dev_err(&pdev->dev, "can't ioremap memory from BAR %d.\n", bar);
		devm_kfree(&pdev->dev, region);
		return NULL;
	}

	list_add(&region->node, &drvdata->regions);
	return region->ioaddr;
}

static int parse_start_from(struct build_feature_devs_info *binfo, int bar)
{
	binfo->ioaddr = cci_pci_ioremap_bar(binfo->pdev, bar);
	if (!binfo->ioaddr)
		return -ENOMEM;

	binfo->current_bar = bar;
	binfo->ioend = binfo->ioaddr + pci_resource_len(binfo->pdev, bar);
	return 0;
}

static int parse_start(struct build_feature_devs_info *binfo)
{
	/* fpga feature list starts from BAR 0 */
	return parse_start_from(binfo, 0);
}

/* switch the memory mapping to BAR# @bar */
static int parse_switch_to(struct build_feature_devs_info *binfo, int bar)
{
	return parse_start_from(binfo, bar);
}

static int attach_port_dev(struct platform_device *pdev, int port_id)
{
	struct feature_fme_header *fme_hdr;
	struct feature_fme_port port;
	struct device *pci_dev = fpga_feature_dev_to_pcidev(pdev);
	struct cci_drvdata *drvdata = dev_get_drvdata(pci_dev);
	struct platform_device *port_dev;
	int ret;

	fme_hdr = get_feature_ioaddr_by_id(&pdev->dev,  FME_FEATURE_ID_HEADER);

	mutex_lock(&drvdata->lock);
	port.csr = readq(&fme_hdr->port[port_id]);
	if (port.afu_access_control == FME_AFU_ACCESS_VF) {
		dev_dbg(&pdev->dev, "port_%d has already been turned to VF.\n",
			port_id);
		mutex_unlock(&drvdata->lock);
		return -EBUSY;
	}

	port_dev = cci_pci_lookup_port_by_id(to_pci_dev(pci_dev), port_id);
	if (!port_dev) {
		dev_err(&pdev->dev, "port_%d is not detected.\n", port_id);
		ret = -EINVAL;
		goto exit;
	}

	if (device_is_registered(&port_dev->dev)) {
		dev_dbg(pci_dev, "port_%d is not released.\n", port_id);
		ret = -EBUSY;
		goto exit;
	}

	dev_dbg(pci_dev, "now re-assign port_%d:%s\n", port_id, port_dev->name);

	ret = platform_device_add(port_dev);
	if (ret)
		goto exit;

	get_device(&port_dev->dev);
	feature_dev_use_end(dev_get_platdata(&port_dev->dev));
	drvdata->released_port_num--;
exit:
	mutex_unlock(&drvdata->lock);
	return ret;
}

static int detach_port_dev(struct platform_device *pdev, int port_id)
{
	struct device *dev = fpga_feature_dev_to_pcidev(pdev);
	struct cci_drvdata *drvdata = dev_get_drvdata(dev);
	struct platform_device *port_dev;
	int ret;

	mutex_lock(&drvdata->lock);
	port_dev = cci_pci_lookup_port_by_id(to_pci_dev(dev), port_id);
	if (!port_dev) {
		dev_err(&pdev->dev, "port_%d is not detected.\n", port_id);
		ret = -EINVAL;
		goto exit;
	}

	if (!device_is_registered(&port_dev->dev)) {
		dev_dbg(&pdev->dev,
		   "port_%d is released or already assigned a VF.\n", port_id);
		ret = -EBUSY;
		goto exit;
	}

	ret = feature_dev_use_excl_begin(dev_get_platdata(&port_dev->dev));
	if (ret)
		goto exit;

	platform_device_del(port_dev);
	put_device(&port_dev->dev);
	drvdata->released_port_num++;
exit:
	mutex_unlock(&drvdata->lock);
	return ret;
}

static int
config_port(struct platform_device *pdev, u32 port_id, bool release)
{
	/* Todo: some potential check */
	if (release)
		return detach_port_dev(pdev, port_id);

	return attach_port_dev(pdev, port_id);
}

static struct platform_device *fpga_for_each_port(struct platform_device *pdev,
		     void *data, int (*match)(struct platform_device *, void *))
{
	struct device *pci_dev = fpga_feature_dev_to_pcidev(pdev);
	struct cci_drvdata *drvdata = dev_get_drvdata(pci_dev);
	struct feature_platform_data *pdata;
	struct platform_device *port_dev;

	mutex_lock(&drvdata->lock);
	list_for_each_entry(pdata, &drvdata->port_dev_list, node) {
		port_dev = pdata->dev;

		if (match(port_dev, data) && get_device(&port_dev->dev))
			goto exit;
	}
	port_dev = NULL;
exit:
	mutex_unlock(&drvdata->lock);
	return port_dev;
}

static struct build_feature_devs_info *
build_info_alloc_and_init(struct pci_dev *pdev)
{
	struct build_feature_devs_info *binfo;

	binfo = devm_kzalloc(&pdev->dev, sizeof(*binfo), GFP_KERNEL);
	if (binfo)
		binfo->pdev = pdev;

	return binfo;
}

static enum fpga_id_type feature_dev_id_type(struct platform_device *pdev)
{
	if (!strcmp(pdev->name, FPGA_FEATURE_DEV_FME))
		return FME_ID;

	if (!strcmp(pdev->name, FPGA_FEATURE_DEV_PORT))
		return PORT_ID;

	WARN_ON(1);
	return FPGA_ID_MAX;
}

/*
 * register current feature device, it is called when we need to switch to
 * another feature parsing or we have parsed all features
 */
static int build_info_commit_dev(struct build_feature_devs_info *binfo)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,8,0) || \
     (defined(RHEL_RELEASE_CODE) && \
        RHEL_RELEASE_CODE >= 0x0705)
#else
	struct cci_drvdata *drvdata = dev_get_drvdata(&binfo->pdev->dev);
	struct msix_entry *msix_entries = drvdata->msix_entries;
#endif
	struct platform_device *fdev = binfo->feature_dev;
	struct feature_platform_data *pdata;
	struct feature_info *finfo, *p;
	int ret, index = 0;

	if (!fdev)
		return 0;

	/*
	 * we need not care the memory which is associated with the
	 * platform device. After call platform_device_unregister(),
	 * it will be automatically freed by device's
	 * release() callback, platform_device_release().
	 */
	pdata = kzalloc(feature_platform_data_size(binfo->feature_num),
			GFP_KERNEL);
	if (!pdata)
		return -ENOMEM;

	if (!strcmp(fdev->name, FPGA_FEATURE_DEV_FME)) {
		pdata->config_port = config_port;
		pdata->fpga_for_each_port = fpga_for_each_port;
	}
	pdata->dev = fdev;
	pdata->num = binfo->feature_num;
	INIT_LIST_HEAD(&pdata->pid_list);
	mutex_init(&pdata->lock);

	fdev->dev.platform_data = pdata;
	fdev->num_resources = binfo->feature_num;
	fdev->resource = kcalloc(binfo->feature_num, sizeof(*fdev->resource),
				 GFP_KERNEL);
	if (!fdev->resource)
		return -ENOMEM;

	/* fill features and resource information for feature dev */
	list_for_each_entry_safe(finfo, p, &binfo->sub_features, node) {
		struct feature *feature = &pdata->features[index];
		struct feature_irq_ctx *ctx = NULL;
		int i;

		/* save resource information for each feature */
		feature->pdev = fdev;
		feature->id = finfo->fid;
		feature->resource_index = index;
		feature->ioaddr = finfo->ioaddr;
		fdev->resource[index++] = finfo->mmio_res;

		if (finfo->vec_cnt) {
			ctx = devm_kcalloc(&binfo->pdev->dev, finfo->vec_cnt,
					   sizeof(*ctx), GFP_KERNEL);
			if (!ctx)
				return -ENOMEM;

			for (i = 0; i < finfo->vec_cnt; i++)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,8,0) || \
     (defined(RHEL_RELEASE_CODE) && \
        RHEL_RELEASE_CODE >= 0x0705)
				ctx[i].irq = pci_irq_vector(binfo->pdev,
							finfo->vec_start + i);
#else
				ctx[i].irq =
				     msix_entries[finfo->vec_start + i].vector;
#endif
			feature->ctx = ctx;
			feature->ctx_num = finfo->vec_cnt;
		}
		list_del(&finfo->node);
		kfree(finfo);
	}

	ret = platform_device_add(binfo->feature_dev);
	if (!ret) {
		struct cci_drvdata *drvdata;

		drvdata = dev_get_drvdata(&binfo->pdev->dev);
		if (feature_dev_id_type(binfo->feature_dev) == PORT_ID)
			cci_pci_add_port_dev(binfo->pdev, binfo->feature_dev);
		else
			drvdata->fme_dev = get_device(&binfo->feature_dev->dev);

		/*
		 * reset it to avoid build_info_free() freeing their resource.
		 *
		 * The resource of successfully registered feature devices
		 * will be freed by platform_device_unregister(). See the
		 * comments in build_info_create_dev().
		 */
		binfo->feature_dev = NULL;
	}

	return ret;
}

static int
build_info_create_dev(struct build_feature_devs_info *binfo,
		      enum fpga_id_type type, void __iomem *ioaddr)
{
	enum fpga_devt_type devt_type = FPGA_DEVT_FME;
	char *name = FPGA_FEATURE_DEV_FME;
	struct platform_device *fdev;
	int ret;

	if (type == PORT_ID) {
		devt_type = FPGA_DEVT_PORT;
		name = FPGA_FEATURE_DEV_PORT;
	}

	/* we will create a new device, commit current device first */
	ret = build_info_commit_dev(binfo);
	if (ret)
		return ret;

	/*
	 * we use -ENODEV as the initialization indicator which indicates
	 * whether the id need to be reclaimed
	 */
	fdev = platform_device_alloc(name, -ENODEV);
	if (!fdev)
		return -ENOMEM;

	binfo->feature_dev = fdev;
	binfo->feature_num = 0;
	INIT_LIST_HEAD(&binfo->sub_features);

	fdev->id = alloc_fpga_id(type, &fdev->dev);
	if (fdev->id < 0)
		return fdev->id;

	fdev->dev.parent = binfo->parent_dev;
	fdev->dev.devt = fpga_get_devt(devt_type, fdev->id);
	return 0;
}

static int remove_feature_dev(struct device *dev, void *data)
{
	struct platform_device *pdev = to_platform_device(dev);

	platform_device_unregister(pdev);
	return 0;
}

static int remove_parent_dev(struct device *dev, void *data)
{
	/* remove platform devices attached in the parent device */
	device_for_each_child(dev, NULL, remove_feature_dev);
	device_unregister(dev);
	return 0;
}

static void remove_all_devs(struct pci_dev *pdev)
{
	/* remove parent device and all its children. */
	device_for_each_child(&pdev->dev, NULL, remove_parent_dev);
}

static void build_info_free(struct build_feature_devs_info *binfo)
{
	struct feature_info *finfo, *p;

	if (!IS_ERR_OR_NULL(binfo->parent_dev))
		remove_all_devs(binfo->pdev);

	/*
	 * it is a valid id, free it. See comments in
	 * build_info_create_dev()
	 */
	if (binfo->feature_dev && binfo->feature_dev->id >= 0) {
		free_fpga_id(feature_dev_id_type(binfo->feature_dev),
			     binfo->feature_dev->id);

		list_for_each_entry_safe(finfo, p, &binfo->sub_features, node) {
			list_del(&finfo->node);
			kfree(finfo);
		}
	}

	platform_device_put(binfo->feature_dev);

	devm_kfree(&binfo->pdev->dev, binfo);
}

static inline u32 feature_size(void __iomem *start)
{
	struct feature_header header, *hdr = start;

	header.csr = readq(hdr);

	/* workaround for private features with invalid size, use 4K instead */
	return header.next_header_offset ? header.next_header_offset : 4096;
}

static u64 feature_id(void __iomem *start)
{
	struct feature_header header, *hdr = start;

	header.csr = readq(hdr);

	if (header.type == FEATURE_TYPE_FIU)
		return FEATURE_ID_FIU_HEADER;
	else if (header.type == FEATURE_TYPE_PRIVATE)
		return header.id;
	else if (header.type == FEATURE_TYPE_AFU)
		return FEATURE_ID_AFU;

	WARN_ON(1);
	return 0;
}

static int
create_feature_instance(struct build_feature_devs_info *binfo,
			void __iomem *start, u64 fid, resource_size_t size,
			unsigned int vec_start, unsigned int vec_cnt)
{
	struct feature_info *finfo;
	resource_size_t res_start;

	size = size ? size : feature_size(start);
	fid = fid ? fid : feature_id(start);
	res_start = pci_resource_start(binfo->pdev, binfo->current_bar);

	finfo = kzalloc(sizeof(*finfo), GFP_KERNEL);
	if (!finfo)
		return -ENOMEM;

	finfo->fid = fid;
	finfo->mmio_res.start = res_start + start - binfo->ioaddr;
	finfo->mmio_res.end = finfo->mmio_res.start + size - 1;
	finfo->mmio_res.flags = IORESOURCE_MEM;
	finfo->vec_start = vec_start;
	finfo->vec_cnt = vec_cnt;
	finfo->ioaddr = start;

	list_add_tail(&finfo->node, &binfo->sub_features);
	binfo->feature_num++;
	return 0;
}

static int parse_feature_fme(struct build_feature_devs_info *binfo,
			     void __iomem *start)
{
	int ret;

	ret = build_info_create_dev(binfo, FME_ID, start);
	if (ret)
		return ret;

	return create_feature_instance(binfo, start, 0, 0, 0, 0);
}

static void parse_feature_irqs(struct build_feature_devs_info *binfo,
			       void __iomem *start, unsigned int *vec_start,
			       unsigned int *vec_cnt)
{
	int cnt = pci_msix_vec_count(binfo->pdev);
	u64 id;

	if (cnt <= 0)
		return;

	id = feature_id(start);

	if (id == PORT_FEATURE_ID_UINT) {
		struct feature_port_uint *port_uint = start;
		struct feature_port_uint_cap uint_cap;

		uint_cap.csr = readq(&port_uint->capability);
		if (uint_cap.intr_num) {
			*vec_start = uint_cap.first_vec_num;
			*vec_cnt = uint_cap.intr_num;
		} else
			dev_dbg(&binfo->pdev->dev, "UAFU doesn't support interrupt\n");
	} else if (id == PORT_FEATURE_ID_ERROR) {
		struct feature_port_error *port_err = start;
		struct feature_port_err_capability port_err_cap;

		port_err_cap.csr = readq(&port_err->error_capability);
		if (port_err_cap.support_intr) {
			*vec_start = port_err_cap.intr_vector_num;
			*vec_cnt = 1;
		} else
			dev_dbg(&binfo->pdev->dev, "Port error doesn't support interrupt\n");
	} else if (id == FME_FEATURE_ID_GLOBAL_ERR) {
		struct feature_fme_err *fme_err = start;
		struct feature_fme_error_capability fme_err_cap;

		fme_err_cap.csr = readq(&fme_err->fme_err_capability);
		if (fme_err_cap.support_intr) {
			*vec_start = fme_err_cap.intr_vector_num;
			*vec_cnt = 1;
		} else
			dev_dbg(&binfo->pdev->dev, "FME error doesn't support interrupt\n");
	}

	if (*vec_start + *vec_cnt > cnt) {
		*vec_start = 0;
		*vec_cnt = 0;
		dev_err(&binfo->pdev->dev, "Inconsistent interrupt number on HW\n");
	}
}

static int parse_feature_fme_private(struct build_feature_devs_info *binfo,
				     struct feature_header *hdr)
{
	unsigned int vec_start = 0, vec_cnt = 0;

	parse_feature_irqs(binfo, hdr, &vec_start, &vec_cnt);

	return create_feature_instance(binfo, hdr, 0, 0, vec_start, vec_cnt);
}

static int parse_feature_port(struct build_feature_devs_info *binfo,
			      void __iomem *start)
{
	int ret;

	ret = build_info_create_dev(binfo, PORT_ID, start);
	if (ret)
		return ret;

	return create_feature_instance(binfo, start, 0, 0, 0, 0);
}

static int parse_feature_port_private(struct build_feature_devs_info *binfo,
				      struct feature_header *hdr)
{
	unsigned int vec_start = 0, vec_cnt = 0;

	parse_feature_irqs(binfo, hdr, &vec_start, &vec_cnt);

	return create_feature_instance(binfo, hdr, 0, 0, vec_start, vec_cnt);
}

static int parse_feature_port_afu(struct build_feature_devs_info *binfo,
				  struct feature_header *hdr)
{
	struct feature_port_header *port_hdr = binfo->ioaddr;
	struct feature_port_capability capability;

	capability.csr = readq(&port_hdr->capability);

	return create_feature_instance(binfo, hdr, PORT_FEATURE_ID_AFU,
				       capability.mmio_size << 10, 0, 0);
}

static int parse_feature_afu(struct build_feature_devs_info *binfo,
			     struct feature_header *hdr)
{
	if (!binfo->feature_dev) {
		dev_err(&binfo->pdev->dev, "this AFU does not belong to any FIU.\n");
		return -EINVAL;
	}

	switch (feature_dev_id_type(binfo->feature_dev)) {
	case PORT_ID:
		return parse_feature_port_afu(binfo, hdr);
	default:
		dev_info(&binfo->pdev->dev, "AFU belonging to FIU %s is not supported yet.\n",
			 binfo->feature_dev->name);
	}

	return 0;
}

static int parse_feature_fiu(struct build_feature_devs_info *binfo,
			     struct feature_header *hdr)
{
	struct feature_fiu_header *fiu_hdr, fiu_header;
	struct feature_header header;
	void __iomem *start = hdr;
	int ret;

	header.csr = readq(hdr);

	switch (header.id) {
	case FEATURE_FIU_ID_FME:
		ret = parse_feature_fme(binfo, hdr);
		binfo->pfme_hdr = hdr;
		if (ret)
			return ret;
		break;
	case FEATURE_FIU_ID_PORT:
		ret = parse_feature_port(binfo, hdr);
		if (ret)
			return ret;
		break;
	default:
		dev_info(&binfo->pdev->dev, "FIU TYPE %d is not supported yet.\n",
			 header.id);
	}

	/* Check FIU's next_afu pointer to AFU */
	fiu_hdr = (struct feature_fiu_header *)(hdr + 1);
	fiu_header.csr = readq(&fiu_hdr->csr);

	if (fiu_header.next_afu) {
		start += fiu_header.next_afu;
		return parse_feature_afu(binfo, start);
	}

	dev_dbg(&binfo->pdev->dev, "No AFUs detected on FIU %d\n",
		header.id);

	return 0;
}

static int parse_feature_private(struct build_feature_devs_info *binfo,
				 struct feature_header *hdr)
{
	struct feature_header header;

	header.csr = readq(hdr);

	if (!binfo->feature_dev) {
		dev_err(&binfo->pdev->dev, "the private feature %x does not belong to any AFU.\n",
			header.id);
		return -EINVAL;
	}

	switch (feature_dev_id_type(binfo->feature_dev)) {
	case FME_ID:
		return parse_feature_fme_private(binfo, hdr);
	case PORT_ID:
		return parse_feature_port_private(binfo, hdr);
	default:
		dev_info(&binfo->pdev->dev, "private feature %x belonging to AFU %s is not supported yet.\n",
			 header.id, binfo->feature_dev->name);
	}
	return 0;
}

static int parse_feature(struct build_feature_devs_info *binfo,
			 struct feature_header *hdr)
{
	struct feature_header header;
	int ret = 0;

	header.csr = readq(hdr);

	switch (header.type) {
	case FEATURE_TYPE_AFU:
		ret = parse_feature_afu(binfo, hdr);
		break;
	case FEATURE_TYPE_PRIVATE:
		ret = parse_feature_private(binfo, hdr);
		break;
	case FEATURE_TYPE_FIU:
		ret = parse_feature_fiu(binfo, hdr);
		break;
	default:
		dev_info(&binfo->pdev->dev,
			 "Feature Type %x is not supported.\n", hdr->type);
	};

	return ret;
}

static int
parse_feature_list(struct build_feature_devs_info *binfo, void __iomem *start)
{
	struct feature_header *hdr, header;
	void __iomem *end = binfo->ioend;
	int ret = 0;

	for (; start < end; start += header.next_header_offset) {
		if (end - start < sizeof(*hdr)) {
			dev_err(&binfo->pdev->dev, "The region is too small to contain a feature.\n");
			ret =  -EINVAL;
			break;
		}

		hdr = (struct feature_header *)start;
		ret = parse_feature(binfo, hdr);
		if (ret)
			break;

		header.csr = readq(hdr);
		if (header.eol || !header.next_header_offset)
			break;
	}

	return ret;
}

static int parse_ports_from_fme(struct build_feature_devs_info *binfo)
{
	struct feature_fme_header *fme_hdr;
	struct feature_fme_port port;
	int i = 0, ret = 0;

	if (binfo->pfme_hdr == NULL) {
		dev_dbg(&binfo->pdev->dev, "VF is detected.\n");
		return ret;
	}

	fme_hdr = binfo->pfme_hdr;

	do {
		port.csr = readq(&fme_hdr->port[i]);
		if (!port.port_implemented)
			break;

		ret = parse_switch_to(binfo, port.port_bar);
		if (ret)
			break;

		ret = parse_feature_list(binfo,
				binfo->ioaddr + port.port_offset);
		if (ret)
			break;
	} while (++i < MAX_FPGA_PORT_NUM);

	return ret;
}

static int create_init_drvdata(struct pci_dev *pdev)
{
	struct cci_drvdata *drvdata;

	drvdata = devm_kzalloc(&pdev->dev, sizeof(*drvdata), GFP_KERNEL);
	if (!drvdata)
		return -ENOMEM;

	drvdata->device_id = alloc_fpga_id(PARENT_ID, NULL);
	if (drvdata->device_id < 0) {
		int ret = drvdata->device_id;

		devm_kfree(&pdev->dev, drvdata);
		return ret;
	}

	mutex_init(&drvdata->lock);
	INIT_LIST_HEAD(&drvdata->port_dev_list);
	INIT_LIST_HEAD(&drvdata->regions);
	drvdata->drv_state = CCI_DRV_STATE_RUNNING;

	dev_set_drvdata(&pdev->dev, drvdata);
	return 0;
}

static void destroy_drvdata(struct pci_dev *pdev)
{
	struct cci_drvdata *drvdata = dev_get_drvdata(&pdev->dev);

	if (drvdata->fme_dev) {
		/* fme device should be unregistered first. */
		WARN_ON(device_is_registered(drvdata->fme_dev));
		free_fpga_id(FME_ID, to_platform_device(drvdata->fme_dev)->id);
		put_device(drvdata->fme_dev);
	}

	cci_pci_remove_port_devs(pdev);
	cci_pci_release_regions(pdev);
	dev_set_drvdata(&pdev->dev, NULL);
	free_fpga_id(PARENT_ID, drvdata->device_id);
	devm_kfree(&pdev->dev, drvdata);
}

static struct class *fpga_class;

struct device *fpga_create_parent_dev(struct pci_dev *pdev, int id)
{
	struct device *dev;

	dev = device_create(fpga_class, &pdev->dev, MKDEV(0, 0), NULL,
			    "intel-fpga-dev.%d", id);
	if (IS_ERR(dev)) {
		dev_err(&pdev->dev, "create parent device failed %ld.\n",
			PTR_ERR(dev));
		return dev;
	}

	/*
	 * it is safe to modify some device fields here as:
	 *   a) the device is not attached to any bus, i.e, no driver
	 *      will match with this device;
	 *   b) it is a single device, i.e, no child will access its
	 *      resource.
	 */
	dev->dma_mask = pdev->dev.dma_mask;
	dev->dma_parms = pdev->dev.dma_parms;
	dev->coherent_dma_mask = pdev->dev.coherent_dma_mask;
	return dev;
}

static int cci_pci_create_feature_devs(struct pci_dev *pdev)
{
	struct cci_drvdata *drvdata = dev_get_drvdata(&pdev->dev);
	struct build_feature_devs_info *binfo;
	int ret;

	binfo = build_info_alloc_and_init(pdev);
	if (!binfo)
		return -ENOMEM;

	binfo->parent_dev = fpga_create_parent_dev(pdev, drvdata->device_id);
	if (IS_ERR(binfo->parent_dev)) {
		ret = PTR_ERR(binfo->parent_dev);
		goto free_binfo_exit;
	}

	ret = parse_start(binfo);
	if (ret)
		goto free_binfo_exit;

	ret = parse_feature_list(binfo, binfo->ioaddr);
	if (ret)
		goto free_binfo_exit;

	ret = parse_ports_from_fme(binfo);
	if (ret)
		goto free_binfo_exit;

	ret = build_info_commit_dev(binfo);
	if (ret)
		goto free_binfo_exit;

	/*
	 * everything is okay, reset ->parent_dev to stop it being
	 * freed by build_info_free()
	 */
	binfo->parent_dev = NULL;

free_binfo_exit:
	build_info_free(binfo);
	return ret;
}

/* PCI Device ID */
#define PCIe_DEVICE_ID_RCiEP0_MCP         0xBCBD
#define PCIe_DEVICE_ID_RCiEP0_SKX_P       0xBCC0
#define PCIe_DEVICE_ID_RCiEP0_DCP         0x09C4
#define PCIe_DEVICE_ID_DCP_0B2B           0x0B2B
#define PCIe_DEVICE_ID_PAC_N3000          0x0B30

/* VF Device */
#define PCIe_DEVICE_ID_VF_MCP             0xBCBF
#define PCIe_DEVICE_ID_VF_SKX_P           0xBCC1
#define PCIe_DEVICE_ID_VF_DCP             0x09C5
#define PCIe_DEVICE_ID_VF_DCP_0B2C        0x0B2C
#define PCIe_DEVICE_ID_VF_PAC_N3000       0x0B31

static struct pci_device_id cci_pcie_id_tbl[] = {
	{PCI_DEVICE(PCI_VENDOR_ID_INTEL, PCIe_DEVICE_ID_RCiEP0_MCP),},
	{PCI_DEVICE(PCI_VENDOR_ID_INTEL, PCIe_DEVICE_ID_VF_MCP),},
	{PCI_DEVICE(PCI_VENDOR_ID_INTEL, PCIe_DEVICE_ID_RCiEP0_SKX_P),},
	{PCI_DEVICE(PCI_VENDOR_ID_INTEL, PCIe_DEVICE_ID_VF_SKX_P),},
	{PCI_DEVICE(PCI_VENDOR_ID_INTEL, PCIe_DEVICE_ID_RCiEP0_DCP),},
	{PCI_DEVICE(PCI_VENDOR_ID_INTEL, PCIe_DEVICE_ID_VF_DCP),},
	{PCI_DEVICE(PCI_VENDOR_ID_INTEL, PCIe_DEVICE_ID_DCP_0B2B),},
	{PCI_DEVICE(PCI_VENDOR_ID_INTEL, PCIe_DEVICE_ID_VF_DCP_0B2C),},
	{PCI_DEVICE(PCI_VENDOR_ID_INTEL, PCIe_DEVICE_ID_PAC_N3000),},
	{PCI_DEVICE(PCI_VENDOR_ID_INTEL, PCIe_DEVICE_ID_VF_PAC_N3000),},
	{0,}
};
MODULE_DEVICE_TABLE(pci, cci_pcie_id_tbl);

static void port_config_vf(struct device *fme_dev, int port_id, bool is_vf)
{
	struct feature_fme_header *fme_hdr;
	struct feature_fme_port port;
	int type = is_vf ? FME_AFU_ACCESS_VF : FME_AFU_ACCESS_PF;

	fme_hdr = get_feature_ioaddr_by_id(fme_dev, FME_FEATURE_ID_HEADER);

	WARN_ON(!fme_hdr);

	port.csr = readq(&fme_hdr->port[port_id]);
	WARN_ON(!port.port_implemented);

	port.afu_access_control = type;
	writeq(port.csr, &fme_hdr->port[port_id]);
}

static int cci_pci_sriov_configure(struct pci_dev *pcidev, int num_vfs)
{
	int ret;
	int vf_ports = 0;
	struct device *fme_dev;
	struct cci_drvdata *drvdata = dev_get_drvdata(&pcidev->dev);
	struct list_head *port_list = &drvdata->port_dev_list;
	struct feature_platform_data *pdata;

	mutex_lock(&drvdata->lock);

	fme_dev = drvdata->fme_dev;
	WARN_ON(!fme_dev);

	if (drvdata->released_port_num < num_vfs) {
		ret = -EBUSY;
		goto unlock_exit;
	}

	if (!num_vfs)
		pci_disable_sriov(pcidev);

	list_for_each_entry(pdata, port_list, node) {
		int id = fpga_port_id(pdata->dev);

		if (device_is_registered(&pdata->dev->dev))
			continue;

		if (!num_vfs) {
			port_config_vf(fme_dev, id, false);
			dev_dbg(&pcidev->dev, "port_%d is turned to PF.\n", id);
			continue;
		}

		port_config_vf(fme_dev, id, true);
		dev_dbg(&pcidev->dev, "port_%d is turned to VF.\n", id);
		if (++vf_ports == num_vfs)
			break;
	}

	if (num_vfs) {
		ret = pci_enable_sriov(pcidev, num_vfs);
		if (ret) {
			list_for_each_entry(pdata, port_list, node) {
				int id = fpga_port_id(pdata->dev);

				if (device_is_registered(&pdata->dev->dev))
					continue;

				port_config_vf(fme_dev, id, false);
			}

			goto unlock_exit;
		}
	}

	ret = num_vfs;
unlock_exit:
	mutex_unlock(&drvdata->lock);
	return ret;
}

static int cci_pci_alloc_irq(struct pci_dev *pcidev)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,8,0) || \
     (defined(RHEL_RELEASE_CODE) && \
        RHEL_RELEASE_CODE >= 0x0705)
#else
	struct cci_drvdata *drvdata = dev_get_drvdata(&pcidev->dev);
	int i = 0;
#endif
	int nvec = pci_msix_vec_count(pcidev);
	int ret = 0;

	if (nvec <= 0) {
		dev_dbg(&pcidev->dev, "fpga interrupt not supported\n");
		return 0;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,8,0) || \
     (defined(RHEL_RELEASE_CODE) && \
        RHEL_RELEASE_CODE >= 0x0705)
	ret = pci_alloc_irq_vectors(pcidev, nvec, nvec, PCI_IRQ_MSIX);
	if (ret < 0)
		return ret;
#else
	drvdata->msix_entries = devm_kmalloc(&pcidev->dev,
					     (sizeof(struct msix_entry) * nvec),
					     GFP_KERNEL);
	if (!drvdata->msix_entries)
		return -ENOMEM;

	for (i = 0; i < nvec; i++) {
		drvdata->msix_entries[i].entry = i;
		drvdata->msix_entries[i].vector = 0;
	}

	ret = pci_enable_msix_exact(pcidev, drvdata->msix_entries, nvec);
	if (ret) {
		devm_kfree(&pcidev->dev, drvdata->msix_entries);
		drvdata->msix_entries = NULL;
		return ret;
	}
#endif

	return 0;
}

static void cci_pci_free_irq(struct pci_dev *pcidev)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,8,0) || \
     (defined(RHEL_RELEASE_CODE) && \
        RHEL_RELEASE_CODE >= 0x0705)
	pci_free_irq_vectors(pcidev);
#else
	struct cci_drvdata *drvdata = dev_get_drvdata(&pcidev->dev);

	if (drvdata->msix_entries) {
		pci_disable_msix(pcidev);
		devm_kfree(&pcidev->dev, drvdata->msix_entries);
		drvdata->msix_entries = NULL;
	}
#endif
}

static void stop_afus(struct cci_drvdata *drvdata)
{
	struct feature_platform_data *pdata;
	struct task_struct *task;
	struct pid_info *pid;
	unsigned int timeout;
	int running;

	list_for_each_entry(pdata, &drvdata->port_dev_list, node) {
		if (!device_is_registered(&pdata->dev->dev))
			continue;

		mutex_lock(&pdata->lock);
		list_for_each_entry(pid, &pdata->pid_list, node) {
			task = pid_task(pid->pid, PIDTYPE_PID);
			if (task == current)
				continue;
			dev_info(&pdata->dev->dev, "SIGSTOP to %d.\n",
				 pid->pid->numbers[0].nr);
			kill_pid(pid->pid, SIGSTOP, 1);
		}
		pdata->unplug = 1;
		mutex_unlock(&pdata->lock);

		timeout = 1000;	/* 1000ms */
		while (timeout) {
			running = 0;
			mutex_lock(&pdata->lock);
			list_for_each_entry(pid, &pdata->pid_list, node) {
				task = pid_task(pid->pid, PIDTYPE_PID);
				if (task == current)
					continue;
				if (!task_is_stopped(task))
					running++;
			}
			mutex_unlock(&pdata->lock);
			if (!running)
				break;
			timeout--;
			mdelay(1);
		};

		if (!timeout)
			dev_err(&pdata->dev->dev,
				"unplug/remove device failed.\n");

		mutex_lock(&pdata->lock);
		__fpga_port_disable(pdata->dev);
		list_for_each_entry(pid, &pdata->pid_list, node) {
			task = pid_task(pid->pid, PIDTYPE_PID);
			if (task == current)
				continue;
			dev_info(&pdata->dev->dev, "SIGHUP & SIGCONT to %d.\n",
				 pid->pid->numbers[0].nr);
			kill_pid(pid->pid, SIGHUP, 1);
			kill_pid(pid->pid, SIGCONT, 1);
		}
		mutex_unlock(&pdata->lock);

		timeout = 1000;	/* 1000ms */
		while (timeout) {
			mutex_lock(&pdata->lock);
			if (pdata->open_count == 0) {
				mutex_unlock(&pdata->lock);
				break;
			}
			mutex_unlock(&pdata->lock);
			timeout--;
			mdelay(1);
		};
		if (!timeout)
			dev_err(&pdata->dev->dev, "kill fpga pids failed %d.\n",
				pdata->open_count);
		else
			dev_info(&pdata->dev->dev, "AFU stopped\n");
	}
}

static void start_afus(struct cci_drvdata *drvdata)
{
	struct feature_platform_data *pdata;
	list_for_each_entry(pdata, &drvdata->port_dev_list, node) {
		if (!device_is_registered(&pdata->dev->dev))
			continue;
		mutex_lock(&pdata->lock);
		__fpga_port_enable(pdata->dev);
		pdata->unplug = 0;
		mutex_unlock(&pdata->lock);
	}
}

static ssize_t
drv_cntrl_store(struct device *dev, struct device_attribute *attr,
		const char *buf, size_t count)
{
	struct cci_drvdata *drvdata = dev_get_drvdata(dev);
	u8 new_state;

	int err;

	err = kstrtou8(buf, 0, &new_state);
	if (err)
		return err;

	mutex_lock(&drvdata->lock);

	switch (drvdata->drv_state) {
	case CCI_DRV_STATE_RUNNING:
		if (new_state == CCI_DRV_STATE_PORTS_STOPPED) {
			drvdata->drv_state = CCI_DRV_STATE_PORTS_STOPPED;
			stop_afus(drvdata);
		} else if (new_state != CCI_DRV_STATE_RUNNING) {
			count = -EINVAL;
			dev_err(dev, "bad state change from %d to %d\n",
				drvdata->drv_state, new_state);
		}
		break;

	case CCI_DRV_STATE_PORTS_STOPPED:
		if (new_state == CCI_DRV_STATE_RUNNING) {
			drvdata->drv_state = CCI_DRV_STATE_RUNNING;
			start_afus(drvdata);
		} else if (new_state != CCI_DRV_STATE_PORTS_STOPPED) {
			count = -EINVAL;
			dev_err(dev, "bad state change from %d to %d\n",
				drvdata->drv_state, new_state);
		}
		break;

	default:
		dev_err(dev, "unexpected state %d\n", drvdata->drv_state);
		count = -EINVAL;
		break;
	}

	mutex_unlock(&drvdata->lock);

	return count;
}

static ssize_t
drv_cntrl_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct cci_drvdata *drvdata = dev_get_drvdata(dev);
	int ret;

	mutex_lock(&drvdata->lock);

	switch (drvdata->drv_state) {
	case CCI_DRV_STATE_RUNNING:
		ret = scnprintf(buf, PAGE_SIZE, "running");
		break;

	case CCI_DRV_STATE_PORTS_STOPPED:
		ret = scnprintf(buf, PAGE_SIZE, "ports_stopped");
		break;

	default:
		ret = scnprintf(buf, PAGE_SIZE, "unknown");
		break;
	}

	mutex_unlock(&drvdata->lock);

	return ret;
}

static DEVICE_ATTR_RW(drv_cntrl);

static const struct attribute *cci_pci_attrs[] = {
	&dev_attr_drv_cntrl.attr,
	NULL,
};

static
int cci_pci_probe(struct pci_dev *pcidev, const struct pci_device_id *pcidevid)
{
	int ret;

	ret = pci_enable_device(pcidev);
	if (ret < 0) {
		dev_err(&pcidev->dev, "Failed to enable device %d.\n", ret);
		goto exit;
	}

	ret = pci_enable_pcie_error_reporting(pcidev);
	if (ret && ret != -EINVAL)
		dev_info(&pcidev->dev, "PCIE AER unavailable %d.\n", ret);

	ret = pci_request_regions(pcidev, DRV_NAME);
	if (ret) {
		dev_err(&pcidev->dev, "Failed to request regions.\n");
		goto disable_error_report_exit;
	}

	pci_set_master(pcidev);
	pci_save_state(pcidev);

	if (!dma_set_mask(&pcidev->dev, DMA_BIT_MASK(64))) {
		dma_set_coherent_mask(&pcidev->dev, DMA_BIT_MASK(64));
	} else if (!dma_set_mask(&pcidev->dev, DMA_BIT_MASK(32))) {
		dma_set_coherent_mask(&pcidev->dev, DMA_BIT_MASK(32));
	} else {
		ret = -EIO;
		dev_err(&pcidev->dev, "No suitable DMA support available.\n");
		goto release_region_exit;
	}

	ret = create_init_drvdata(pcidev);
	if (ret)
		goto release_region_exit;

	ret = cci_pci_alloc_irq(pcidev);
	if (ret)
		goto destroy_drvdata_exit;

	ret = cci_pci_create_feature_devs(pcidev);
	if (ret)
		goto free_irq_exit;

	ret = sysfs_create_files(&pcidev->dev.kobj, cci_pci_attrs);

	if (!ret)
		goto exit;

free_irq_exit:
	cci_pci_free_irq(pcidev);
destroy_drvdata_exit:
	destroy_drvdata(pcidev);
release_region_exit:
	pci_release_regions(pcidev);
disable_error_report_exit:
	pci_disable_pcie_error_reporting(pcidev);
	pci_disable_device(pcidev);
exit:
	return ret;
}

static
void cci_pci_remove(struct pci_dev *pcidev)
{
	struct cci_drvdata *drvdata = dev_get_drvdata(&pcidev->dev);

	mutex_lock(&drvdata->lock);
	stop_afus(drvdata);
	mutex_unlock(&drvdata->lock);

	sysfs_remove_files(&pcidev->dev.kobj, cci_pci_attrs);

	/* disable sriov. */
	if (dev_is_pf(&pcidev->dev))
		cci_pci_sriov_configure(pcidev, 0);

	remove_all_devs(pcidev);

	pci_disable_pcie_error_reporting(pcidev);

	cci_pci_free_irq(pcidev);
	destroy_drvdata(pcidev);
	pci_release_regions(pcidev);
	pci_disable_device(pcidev);
}

static struct pci_driver cci_pci_driver = {
	.name = DRV_NAME,
	.id_table = cci_pcie_id_tbl,
	.probe = cci_pci_probe,
	.remove = cci_pci_remove,
	.sriov_configure = cci_pci_sriov_configure
};

static int __init ccidrv_init(void)
{
	int ret;

	pr_info("Intel(R) FPGA PCIe Driver: Version %s\n", DRV_VERSION);

	fpga_ids_init();

	ret = fpga_chardev_init();
	if (ret)
		goto exit_ids;

	fpga_class = class_create(THIS_MODULE, "fpga");
	if (IS_ERR(fpga_class)) {
		ret = PTR_ERR(fpga_class);
		goto exit_chardev;
	}

	ret = pci_register_driver(&cci_pci_driver);
	if (ret) {
		class_destroy(fpga_class);
exit_chardev:
		fpga_chardev_uinit();
exit_ids:
		fpga_ids_destroy();
	}

	return ret;
}

static void __exit ccidrv_exit(void)
{
	pci_unregister_driver(&cci_pci_driver);
	class_destroy(fpga_class);
	fpga_chardev_uinit();
	fpga_ids_destroy();
}

module_init(ccidrv_init);
module_exit(ccidrv_exit);

MODULE_DESCRIPTION("FPGA PCIe Device Drive");
MODULE_AUTHOR("Intel Corporation");
MODULE_LICENSE("GPL v2");
