/*
 *  Copyright Intel Corporation (C) 2017.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Based on the i2c-axxia.c driver.
 */
#include <linux/clk.h>
#include <linux/clkdev.h>
#include <linux/err.h>
#include <linux/i2c.h>
#include <linux/i2c/altera.h>
#include <linux/iopoll-4-12.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/version.h>

#define ALTR_I2C_TFR_CMD	0x00	/* Transfer Command register */
#define     ALTR_I2C_TFR_CMD_STA	BIT(9)	/* send START before byte */
#define     ALTR_I2C_TFR_CMD_STO	BIT(8)	/* send STOP after byte */
#define     ALTR_I2C_TFR_CMD_RW_D	BIT(0)	/* Direction of transfer */
#define ALTR_I2C_RX_DATA	0x04	/* RX data FIFO register */
#define ALTR_I2C_CTRL		0x08	/* Control register */
#define     ALTR_I2C_CTRL_RXT_SHFT	4	/* RX FIFO Threshold */
#define     ALTR_I2C_CTRL_TCT_SHFT	2	/* TFER CMD FIFO Threshold */
#define     ALTR_I2C_CTRL_BSPEED	BIT(1)	/* Bus Speed (1=Fast) */
#define     ALTR_I2C_CTRL_EN	BIT(0)	/* Enable Core (1=Enable) */
#define ALTR_I2C_ISER		0x0C	/* Interrupt Status Enable register */
#define     ALTR_I2C_ISER_RXOF_EN	BIT(4)	/* Enable RX OVERFLOW IRQ */
#define     ALTR_I2C_ISER_ARB_EN	BIT(3)	/* Enable ARB LOST IRQ */
#define     ALTR_I2C_ISER_NACK_EN	BIT(2)	/* Enable NACK DET IRQ */
#define     ALTR_I2C_ISER_RXRDY_EN	BIT(1)	/* Enable RX Ready IRQ */
#define     ALTR_I2C_ISER_TXRDY_EN	BIT(0)	/* Enable TX Ready IRQ */
#define ALTR_I2C_ISR		0x10	/* Interrupt Status register */
#define     ALTR_I2C_ISR_RXOF		BIT(4)	/* RX OVERFLOW IRQ */
#define     ALTR_I2C_ISR_ARB		BIT(3)	/* ARB LOST IRQ */
#define     ALTR_I2C_ISR_NACK		BIT(2)	/* NACK DET IRQ */
#define     ALTR_I2C_ISR_RXRDY		BIT(1)	/* RX Ready IRQ */
#define     ALTR_I2C_ISR_TXRDY		BIT(0)	/* TX Ready IRQ */
#define ALTR_I2C_STATUS		0x14	/* Status register */
#define     ALTR_I2C_STAT_CORE		BIT(0)	/* Core Status (0=idle) */
#define ALTR_I2C_TC_FIFO_LVL	0x18	/* Transfer FIFO LVL register */
#define ALTR_I2C_RX_FIFO_LVL	0x1C	/* Receive FIFO LVL register */
#define ALTR_I2C_SCL_LOW	0x20	/* SCL low count register */
#define ALTR_I2C_SCL_HIGH	0x24	/* SCL high count register */
#define ALTR_I2C_SDA_HOLD	0x28	/* SDA hold count register */

#define ALTR_I2C_ALL_IRQ	(ALTR_I2C_ISR_RXOF | ALTR_I2C_ISR_ARB | \
				 ALTR_I2C_ISR_NACK | ALTR_I2C_ISR_RXRDY | \
				 ALTR_I2C_ISR_TXRDY)

#define ALTR_I2C_THRESHOLD	0	/* IRQ Threshold at 1 element */
#define ALTR_I2C_DFLT_FIFO_SZ	4
#define ALTR_I2C_TIMEOUT	100000	/* 100ms */
#define ALTR_I2C_XFER_TIMEOUT	(msecs_to_jiffies(250))

/**
 * altr_i2c_dev - I2C device context
 * @regmap: pointer to regmap.
 * @msg: pointer to current message
 * @msg_len: number of bytes transferred in msg
 * @msg_err: error code for completed message
 * @msg_complete: xfer completion object
 * @dev: device reference
 * @adapter: core i2c abstraction
 * @i2c_clk: clock reference for i2c input clock
 * @bus_clk_rate: current i2c bus clock rate
 * @buf: ptr to msg buffer for easier use.
 * @fifo_size: size of the FIFO passed in.
 * @irq: irq number.
 * @isr_mask: cached copy of local ISR enables.
 * @isr_status: cached copy of local ISR status.
 * @lock: spinlock for IRQ synchronization.
 */
struct altr_i2c_dev {
	struct regmap *regmap;
	struct i2c_msg *msg;
	size_t msg_len;
	int msg_err;
	struct completion msg_complete;
	struct device *dev;
	struct i2c_adapter adapter;
	struct clk *i2c_clk;
	u32 bus_clk_rate;
	u8 *buf;
	u32 fifo_size;
	int irq;
	u32 isr_mask;
	u32 isr_status;
	spinlock_t lock;	/* IRQ synchronization */
};

static const struct regmap_config altr_i2c_config = {
	.reg_bits = 32,
	.reg_stride = 4,
	.val_bits = 32,
	.fast_io = true,
};

static void
altr_i2c_int_enable(struct altr_i2c_dev *idev, u32 mask, bool enable)
{
	unsigned long flags;
	u32 int_en;

	spin_lock_irqsave(&idev->lock, flags);

	regmap_read(idev->regmap, ALTR_I2C_ISER, &int_en);
	if (enable)
		idev->isr_mask = int_en | mask;
	else
		idev->isr_mask = int_en & ~mask;

	regmap_write(idev->regmap, ALTR_I2C_ISER, idev->isr_mask);

	spin_unlock_irqrestore(&idev->lock, flags);
}

static void altr_i2c_int_clear(struct altr_i2c_dev *idev, u32 mask)
{
	u32 int_en;

	regmap_read(idev->regmap, ALTR_I2C_ISR, &int_en);
	regmap_write(idev->regmap, ALTR_I2C_ISR, int_en | mask);
}

static void altr_i2c_core_disable(struct altr_i2c_dev *idev)
{
	u32 tmp;

	regmap_read(idev->regmap, ALTR_I2C_CTRL, &tmp);
	regmap_write(idev->regmap, ALTR_I2C_CTRL, tmp & ~ALTR_I2C_CTRL_EN);
}

static void altr_i2c_core_enable(struct altr_i2c_dev *idev)
{
	u32 tmp;

	regmap_read(idev->regmap, ALTR_I2C_CTRL, &tmp);
	regmap_write(idev->regmap, ALTR_I2C_CTRL, tmp | ALTR_I2C_CTRL_EN);
}

static void altr_i2c_reset(struct altr_i2c_dev *idev)
{
	altr_i2c_core_disable(idev);
	altr_i2c_core_enable(idev);
}

static inline void altr_i2c_stop(struct altr_i2c_dev *idev)
{
	regmap_write(idev->regmap, ALTR_I2C_TFR_CMD, ALTR_I2C_TFR_CMD_STO);
}

static void altr_i2c_init(struct altr_i2c_dev *idev)
{
/**
 * on 3.10 kernel, some issue on clk_register_fixed_rate(),
 * so hardcore on 3.10.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,6,0)
	u32 divisor = 100000000/ idev->bus_clk_rate;
	u32 clk_mhz = 100000000 / 1000000;
#else
	u32 divisor = clk_get_rate(idev->i2c_clk) / idev->bus_clk_rate;
	u32 clk_mhz = clk_get_rate(idev->i2c_clk) / 1000000;
#endif
	u32 tmp = (ALTR_I2C_THRESHOLD << ALTR_I2C_CTRL_RXT_SHFT) |
		  (ALTR_I2C_THRESHOLD << ALTR_I2C_CTRL_TCT_SHFT);
	u32 t_high, t_low;

	if (idev->bus_clk_rate <= 100000) {
		tmp &= ~ALTR_I2C_CTRL_BSPEED;
		/* Standard mode SCL 50/50 */
		t_high = divisor * 1 / 2;
		t_low = divisor * 1 / 2;
	} else {
		tmp |= ALTR_I2C_CTRL_BSPEED;
		/* Fast mode SCL 33/66 */
		t_high = divisor * 1 / 3;
		t_low = divisor * 2 / 3;
	}
	regmap_write(idev->regmap, ALTR_I2C_CTRL, tmp);

	dev_dbg(idev->dev, "rate=%uHz per_clk=%uMHz -> ratio=1:%u\n",
		idev->bus_clk_rate, clk_mhz, divisor);

	/* Reset controller */
	altr_i2c_reset(idev);

	/* SCL High Time */
	regmap_write(idev->regmap, ALTR_I2C_SCL_HIGH, t_high);
	/* SCL Low Time */
	regmap_write(idev->regmap, ALTR_I2C_SCL_LOW, t_low);
	/* SDA Hold Time, 300ns */
	regmap_write(idev->regmap, ALTR_I2C_SDA_HOLD,
		     (div_u64(300 * clk_mhz, 1000)));

	/* Mask all master interrupt bits */
	altr_i2c_int_enable(idev, ALTR_I2C_ALL_IRQ, false);
}

/**
 * altr_i2c_transfer - On the last byte to be transmitted, send
 * a Stop bit on the last byte.
 */
static void altr_i2c_transfer(struct altr_i2c_dev *idev, u32 data)
{
	/* On the last byte to be transmitted, send STOP */
	if (idev->msg_len == 1)
		data |= ALTR_I2C_TFR_CMD_STO;
	if (idev->msg_len > 0)
		regmap_write(idev->regmap, ALTR_I2C_TFR_CMD, data);
}

/**
 * altr_i2c_empty_rx_fifo - Fetch data from RX FIFO until end of
 * transfer. Send a Stop bit on the last byte.
 */
static void altr_i2c_empty_rx_fifo(struct altr_i2c_dev *idev)
{
	u32 rx_fifo_avail, data;
	int bytes_to_transfer;

	regmap_read(idev->regmap, ALTR_I2C_RX_FIFO_LVL, &rx_fifo_avail);
	bytes_to_transfer = min_t(size_t, rx_fifo_avail, idev->msg_len);

	while (bytes_to_transfer-- > 0) {
		regmap_read(idev->regmap, ALTR_I2C_RX_DATA, &data);
		*idev->buf++ = data;
		idev->msg_len--;
		altr_i2c_transfer(idev, 0);
	}
}

/**
 * altr_i2c_fill_tx_fifo - Fill TX FIFO from current message buffer.
 * @return: Number of bytes left to transfer.
 */
static int altr_i2c_fill_tx_fifo(struct altr_i2c_dev *idev)
{
	int bytes_to_transfer, ret;
	size_t tx_fifo_avail;
	u32 size;

	regmap_read(idev->regmap, ALTR_I2C_TC_FIFO_LVL, &size);
	tx_fifo_avail = idev->fifo_size - size;
	bytes_to_transfer = min(tx_fifo_avail, idev->msg_len);
	ret = idev->msg_len - bytes_to_transfer;

	while (bytes_to_transfer-- > 0) {
		altr_i2c_transfer(idev, *idev->buf++);
		idev->msg_len--;
	}

	return ret;
}

static irqreturn_t altr_i2c_isr_quick(int irq, void *_dev)
{
	struct altr_i2c_dev *idev = _dev;
	irqreturn_t ret = IRQ_HANDLED;
	unsigned int status;

	/* Read IRQ status but only interested in Enabled IRQs. */
	regmap_read(idev->regmap, ALTR_I2C_ISR, &status);
	idev->isr_status = status & idev->isr_mask;
	if (idev->isr_status)
		ret = IRQ_WAKE_THREAD;

	return ret;
}

/* could be used in atomic context */
#define altr_i2c_read_poll_timeout(map, addr, val, cond, delay_us, timeout_us) \
({ \
	ktime_t timeout = ktime_add_us(ktime_get(), timeout_us); \
	int pollret; \
	for (;;) { \
		pollret = regmap_read((map), (addr), &(val)); \
		if (pollret) \
			break; \
		if (cond) \
			break; \
		if (timeout_us && ktime_compare(ktime_get(), timeout) > 0) { \
			pollret = regmap_read((map), (addr), &(val)); \
			break; \
		} \
		if (delay_us) \
			udelay(delay_us); \
	} \
	pollret ?: ((cond) ? 0 : -ETIMEDOUT); \
})

/**
 * altr_handle_i2c_events - handle all the pending events in isr_status
 * @idev: the altera i2c device.
 * @return: true if all events are handled and no more event is expected.
 * false if further events are expected so we need to keep on monitoring.
 */
static bool altr_handle_all_i2c_events(struct altr_i2c_dev *idev)
{
	u32 status = idev->isr_status;
	bool read, finish = false;
	int ret;

	read = (idev->msg->flags & I2C_M_RD) != 0;

	/* handle Lost Arbitration */
	if (unlikely(status & ALTR_I2C_ISR_ARB)) {
		altr_i2c_int_clear(idev, ALTR_I2C_ISR_ARB);
		idev->msg_err = -EAGAIN;
		finish = true;
	} else if (unlikely(status & ALTR_I2C_ISR_NACK)) {
		dev_dbg(idev->dev, "Could not get ACK\n");
		idev->msg_err = -ENXIO;
		altr_i2c_int_clear(idev, ALTR_I2C_ISR_NACK);
		altr_i2c_stop(idev);
		finish = true;
	} else if (read && unlikely(status & ALTR_I2C_ISR_RXOF)) {
		/* handle RX FIFO Overflow */
		altr_i2c_empty_rx_fifo(idev);
		altr_i2c_int_clear(idev, ALTR_I2C_ISR_RXRDY);
		altr_i2c_stop(idev);
		dev_err(idev->dev, "RX FIFO Overflow\n");
		finish = true;
	} else if (read && (status & ALTR_I2C_ISR_RXRDY)) {
		/* RX FIFO needs service? */
		altr_i2c_empty_rx_fifo(idev);
		altr_i2c_int_clear(idev, ALTR_I2C_ISR_RXRDY);
		if (!idev->msg_len)
			finish = true;
	} else if (!read && (status & ALTR_I2C_ISR_TXRDY)) {
		/* TX FIFO needs service? */
		altr_i2c_int_clear(idev, ALTR_I2C_ISR_TXRDY);
		if (idev->msg_len > 0)
			altr_i2c_fill_tx_fifo(idev);
		else
			finish = true;
	} else {
		dev_warn(idev->dev, "Unexpected events: 0x%x\n", status);
		altr_i2c_int_clear(idev, ALTR_I2C_ALL_IRQ);
	}

	if (finish) {
		/* Wait for the Core to finish */
		ret = altr_i2c_read_poll_timeout(idev->regmap, ALTR_I2C_STATUS,
						 status,
						 !(status & ALTR_I2C_STAT_CORE),
						 1, ALTR_I2C_TIMEOUT);
		if (ret)
			dev_err(idev->dev, "message timeout\n");
		altr_i2c_int_enable(idev, ALTR_I2C_ALL_IRQ, false);
		altr_i2c_int_clear(idev, ALTR_I2C_ALL_IRQ);
		dev_dbg(idev->dev, "Message Complete\n");
	}

	return finish;
}

static irqreturn_t altr_i2c_isr(int irq, void *_dev)
{
	struct altr_i2c_dev *idev = _dev;

	if (!idev->msg) {
		dev_warn(idev->dev, "unexpected interrupt\n");
		altr_i2c_int_clear(idev, ALTR_I2C_ALL_IRQ);
		return IRQ_HANDLED;
	}

	if (altr_handle_all_i2c_events(idev))
		complete(&idev->msg_complete);

	return IRQ_HANDLED;
}

/**
 * altr_i2c_poll_for_completion - handle all the pending events in isr_status
 * @idev: the altera i2c device.
 * @timeout: timeout value from caller in jiffies.
 * @return: 0 on timeout, otherwise returns remaining time in jiffies.
 */
static unsigned long altr_i2c_poll_for_completion(struct altr_i2c_dev *idev,
						  unsigned long timeout)
{
	unsigned long poll_timeout = jiffies + timeout;
	unsigned int status;
	long remain_time;
	bool finish;

	do {
		for (;;) {
			/*
			 * Read IRQ status but only interested in Enabled
			 * IRQs.
			 */
			regmap_read(idev->regmap, ALTR_I2C_ISR, &status);

			idev->isr_status = status & idev->isr_mask;
			if (idev->isr_status)
				break;

			if (time_after(jiffies, poll_timeout))
				return 0;

			usleep_range(5, 10);
		}

		finish = altr_handle_all_i2c_events(idev);

		if (time_after(jiffies, poll_timeout))
			return 0;

	} while (!finish);

	remain_time = poll_timeout - jiffies;

	return remain_time < 0 ? 0 : remain_time;
}

static unsigned long altr_i2c_wait_for_completion(struct altr_i2c_dev *idev,
						  unsigned long timeout)
{
	if (idev->irq < 0)
		return altr_i2c_poll_for_completion(idev, timeout);

	reinit_completion(&idev->msg_complete);
	return wait_for_completion_timeout(&idev->msg_complete, timeout);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 7, 0) || \
     (defined(RHEL_RELEASE_CODE) && \
        RHEL_RELEASE_CODE >= 0x0705)
#else
static inline u8 i2c_8bit_addr_from_msg(const struct i2c_msg *msg)
{
	return (msg->addr << 1) | (msg->flags & I2C_M_RD ? 1 : 0);
}
#endif

static int altr_i2c_xfer_msg(struct altr_i2c_dev *idev, struct i2c_msg *msg)
{
	u32 imask = ALTR_I2C_ISR_RXOF | ALTR_I2C_ISR_ARB | ALTR_I2C_ISR_NACK;
	unsigned long time_left;
	u32 value;
	u8 addr = i2c_8bit_addr_from_msg(msg);

	idev->msg = msg;
	idev->msg_len = msg->len;
	idev->buf = msg->buf;
	idev->msg_err = 0;
	altr_i2c_core_enable(idev);

	/* Make sure RX FIFO is empty */
	do {
		regmap_read(idev->regmap, ALTR_I2C_RX_DATA, &value);
		regmap_read(idev->regmap, ALTR_I2C_RX_FIFO_LVL, &value);
	} while (value);

	regmap_write(idev->regmap, ALTR_I2C_TFR_CMD,
		     ALTR_I2C_TFR_CMD_STA | addr);

	if ((msg->flags & I2C_M_RD) != 0) {
		imask |= ALTR_I2C_ISER_RXOF_EN | ALTR_I2C_ISER_RXRDY_EN;
		altr_i2c_int_enable(idev, imask, true);
		/* write the first byte to start the RX */
		altr_i2c_transfer(idev, 0);
	} else {
		imask |= ALTR_I2C_ISR_TXRDY;
		altr_i2c_int_enable(idev, imask, true);
		altr_i2c_fill_tx_fifo(idev);
	}

	time_left = altr_i2c_wait_for_completion(idev, ALTR_I2C_XFER_TIMEOUT);

	altr_i2c_int_enable(idev, imask, false);

	regmap_read(idev->regmap, ALTR_I2C_STATUS, &value);
	if (value & ALTR_I2C_STAT_CORE)
		dev_err(idev->dev, "Core Status not IDLE...\n");

	if (time_left == 0) {
		idev->msg_err = -ETIMEDOUT;
		dev_dbg(idev->dev, "Transaction timed out.\n");
	}

	altr_i2c_core_disable(idev);

	return idev->msg_err;
}

static int
altr_i2c_xfer(struct i2c_adapter *adap, struct i2c_msg *msgs, int num)
{
	struct altr_i2c_dev *idev = i2c_get_adapdata(adap);
	int i, ret = 0;

	for (i = 0; i < num; i++) {
		ret = altr_i2c_xfer_msg(idev, msgs++);
		if (ret)
			return ret;
	}
	return num;
}

static u32 altr_i2c_func(struct i2c_adapter *adap)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}

static const struct i2c_algorithm altr_i2c_algo = {
	.master_xfer = altr_i2c_xfer,
	.functionality = altr_i2c_func,
};

static int altr_i2c_probe(struct platform_device *pdev)
{
	struct altr_i2c_platform_data *pdata = dev_get_platdata(&pdev->dev);
	struct altr_i2c_dev *idev = NULL;
	struct resource *res;
	void __iomem *base;
	int ret;
	u32 val;
	u8 i;

	idev = devm_kzalloc(&pdev->dev, sizeof(*idev), GFP_KERNEL);
	if (!idev)
		return -ENOMEM;

	if (pdata && pdata->regmap_ctx) {
		idev->regmap = devm_regmap_init(&pdev->dev, &pdata->regbus,
						pdata->regmap_ctx,
						&altr_i2c_config);
	} else {
		res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
		base = devm_ioremap_resource(&pdev->dev, res);
		if (IS_ERR(base))
			return PTR_ERR(base);

		idev->regmap = devm_regmap_init_mmio(&pdev->dev, base,
						     &altr_i2c_config);
	}

	if (IS_ERR(idev->regmap))
		return PTR_ERR(idev->regmap);

	idev->dev = &pdev->dev;
	init_completion(&idev->msg_complete);
	spin_lock_init(&idev->lock);

	if (pdata) {
		idev->i2c_clk = pdata->i2c_clk;
		idev->fifo_size = pdata->fifo_size;
		idev->bus_clk_rate = pdata->bus_clk_rate;
	} else {
		idev->i2c_clk = devm_clk_get(&pdev->dev, NULL);
		if (IS_ERR(idev->i2c_clk)) {
			dev_err(&pdev->dev, "missing clock\n");
			return PTR_ERR(idev->i2c_clk);
		}

		val = device_property_read_u32(idev->dev, "fifo-size",
					       &idev->fifo_size);
		if (val) {
			dev_err(&pdev->dev, "FIFO size set to default of %d\n",
				ALTR_I2C_DFLT_FIFO_SZ);
			idev->fifo_size = ALTR_I2C_DFLT_FIFO_SZ;
		}

		val = device_property_read_u32(idev->dev, "clock-frequency",
					       &idev->bus_clk_rate);
		if (val) {
			dev_err(&pdev->dev, "Default to 100kHz\n");
			idev->bus_clk_rate = 100000;	/* default clock rate */
		}
	}

	if (idev->bus_clk_rate > 400000) {
		dev_err(&pdev->dev, "invalid clock-frequency %d\n",
			idev->bus_clk_rate);
		return -EINVAL;
	}

	idev->irq = platform_get_irq(pdev, 0);
	if (idev->irq >= 0) {
		ret = devm_request_threaded_irq(&pdev->dev, idev->irq,
						altr_i2c_isr_quick,
						altr_i2c_isr,
						IRQF_ONESHOT, pdev->name,
						idev);
		if (ret) {
			dev_err(&pdev->dev, "failed to claim IRQ %d\n",
				idev->irq);
			return ret;
		}
	}

	ret = clk_prepare_enable(idev->i2c_clk);
	if (ret) {
		dev_err(&pdev->dev, "failed to enable clock\n");
		return ret;
	}

	altr_i2c_init(idev);

	i2c_set_adapdata(&idev->adapter, idev);
	strlcpy(idev->adapter.name, pdev->name, sizeof(idev->adapter.name));
	idev->adapter.owner = THIS_MODULE;
	idev->adapter.algo = &altr_i2c_algo;
	idev->adapter.dev.parent = &pdev->dev;
	idev->adapter.dev.of_node = pdev->dev.of_node;

	platform_set_drvdata(pdev, idev);

	ret = i2c_add_adapter(&idev->adapter);
	if (ret) {
		clk_disable_unprepare(idev->i2c_clk);
		return ret;
	}

	if (pdata) {
		/* add in known devices to the bus */
		for (i = 0; i < pdata->num_devices; i++)
			i2c_new_device(&idev->adapter, pdata->devices + i);
	}

	dev_info(&pdev->dev, "Altera SoftIP I2C Probe Complete\n");

	return 0;
}

static int altr_i2c_remove(struct platform_device *pdev)
{
	struct altr_i2c_dev *idev = platform_get_drvdata(pdev);

	clk_disable_unprepare(idev->i2c_clk);
	i2c_del_adapter(&idev->adapter);

	return 0;
}

/* Match table for of_platform binding */
static const struct of_device_id altr_i2c_of_match[] = {
	{ .compatible = "altr,softip-i2c-v1.0" },
	{},
};
MODULE_DEVICE_TABLE(of, altr_i2c_of_match);

static struct platform_driver altr_i2c_driver = {
	.probe = altr_i2c_probe,
	.remove = altr_i2c_remove,
	.driver = {
		.name = ALTERA_I2C_DRV_NAME,
		.of_match_table = altr_i2c_of_match,
	},
};

module_platform_driver(altr_i2c_driver);

MODULE_DESCRIPTION("Altera Soft IP I2C bus driver");
MODULE_AUTHOR("Thor Thayer <thor.thayer@linux.intel.com>");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS("platform:" ALTERA_I2C_DRV_NAME);
