#include <linux/seq_file.h>
#include <linux/debugfs.h>
#include <linux/mutex.h>
#include <linux/regmap.h>
#include <linux/slab.h>
#include <linux/device.h>

#include "intel-max10-debug.h"

#ifdef CONFIG_DEBUG_FS

static ssize_t user_buffer_to_uint(const char __user *buffer, size_t count,
				   loff_t *ppos, unsigned int *value)
{
	char buf[24];
	ssize_t len;
	int ret;

	if (*ppos != 0)
		return -EINVAL;

	if (count >= sizeof(buf))
		return -ENOSPC;

	len = simple_write_to_buffer(buf, sizeof(buf) - 1,
				     ppos, buffer, count);
	if (len < 0)
		return len;

	buf[len] = '\0';
	ret = kstrtouint(buf, 16, value);
	if (ret)
		return -EIO;

	return len;
}

static ssize_t uint_to_user_buffer(char __user *buffer, size_t count,
				   loff_t *ppos, unsigned int value)
{
	char buf[24];
	int ret;

	if (*ppos != 0)
		return 0;

	ret = snprintf(buf, sizeof(buf) - 1, "0x%x\n", value);

	return simple_read_from_buffer(buffer, count, ppos, buf, ret);
}

static ssize_t regaddr_write(struct file *file, const char __user *buffer,
			     size_t count, loff_t *ppos)
{
	struct intel_max10_debug *debug = file->private_data;
	int ret;

	mutex_lock(&debug->lock);
	ret = user_buffer_to_uint(buffer, count, ppos, &debug->regaddr);
	mutex_unlock(&debug->lock);

	return ret;
}

static ssize_t regaddr_read(struct file *file, char __user *buffer,
			    size_t count, loff_t *ppos)
{
	struct intel_max10_debug *debug = file->private_data;
	int ret;

	mutex_lock(&debug->lock);
	ret =  uint_to_user_buffer(buffer, count, ppos, debug->regaddr);
	mutex_unlock(&debug->lock);

	return ret;
}

static const struct file_operations regaddr_fops = {
	.owner = THIS_MODULE,
	.open = simple_open,
	.read = regaddr_read,
	.write = regaddr_write,
};

static ssize_t regval_write(struct file *file, const char __user *buffer,
			    size_t count, loff_t *ppos)
{
	struct intel_max10_debug *debug = file->private_data;
	u32 val;
	int ret;

	ret = user_buffer_to_uint(buffer, count, ppos, &val);
	if (ret < 0)
		return ret;

	mutex_lock(&debug->lock);
	ret = regmap_write(debug->map, debug->regaddr, val);
	mutex_unlock(&debug->lock);

	return ret ?: count;
}

static ssize_t regval_read(struct file *file, char __user *buffer,
			   size_t count, loff_t *ppos)
{
	struct intel_max10_debug *debug = file->private_data;
	u32 val;
	int ret;

	mutex_lock(&debug->lock);
	ret = regmap_read(debug->map, debug->regaddr, &val);
	mutex_unlock(&debug->lock);
	if (ret)
		return ret;

	return uint_to_user_buffer(buffer, count, ppos, val);
}

static const struct file_operations regval_fops = {
	.owner = THIS_MODULE,
	.open = simple_open,
	.read = regval_read,
	.write = regval_write,
};

struct intel_max10_debug *intel_max10_debug_init(struct device *dev,
						 struct regmap *map)
{
	struct intel_max10_debug *debug;

	if (!dev || !map)
		return NULL;

	debug = devm_kzalloc(dev, sizeof(*debug), GFP_KERNEL);
	if (!debug)
		return NULL;

	debug->debugfs = debugfs_create_dir(dev_name(dev), NULL);
	if (!debug->debugfs) {
		dev_err(dev, "Failed to create max10 debugfs\n");
		return NULL;
	}

	mutex_init(&debug->lock);
	debug->map = map;

	if (!debugfs_create_file("regaddr", 0644, debug->debugfs,
				 debug, &regaddr_fops))
		goto err_out;

	if (!debugfs_create_file("regval", 0644, debug->debugfs,
				 debug, &regval_fops))
		goto err_out;

	return debug;

err_out:
	debugfs_remove_recursive(debug->debugfs);
	return NULL;
}

void intel_max10_debug_exit(struct intel_max10_debug *debug)
{
	if (!debug)
		return;

	debugfs_remove_recursive(debug->debugfs);
	mutex_destroy(&debug->lock);
}

#else /* !CONFIG_DEBUG_FS */
struct intel_max10_debug *intel_max10_debug_init(struct device *dev,
						 struct regmap *map)
{
	return NULL;
}

void intel_max10_debug_exit(struct intel_max10_debug *debug)
{
}
#endif /* CONFIG_DEBUG_FS */
