// SPDX-License-Identifier: GPL-2.0
/*
 * Intel MAX10 MFD Driver
 *
 * Copyright (C) 2018 Intel Corporation. All rights reserved.
 *
 */

#include <linux/bitfield-intel-pac.h>
#include <linux/c827_retimer.h>
#include <linux/mutex.h>
#include <linux/regmap-intel-pac.h>
#include <linux/module.h>
#include <linux/mfd/core.h>
#include <linux/init.h>
#include <linux/spi/spi.h>
#include <linux/i2c/altera.h>
#include <linux/mtd/intel-generic-qspi.h>
#include <linux/clk-provider.h>
#include <linux/mtd/intel-on-chip-flash.h>
#include <linux/ifpga-sec-mgr.h>
#include <linux/fpga/ifpga-sec-mgr.h>
#include <linux/version.h>

#include "fdt_reg.h"
#include "intel-spi-avmm.h"

#include "intel-max10-debug.h"

struct max10_id {
	char compatible[128];
};

#define MAX10_PAC	"intel,max10"
#define MAX10_PAC_N3000	"intel,max10-pac-n3000"

static struct max10_id max10_id_table[] = {
	{.compatible = MAX10_PAC,},
	{.compatible = MAX10_PAC_N3000,},
	{}
};

struct staging_area {
	u32	base;
	u32	size;
};

/**
 * struct intel_max10 - Intel Max10 MFD device private data structure
 * @dev: this device
 * @bridge: spi to avmm bridge used for register access.
 * @regmap: the regmap used to access registers by max10 itself
 * @flags: used to indicate max10 hardware capability.
 * @sobj_list: list of sensor objects.
 * @id: used to distinguish specific max10 implementation.
 */
struct intel_max10 {
	struct device *dev;
	struct spi_avmm_bridge *bridge;
	struct regmap *regmap;
	unsigned int base;
	unsigned int flags;

	struct platform_device *fpga_flash;
	bool fpga_flash_enabled;
	struct mutex fpga_flash_lock;

	struct platform_device *bmcfw_flash;
	bool bmcfw_flash_enabled;
	struct rw_semaphore bmcfw_lock;

	struct platform_device *i2c_master2;
	bool i2c_master2_enabled;
	struct mutex i2c_master2_lock;

	struct platform_device *bmcimg_flash;
	bool bmcimg_flash_enabled;
	struct mutex bmcimg_flash_lock;

	struct list_head sobj_list;

	struct max10_id *id;

	struct intel_max10_debug *debug;

	struct platform_device *pv0_pdev;
	struct platform_device *pv1_pdev;

	struct ifpga_sec_mgr *imgr;
	struct staging_area staging_area;

	struct platform_device *tmp_igsf;

	bool fdt_exist;
};

/* max10 capability flags */
#define MAX10_FLAGS_NO_I2C2		BIT(0)
#define MAX10_FLAGS_MAC_CACHE		BIT(1)
#define MAX10_FLAGS_SECURE		BIT(2)
#define MAX10_FLAGS_TMP_IGSF		BIT(3)

#define REGMAP_STRIDE			4
static struct regmap_config intel_max10_regmap_config = {
	.reg_bits = 32,
	.val_bits = 32,
	.reg_stride = REGMAP_STRIDE,
	.cache_type = REGCACHE_NONE,
	.max_register = 0x200000fc,
};

static inline bool is_max10_pac_n3000(struct intel_max10 *max10)
{
	return max10->id && !strcmp(max10->id->compatible, MAX10_PAC_N3000);
}

/* On-Chip Flash Dual Configuration IP */
#define DUAL_CFG_CTRL0		0x200020
#define DUAL_CFG_CTRL1		0x200024

/* System Registers */
#define MAX10_BASE_ADDR		0x300400
#define MAX10_SEC_BASE_ADDR	0x300800
/* Register offset of system registers */
#define NIOS2_FW_VERSION	0x0
#define MAX10_MACADDR1		0x10
#define   MAX10_MAC_BYTE4	GENMASK(7, 0)
#define   MAX10_MAC_BYTE3	GENMASK(15, 8)
#define   MAX10_MAC_BYTE2	GENMASK(23, 16)
#define   MAX10_MAC_BYTE1	GENMASK(31, 24)
#define MAX10_MACADDR2		0x14
#define   MAX10_MAC_BYTE6	GENMASK(7, 0)
#define   MAX10_MAC_BYTE5	GENMASK(15, 8)
#define   MAX10_MAC_COUNT	GENMASK(23, 16)
#define RSU_REG			0x2c
#define   FPGA_RECONF_PAGE	GENMASK(2, 0)
#define   FPGA_RP_LOAD		BIT(3)
#define   NIOS2_PRERESET	BIT(4)
#define   NIOS2_HANG		BIT(5)
#define   RSU_ENABLE		BIT(6)
#define   NIOS2_RESET		BIT(7)
#define   NIOS2_I2C2_POLL_STOP	BIT(13)
#define   PKVL_EEPROM_LOAD	BIT(31)
#define FPGA_RECONF_REG		0x30
#define   COUNTDOWN_START	BIT(18)
#define   SFPGA_RECONF_PAGE	GENMASK(22, 20)
#define   SFPGA_RP_LOAD		BIT(23)
#define MAX10_TEST_REG		0x3c
#define MAX10_BUILD_VER		0x68
#define   MAX10_VERSION_MAJOR	GENMASK(23, 16)
#define   PCB_INFO		GENMASK(31, 24)
#define FPGA_PAGE_INFO		0x6c
#define DT_AVAIL_REG		0x90
#define   DT_AVAIL		BIT(0)
#define DT_BASE_ADDR_REG	0x94
#define MAX10_DOORBELL		0x400
#define   RSU_REQUEST		BIT(0)
#define   SEC_PROGRESS		GENMASK(7, 4)
#define   HOST_STATUS		GENMASK(11, 8)
#define   SEC_STATUS		GENMASK(23, 16)
#define   PKVL_EEPROM_LOAD_SEC	BIT(24)
#define   PKVL1_POLL_EN		BIT(25)
#define   PKVL2_POLL_EN		BIT(26)
#define   CONFIG_SEL		BIT(28)
#define   REBOOT_REQ		BIT(29)
#define MAX10_AUTH_RESULT	0x404


/* Progress states */
#define RSU_PROG_IDLE			0x0
#define RSU_PROG_PREPARE		0x1
#define RSU_PROG_READY			0x3
#define RSU_PROG_AUTHENTICATING		0x4
#define RSU_PROG_COPYING		0x5
#define RSU_PROG_UPDATE_CANCEL		0x6
#define RSU_PROG_PROGRAM_KEY_HASH	0x7
#define RSU_PROG_RSU_DONE		0x8
#define RSU_PROG_PKVL_PROM_DONE		0x9

/* Device and error states */
#define RSU_STAT_NORMAL			0x0
#define RSU_STAT_TIMEOUT		0x1
#define RSU_STAT_AUTH_FAIL		0x2
#define RSU_STAT_COPY_FAIL		0x3
#define RSU_STAT_FATAL			0x4
#define RSU_STAT_PKVL_REJECT		0x5
#define RSU_STAT_NON_INC		0x6
#define RSU_STAT_ERASE_FAIL		0x7
#define RSU_STAT_WEAROUT		0x8
#define RSU_STAT_NIOS_OK		0x80
#define RSU_STAT_USER_OK		0x81
#define RSU_STAT_FACTORY_OK		0x82
#define RSU_STAT_USER_FAIL		0x83
#define RSU_STAT_FACTORY_FAIL		0x84
#define RSU_STAT_NIOS_FLASH_ERR		0x85
#define RSU_STAT_FPGA_FLASH_ERR		0x86

#define HOST_STATUS_IDLE		0x0
#define HOST_STATUS_WRITE_DONE		0x1
#define HOST_STATUS_ABORT_RSU		0x2

#define sec_prog(doorbell)		FIELD_GET(SEC_PROGRESS, doorbell)
#define sec_stat(doorbell)		FIELD_GET(SEC_STATUS, doorbell)

#define NIOS2_HANG_POLL_INTERVAL_US	1000
#define NIOS2_HANG_POLL_TIMEOUT_US	(5000 * NIOS2_HANG_POLL_INTERVAL_US)

#define NIOS2_I2C2_POLL_STOP_DELAY_MS	2000	/* > 1 sec poll interval */

#define DFT_MAX_SIZE			0x7e0000

/* interval 100ms and timeout 20s */
#define SEC_START_INTERVAL_US		(100 * 1000)
#define SEC_START_TIMEOUT_US		(20 * 1000 * 1000)

/* interval 100ms and timeout 5s */
#define NIOS_HANDSHAKE_INTERVAL_US	(100 * 1000)
#define NIOS_HANDSHAKE_TIMEOUT_US	(5 * 1000 * 1000)

/*
 * register access helper functions.
 *
 * max10_raw_write - write max10 register per addr
 * max10_raw_bulk_read - bulk_read max10 registers per addr
 * max10_raw_bulk_write - bulk_write max10 registers per addr
 * max10_sys_read - read max10 system register per offset
 * max10_sys_update_bits - update max10 system register per offset
 * max10_handshake_read - read max10 handshake system register per offset
 */
static int max10_raw_write(struct intel_max10 *max10, unsigned int addr,
			   unsigned int val)
{
	int ret;

	ret = regmap_write(max10->regmap, addr, val);
	if (ret)
		dev_err(max10->dev, "fail to write raw reg %x val %x: %d\n",
			addr, val, ret);

	return ret;
}

static int max10_raw_bulk_read(struct intel_max10 *max10, unsigned int addr,
			       void *val, size_t cnt)
{
	int ret;

	ret = regmap_bulk_read(max10->regmap, addr, val, cnt);
	if (ret)
		dev_err(max10->dev, "fail to read raw reg %x cnt %zx: %d\n",
			addr, cnt, ret);

	return ret;
}

static int max10_raw_bulk_write(struct intel_max10 *max10, unsigned int addr,
				void *val, size_t cnt)
{
	int ret;

	ret = regmap_bulk_write(max10->regmap, addr, val, cnt);
	if (ret)
		dev_err(max10->dev, "fail to write raw reg %x cnt %zx: %d\n",
			addr, cnt, ret);

	return ret;
}

static int max10_sys_read(struct intel_max10 *max10, unsigned int offset,
			  unsigned int *val)
{
	struct device *dev = max10->dev;
	int ret;

	ret = regmap_read(max10->regmap, max10->base + offset, val);
	if (ret)
		dev_err(dev, "fail to read system reg %x: %d\n", offset, ret);

	return ret;
}

static int max10_sys_update_bits(struct intel_max10 *max10, unsigned int offset,
				 unsigned int msk, unsigned int val)
{
	struct device *dev = max10->dev;
	int ret;

	ret = regmap_update_bits(max10->regmap, max10->base + offset, msk, val);
	if (ret)
		dev_err(dev, "fail to update system reg %x: %d\n", offset, ret);

	return ret;
}

static int max10_spi_setup(struct spi_device *spi, u8 bits_per_word)
{
	spi->mode = SPI_MODE_1;
	spi->bits_per_word = bits_per_word;

	return spi_setup(spi);
}

static int check_rot_handshake(struct intel_max10 *max10)
{
	u32 doorbell;
	int ret;

	ret = max10_sys_read(max10, MAX10_DOORBELL, &doorbell);
	if (ret)
		return ret;

	switch (sec_prog(doorbell)) {
	case RSU_PROG_PREPARE:
	case RSU_PROG_AUTHENTICATING:
	case RSU_PROG_COPYING:
	case RSU_PROG_UPDATE_CANCEL:
	case RSU_PROG_PROGRAM_KEY_HASH:
		return -EBUSY;
	default:
		if (sec_stat(doorbell) == RSU_STAT_USER_OK)
			return -EBUSY;
	}
	return 0;
}

static int enable_rot_handshake(struct intel_max10 *max10)
{
	int ret = 0;

	down_write(&max10->bmcfw_lock);
	if (max10->bmcfw_flash_enabled) {
		ret = check_rot_handshake(max10);
		if (ret) {
			up_write(&max10->bmcfw_lock);
			return ret;
		}
		max10->bmcfw_flash_enabled = false;
	}
	up_write(&max10->bmcfw_lock);

	return 0;
}

static int bmcfw_handshake_enter(struct intel_max10 *max10)
{
	int ret;

	if (max10->flags & MAX10_FLAGS_SECURE && max10->bmcfw_flash_enabled) {
		ret = enable_rot_handshake(max10);
		if (ret)
			return ret;
	}

	if (!down_read_trylock(&max10->bmcfw_lock))
		return -EBUSY;

	if (max10->bmcfw_flash_enabled) {
		up_read(&max10->bmcfw_lock);
		return -EBUSY;
	}

	return 0;
}

static void bmcfw_handshake_exit(struct intel_max10 *max10)
{
	up_read(&max10->bmcfw_lock);
}

static int max10_handshake_read(struct intel_max10 *max10, unsigned int offset,
				unsigned int *val)
{
	int ret;

	ret = bmcfw_handshake_enter(max10);
	if (ret)
		return ret;

	ret = max10_sys_read(max10, offset, val);

	bmcfw_handshake_exit(max10);

	return ret;
}

/* switch on - flash chip accessible, off - forbid to access flash chip */
static int max10_flash_switch_set(struct intel_max10 *max10, bool on)
{
	return max10_sys_update_bits(max10, RSU_REG, RSU_ENABLE,
				     on ? RSU_ENABLE : 0);
}

static const char *devm_fdt_strdup(struct device *dev,
				   const void *fdt, int nodeoffset,
				   const char *name, int *lenp)
{
	const char *str;

	str = fdt_getprop(fdt, nodeoffset, name, lenp);
	if (!str)
		return NULL;

	return devm_kmemdup(dev, str, strlen(str) + 1, GFP_KERNEL);
}

static struct max10_id *max10_match_compatible(const char *fdt_root)
{
	struct max10_id *id = max10_id_table;

	for (; id->compatible[0]; id++) {
		if (fdt_node_check_compatible(fdt_root, 0, id->compatible))
			continue;

		return id;
	}

	return NULL;
}

/* fetch sub device table from hw */
static char *fetch_subdev_table(struct intel_max10 *max10)
{
	int stride = REGMAP_STRIDE;
	struct max10_id *id;
	struct fdt_header hdr;
	char *fdt_root = NULL;
	u32 dt_size, dt_addr, val;
	int ret;

	if (!IS_ALIGNED(sizeof(hdr), stride))
		return NULL;

	if (max10_sys_read(max10, DT_AVAIL_REG, &val))
		return NULL;

	if (!(val & DT_AVAIL)) {
		dev_err(max10->dev, "DT not available\n");
		return NULL;
	}

	if (max10_sys_read(max10, DT_BASE_ADDR_REG, &dt_addr))
		return NULL;

	max10_flash_switch_set(max10, true);

	ret = max10_raw_bulk_read(max10, dt_addr, &hdr, sizeof(hdr) / stride);
	if (ret)
		goto done;

	ret = fdt_check_header(&hdr);
	if (ret) {
		dev_err(max10->dev, "fdt check header failed %d\n", ret);
		goto done;
	}

	dt_size = fdt_totalsize(&hdr);
	if (dt_size > DFT_MAX_SIZE || !IS_ALIGNED(dt_size, stride)) {
		dev_err(max10->dev, "bad dt size (%u)\n", dt_size);
		ret = -EINVAL;
		goto done;
	}

	fdt_root = kmalloc(dt_size, GFP_KERNEL);
	if (!fdt_root) {
		ret = -ENOMEM;
		goto done;
	}

	ret = max10_raw_bulk_read(max10, dt_addr, fdt_root, dt_size / stride);
	if (ret)
		goto done;

	id = max10_match_compatible(fdt_root);
	if (!id) {
		ret = -ENODEV;
		dev_err(max10->dev, "max10 compatible not found\n");
		goto done;
	}

	max10->id = id;

done:
	max10_flash_switch_set(max10, false);

	if (ret && fdt_root) {
		kfree(fdt_root);
		return NULL;
	}

	return fdt_root;
}

#define SENSOR_REG_VALUE	0x0
#define SENSOR_REG_HIGH_WARN	0x1
#define SENSOR_REG_HIGH_FATAL	0x2
#define SENSOR_REG_LOW_WARN	0x3
#define SENSOR_REG_LOW_FATAL	0x4
#define SENSOR_REG_HYSTERESIS	0x5
#define SENSOR_REG_MAX		0x6

static const char * const sensor_reg_name[] = {
	"value",
	"high_warn",
	"high_fatal",
	"low_warn",
	"low_fatal",
	"hysteresis",
};

struct sensor_reg {
	unsigned int regoff;
	size_t size;
};

struct sensor_info {
	const char *name;
	const char *type;
	unsigned int id;
	unsigned int multiplier;
	struct sensor_reg regs[SENSOR_REG_MAX];
};

struct sensor_obj {
	struct sensor_info *info;
	struct kobject kobj;
	const struct attribute_group **attr_groups;
	struct list_head node;
	void *priv;
};

struct sensor_obj_attribute {
	struct attribute attr;
	ssize_t (*show)(struct sensor_obj *sobj, char *buf);
	ssize_t (*store)(struct sensor_obj *sobj, const char *buf, size_t n);
};

#define to_sensor_obj_attr(_attr)					\
		container_of(_attr, struct sensor_obj_attribute, attr)
#define to_sensor_obj(_kobj) \
		container_of(_kobj, struct sensor_obj, kobj)

static ssize_t sensor_obj_attr_show(struct kobject *kobj,
				    struct attribute *attr, char *buf)
{
	struct sensor_obj_attribute *sensor_attr = to_sensor_obj_attr(attr);
	struct sensor_obj *sobj = to_sensor_obj(kobj);
	ssize_t ret = -EIO;

	if (sensor_attr->show)
		ret = sensor_attr->show(sobj, buf);
	return ret;
};

static ssize_t sensor_obj_attr_store(struct kobject *kobj,
				     struct attribute *attr, const char *buf,
				     size_t n)
{
	struct sensor_obj_attribute *sensor_attr = to_sensor_obj_attr(attr);
	struct sensor_obj *sobj = to_sensor_obj(kobj);
	ssize_t ret = -EIO;

	if (sensor_attr->store)
		ret = sensor_attr->store(sobj, buf, n);
	return ret;
}

static const struct sysfs_ops sensor_obj_sysfs_ops = {
	.show = sensor_obj_attr_show,
	.store = sensor_obj_attr_store,
};

static void sensor_obj_release(struct kobject *kobj)
{
	kfree(to_sensor_obj(kobj));
}

static struct kobj_type sensor_obj_ktype = {
	.sysfs_ops = &sensor_obj_sysfs_ops,
	.release = sensor_obj_release,
};

static struct sensor_obj *
create_sensor_obj(struct kobject *parent, struct sensor_info *info,
		  const struct attribute_group **groups, const char *name,
		  void *priv)
{
	struct sensor_obj *sobj;
	int ret;

	sobj = kzalloc(sizeof(*sobj), GFP_KERNEL);
	if (!sobj)
		return ERR_PTR(-ENOMEM);

	sobj->info = info;
	sobj->priv = priv;
	sobj->attr_groups = groups;
	INIT_LIST_HEAD(&sobj->node);

	ret = kobject_init_and_add(&sobj->kobj, &sensor_obj_ktype, parent,
				   "%s%d", name, info->id);
	if (ret)
		goto put_exit;

	if (sobj->attr_groups) {
		ret = sysfs_create_groups(&sobj->kobj, sobj->attr_groups);
		if (ret)
			goto put_exit;
	}

	return sobj;

put_exit:
	kobject_put(&sobj->kobj);
	return ERR_PTR(ret);
}

static void destroy_sensor_obj(struct sensor_obj *sobj)
{
	kobject_put(&sobj->kobj);
}

#define SENSOR_OBJ_ATTR_RO(_name)				\
	struct sensor_obj_attribute sensor_obj_attr_##_name = __ATTR_RO(_name)

static ssize_t name_show(struct sensor_obj *sobj, char *buf)
{
	return sprintf(buf, "%s\n", sobj->info->name);
}
static SENSOR_OBJ_ATTR_RO(name);

static ssize_t id_show(struct sensor_obj *sobj, char *buf)
{
	return sprintf(buf, "%u\n", sobj->info->id);
}
static SENSOR_OBJ_ATTR_RO(id);

static ssize_t type_show(struct sensor_obj *sobj, char *buf)
{
	return sprintf(buf, "%s\n", sobj->info->type);
}
static SENSOR_OBJ_ATTR_RO(type);

static bool sensor_reg_valid(struct sensor_reg *reg)
{
	return !!reg->size;
}

static int sensor_reg_read(struct sensor_obj *sobj, unsigned int reg_idx,
			   char *buf)
{
	struct intel_max10 *max10 = sobj->priv;
	struct sensor_info *info = sobj->info;
	struct sensor_reg *regs = info->regs;
	int ret = -ENODEV;
	u32 value;

	if (reg_idx < SENSOR_REG_MAX && sensor_reg_valid(&regs[reg_idx])) {
		ret = max10_handshake_read(max10, regs[reg_idx].regoff, &value);
		if (ret)
			return ret;

		if (value == 0xdeadbeef)
			return -EBUSY;

		return sprintf(buf, "%u\n", value * info->multiplier);
	}

	return ret;
}

#define SENSOR_REG_SHOW(name, reg_idx)					\
static ssize_t name##_show(struct sensor_obj *sobj, char *buf)		\
{									\
	return sensor_reg_read(sobj, reg_idx, buf);			\
}									\
static SENSOR_OBJ_ATTR_RO(name)

SENSOR_REG_SHOW(value, SENSOR_REG_VALUE);
SENSOR_REG_SHOW(high_warn, SENSOR_REG_HIGH_WARN);
SENSOR_REG_SHOW(high_fatal, SENSOR_REG_HIGH_FATAL);
SENSOR_REG_SHOW(low_warn, SENSOR_REG_LOW_WARN);
SENSOR_REG_SHOW(low_fatal, SENSOR_REG_LOW_FATAL);
SENSOR_REG_SHOW(hysteresis, SENSOR_REG_HYSTERESIS);

static struct attribute *sensor_attrs[] = {
	&sensor_obj_attr_name.attr,
	&sensor_obj_attr_id.attr,
	&sensor_obj_attr_type.attr,
	&sensor_obj_attr_value.attr,
	&sensor_obj_attr_high_warn.attr,
	&sensor_obj_attr_high_fatal.attr,
	&sensor_obj_attr_low_warn.attr,
	&sensor_obj_attr_low_fatal.attr,
	&sensor_obj_attr_hysteresis.attr,
	NULL,
};

static struct attribute_group sensor_attr_group = {
	.attrs = sensor_attrs,
};

static const struct attribute_group *sensor_obj_attr_groups[] = {
	&sensor_attr_group,
	NULL,
};

static void max10_sensor_uinit(struct intel_max10 *max10)
{
	struct sensor_obj *sobj, *sobj_tmp;

	list_for_each_entry_safe(sobj, sobj_tmp, &max10->sobj_list, node) {
		list_del(&sobj->node);
		destroy_sensor_obj(sobj);
	}
}

static int max10_sensor_init(struct intel_max10 *max10, const char *fdt_root,
			     int parent)
{
	struct sensor_info *info;
	struct sensor_obj *sobj;
	int i, ret = 0, offset = 0;
	const fdt32_t *num;
	const char *ptr;
	u64 start, size;

	INIT_LIST_HEAD(&max10->sobj_list);

	/* skip sensor initialization if subdevice table is not found */
	if (!fdt_root) {
		dev_dbg(max10->dev, "skip sensor initialization as not DT\n");
		return 0;
	}

	fdt_for_each_subnode(offset, fdt_root, parent) {
		ptr = fdt_get_name(fdt_root, offset, NULL);
		if (!ptr) {
			dev_err(max10->dev, "failed to fdt_get_name()\n");
			continue;
		}

		if (!strstr(ptr, "sensor")) {
			dev_dbg(max10->dev, "%s is not a sensor node\n", ptr);
			continue;
		}

		dev_dbg(max10->dev, "sensor node %s found\n", ptr);

		info = devm_kzalloc(max10->dev, sizeof(*info), GFP_KERNEL);
		if (!info) {
			ret = -ENOMEM;
			goto uinit_sensor;
		}

		info->name = devm_fdt_strdup(max10->dev, fdt_root, offset,
					     "sensor_name", NULL);
		if (!info->name) {
			ret = -EINVAL;
			goto uinit_sensor;
		}

		info->type = devm_fdt_strdup(max10->dev, fdt_root, offset,
					     "type", NULL);
		if (!info->type) {
			ret = -EINVAL;
			goto uinit_sensor;
		}

		for (i = SENSOR_REG_VALUE; i < SENSOR_REG_MAX; i++) {
			ret = fdt_get_named_reg(fdt_root, offset,
						sensor_reg_name[i], &start,
						&size);
			if (ret) {
				dev_dbg(max10->dev, "sensor node %s no %s\n",
					ptr, sensor_reg_name[i]);

				if (i == SENSOR_REG_VALUE) {
					ret = -EINVAL;
					goto uinit_sensor;
				}

				continue;
			}

			/* This is a hack to compatible with non-secure
			 * solution. If sensors are included in root node,
			 * then it's non-secure dtb, which use absolute addr
			 * of non-secure solution.
			 */
			if (parent)
				info->regs[i].regoff = start;
			else
				info->regs[i].regoff = start - MAX10_BASE_ADDR;
			info->regs[i].size = size;
		}

		num = fdt_getprop(fdt_root, offset, "id", NULL);
		if (!num) {
			ret = -EINVAL;
			goto uinit_sensor;
		}

		info->id = fdt32_to_cpu(*num);

		num = fdt_getprop(fdt_root, offset, "multiplier", NULL);

		info->multiplier = num ? fdt32_to_cpu(*num) : 1;

		dev_dbg(max10->dev, "sensor info: %s: %s: %u: %u\n",
			info->name, info->type, info->id, info->multiplier);

		for (i = SENSOR_REG_VALUE; i < SENSOR_REG_MAX; i++)
			dev_dbg(max10->dev, "sensor reg[%d]: %x: %zu\n",
				i, info->regs[i].regoff, info->regs[i].size);

		sobj = create_sensor_obj(&max10->dev->kobj, info,
					 sensor_obj_attr_groups, "sensor",
					 max10);
		if (IS_ERR(sobj)) {
			ret = PTR_ERR(sobj);
			goto uinit_sensor;
		}

		list_add(&sobj->node, &max10->sobj_list);
	}

	return 0;

uinit_sensor:
	max10_sensor_uinit(max10);
	return ret;
}

static int max10_system_manager_init(struct intel_max10 *max10,
				     const char *fdt_root)
{
	int offset, sysmgr_offset = 0;

	if (!fdt_root)
		return -ENODEV;

	fdt_for_each_subnode(offset, fdt_root, 0) {
		if (!fdt_node_check_compatible(fdt_root, offset,
					       "intel-max10,system-manager")) {
			sysmgr_offset = offset;
			break;
		}
	}

	/* If no system-manager node is found, it implies all system subnodes
	 * are included in root node.
	 */
	return max10_sensor_init(max10, fdt_root, sysmgr_offset);
}

static void max10_system_manager_uinit(struct intel_max10 *max10)
{
	if (!max10->fdt_exist)
		return;

	max10_sensor_uinit(max10);
}

static int max10_staging_area_init(struct intel_max10 *max10,
				   const char *fdt_root)
{
	struct device *dev = max10->dev;
	int ret, offset = 0;
	u64 start, size;

	if (!fdt_root)
		return -ENODEV;

	fdt_for_each_subnode(offset, fdt_root, 0) {
		if (fdt_node_check_compatible(fdt_root, offset,
					      "ifpga-sec-mgr,staging-area"))
			continue;

		ret = fdt_get_reg(fdt_root, offset, 0, &start, &size);
		if (!ret) {
			max10->staging_area.base = start;
			max10->staging_area.size = size;

			dev_dbg(dev, "staging area base %llx size %llx\n",
				start, size);
		}

		return ret;
	}

	return -ENODEV;
}

/* PKVL related registers, in system register region */
#define PKVL_POLLING_CTRL		0x80
#define   POLLING_MODE			GENMASK(15, 0)
#define   PKVL_A_PRELOAD		BIT(16)
#define   PKVL_A_PRELOAD_TIMEOUT	BIT(17)
#define   PKVL_A_DATA_TOO_BIG		BIT(18)
#define   PKVL_A_HDR_CHECKSUM		BIT(20)
#define   PKVL_B_PRELOAD		BIT(24)
#define   PKVL_B_PRELOAD_TIMEOUT	BIT(25)
#define   PKVL_B_DATA_TOO_BIG		BIT(26)
#define   PKVL_B_HDR_CHECKSUM		BIT(28)
#define   PKVL_EEPROM_UPG_STATUS	GENMASK(31, 16)
#define PKVL_LINK_STATUS		0x164
#define PKVL_A_VERSION			0x254
#define PKVL_B_VERSION			0x258
#define   SERDES_VERSION		GENMASK(15, 0)
#define   SBUS_VERSION			GENMASK(31, 16)

#define PKVL_PRELOAD		(PKVL_A_PRELOAD | PKVL_B_PRELOAD)
#define PKVL_PRELOAD_TIMEOUT	(PKVL_A_PRELOAD_TIMEOUT |\
				 PKVL_B_PRELOAD_TIMEOUT)
#define PKVL_DATA_TOO_BIG	(PKVL_A_DATA_TOO_BIG | PKVL_B_DATA_TOO_BIG)
#define PKVL_HDR_CHECKSUM	(PKVL_A_HDR_CHECKSUM | PKVL_B_HDR_CHECKSUM)

#define PKVL_UPG_STATUS_MASK	(PKVL_PRELOAD | PKVL_PRELOAD_TIMEOUT |\
				 PKVL_DATA_TOO_BIG | PKVL_HDR_CHECKSUM)
#define PKVL_UPG_STATUS_GOOD	(PKVL_PRELOAD | PKVL_HDR_CHECKSUM)

/* interval 100ms and timeout 2s */
#define PKVL_EEPROM_LOAD_INTERVAL_US	(100 * 1000)
#define PKVL_EEPROM_LOAD_TIMEOUT_US	(2 * 1000 * 1000)

/* interval 100ms and timeout 30s */
#define PKVL_PRELOAD_INTERVAL_US	(100 * 1000)
#define PKVL_PRELOAD_TIMEOUT_US		(30 * 1000 * 1000)

static int get_pkvl_polling_mode(struct intel_max10 *max10, u16 *mode)
{
	int ret;
	unsigned int val;

	ret = max10_sys_read(max10, PKVL_POLLING_CTRL, &val);
	if (ret)
		return ret;

	*mode = (u16)FIELD_GET(POLLING_MODE, val);

	return ret;
}

static int get_pkvl_polling_mode_sec(struct intel_max10 *max10, u16 *mode)
{
	int ret;
	unsigned int val;

	ret = max10_sys_read(max10, MAX10_DOORBELL, &val);
	if (ret)
		return ret;

	*mode = (u16)FIELD_GET(PKVL1_POLL_EN | PKVL2_POLL_EN, val);

	return ret;
}

static ssize_t polling_mode_show(struct device *dev,
				 struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	u16 mode;
	int ret;

	if (max10->flags | MAX10_FLAGS_SECURE)
		ret = get_pkvl_polling_mode_sec(max10, &mode);
	else
		ret = get_pkvl_polling_mode(max10, &mode);

	if (ret)
		return ret;

	return sprintf(buf, "0x%x\n", mode);
}
static DEVICE_ATTR_RO(polling_mode);

static int trigger_pkvl_eeprom_load(struct intel_max10 *max10,
				    unsigned int reg, unsigned int bit_mask)
{
	unsigned int val;
	int ret;

	ret = max10_sys_update_bits(max10, reg, bit_mask, bit_mask);
	if (ret)
		return ret;

	/*
	 * If current NIOS FW supports this PKVL update feature, then it will
	 * clear the same PKVL_EEPROM_LOAD bit in 2 seconds. If current NIOS
	 * FW doesn't support PKVL update feature, then driver needs to clear
	 * up the PKVL_EEPROM_LOAD bit manually and return error code.
	 */
	ret = regmap_read_poll_timeout(max10->regmap, max10->base + reg,
				       val, (!(val & bit_mask)),
				       PKVL_EEPROM_LOAD_INTERVAL_US,
				       PKVL_EEPROM_LOAD_TIMEOUT_US);
	if (ret == -ETIMEDOUT) {
		dev_err(max10->dev, "%s PKVL_EEPROM_LOAD clear timedout\n",
			__func__);
		max10_sys_update_bits(max10, reg, bit_mask, 0);
		ret = -ENODEV;
	} else if (ret) {
		dev_err(max10->dev, "%s poll EEPROM_LOAD error %d\n",
			__func__, ret);
	}

	return ret;
}

static int pkvl_eeprom_load(struct intel_max10 *max10)
{
	return trigger_pkvl_eeprom_load(max10, RSU_REG, PKVL_EEPROM_LOAD);
}

static int pkvl_eeprom_load_sec(struct intel_max10 *max10)
{
	unsigned int doorbell;
	int ret;

	ret = max10_sys_read(max10, MAX10_DOORBELL, &doorbell);
	if (ret)
		return ret;

	switch (sec_prog(doorbell)) {
	case RSU_PROG_IDLE:
	case RSU_PROG_RSU_DONE:
	case RSU_PROG_PKVL_PROM_DONE:
		break;
	default:
		dev_err(max10->dev, "%s Doorbell error status: 0x%08x\n",
			__func__, doorbell);
		return -EBUSY;
	}

	ret = trigger_pkvl_eeprom_load(max10, MAX10_DOORBELL,
				       PKVL_EEPROM_LOAD_SEC);
	if (ret)
		return ret;

	ret = regmap_read_poll_timeout(max10->regmap,
				       max10->base + MAX10_DOORBELL, doorbell,
				       ((sec_prog(doorbell) ==
					 RSU_PROG_PKVL_PROM_DONE) ||
					(sec_stat(doorbell) ==
					 RSU_STAT_PKVL_REJECT)),
				       PKVL_PRELOAD_INTERVAL_US,
				       PKVL_PRELOAD_TIMEOUT_US);
	if (ret) {
		if (ret == -ETIMEDOUT)
			dev_err(max10->dev,
				"%s Doorbell check timedout: 0x%08x\n",
				__func__, doorbell);
		else
			dev_err(max10->dev, "%s poll Doorbell error\n",
				__func__);
		return ret;
	}

	if (sec_stat(doorbell) == RSU_STAT_PKVL_REJECT) {
		dev_err(max10->dev, "%s rejected\n", __func__);
		return -ECANCELED;
	}

	return 0;
}

static ssize_t eeprom_load_store(struct device *dev,
				 struct device_attribute *attr,
				 const char *buf, size_t count)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	unsigned int val;
	int ret;

	ret = kstrtouint(buf, 0, &val);
	if (ret)
		return ret;

	if (val != 1) {
		dev_err(dev, "%s bad eeprom load val = %d\n", __func__, val);
		return -EINVAL;
	}

	/*
	 * hold lock to prevent current NIOS FW upgrade. If NIOS FW upgrade is
	 * on-going (holding NIOS in reset), then return -EBUSY.
	 */
	down_write(&max10->bmcfw_lock);
	if (max10->bmcfw_flash_enabled) {
		ret = -EBUSY;
		goto done;
	}

	if (max10->flags | MAX10_FLAGS_SECURE)
		ret = pkvl_eeprom_load_sec(max10);
	else
		ret = pkvl_eeprom_load(max10);

	if (ret)
		goto done;

	ret = regmap_read_poll_timeout(max10->regmap,
				       max10->base + PKVL_POLLING_CTRL, val,
				       ((val & PKVL_PRELOAD) == PKVL_PRELOAD),
				       PKVL_PRELOAD_INTERVAL_US,
				       PKVL_PRELOAD_TIMEOUT_US);
	if (ret) {
		dev_err(dev, "%s poll PKVL_PRELOAD error %d\n", __func__, ret);
		goto done;
	}

	if ((val & PKVL_UPG_STATUS_MASK) != PKVL_UPG_STATUS_GOOD) {
		dev_err(dev, "%s error detected during upgrade\n", __func__);
		ret = -EIO;
	}

done:
	up_write(&max10->bmcfw_lock);

	return ret ? : count;
}
static DEVICE_ATTR_WO(eeprom_load);

static ssize_t eeprom_update_status_show(struct device *dev,
					 struct device_attribute *attr,
					 char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	unsigned int val;
	int ret;

	ret = max10_handshake_read(max10, PKVL_POLLING_CTRL, &val);
	if (ret)
		return ret;

	val &= PKVL_UPG_STATUS_MASK;

	return sprintf(buf, "0x%x\n",
		       (u16)FIELD_GET(PKVL_EEPROM_UPG_STATUS, val));
}
static DEVICE_ATTR_RO(eeprom_update_status);

static ssize_t pkvl_a_version_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	unsigned int val;
	int ret;

	ret = max10_handshake_read(max10, PKVL_A_VERSION, &val);
	if (ret)
		return ret;

	return sprintf(buf, "%04x.%04x\n",
		       (u16)FIELD_GET(SBUS_VERSION, val),
		       (u16)FIELD_GET(SERDES_VERSION, val));
}
static DEVICE_ATTR_RO(pkvl_a_version);

static ssize_t pkvl_b_version_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	unsigned int val;
	int ret;

	ret = max10_handshake_read(max10, PKVL_B_VERSION, &val);
	if (ret)
		return ret;

	return sprintf(buf, "%04x.%04x\n",
		       (u16)FIELD_GET(SBUS_VERSION, val),
		       (u16)FIELD_GET(SERDES_VERSION, val));
}
static DEVICE_ATTR_RO(pkvl_b_version);

static ssize_t status_show(struct device *dev,
			   struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	unsigned int val;
	int ret;

	ret = max10_handshake_read(max10, PKVL_LINK_STATUS, &val);
	if (ret)
		return ret;

	return sprintf(buf, "0x%x\n", val);
}
static DEVICE_ATTR_RO(status);

static struct attribute *pkvl_attrs[] = {
	&dev_attr_polling_mode.attr,
	&dev_attr_eeprom_load.attr,
	&dev_attr_eeprom_update_status.attr,
	&dev_attr_pkvl_a_version.attr,
	&dev_attr_pkvl_b_version.attr,
	&dev_attr_status.attr,
	NULL,
};

static struct attribute_group pkvl_attr_group = {
	.name	= "pkvl",
	.attrs	= pkvl_attrs,
};

static unsigned int max10_get_parent_id(struct intel_max10 *max10)
{
	struct platform_device *pdev;

	if (!max10->dev || !max10->dev->parent || !max10->dev->parent->parent ||
	    !max10->dev->parent->parent->parent)
		return 0xffff;

	pdev = to_platform_device(max10->dev->parent->parent->parent);

	dev_info(max10->dev, "%s %s\n", __func__, dev_name(&pdev->dev));

	return pdev->id;
}

static struct platform_device *create_c827_pdev(struct intel_max10 *max10,
						unsigned int id)
{
	struct c827_retimer_platdata pdata;
	struct platform_device *pdev;
	int ret;

	pdev = platform_device_alloc(C827_RETIMER_DRV, PLATFORM_DEVID_AUTO);
	if (!pdev)
		return NULL;

	pdata.regmap = max10->regmap;
	pdata.parent_id = max10_get_parent_id(max10);
	pdata.id = id;
	pdata.addr = max10->base + PKVL_LINK_STATUS;

	ret = platform_device_add_data(pdev, &pdata, sizeof(pdata));
	if (ret) {
		platform_device_put(pdev);
		return NULL;
	}

	ret = platform_device_add(pdev);
	if (ret) {
		platform_device_put(pdev);
		return NULL;
	}

	return pdev;
}

static void remove_c827_pdev(struct platform_device *pdev)
{
	if (!pdev)
		return;

	platform_device_unregister(pdev);
}

static int max10_pkvl_init(struct intel_max10 *max10)
{
	int ret;

	if (!is_max10_pac_n3000(max10))
		return 0;

	ret = sysfs_create_group(&max10->dev->kobj, &pkvl_attr_group);
	if (ret)
		return ret;

	/* each max10 has 2 pkvl c827 chip */
	max10->pv0_pdev = create_c827_pdev(max10, 0);
	max10->pv1_pdev = create_c827_pdev(max10, 1);

	if (!max10->pv0_pdev || !max10->pv1_pdev) {
		remove_c827_pdev(max10->pv0_pdev);
		remove_c827_pdev(max10->pv1_pdev);
		sysfs_remove_group(&max10->dev->kobj, &pkvl_attr_group);
		return -ENODEV;
	}

	return ret;
}

static void max10_pkvl_uinit(struct intel_max10 *max10)
{
	if (!is_max10_pac_n3000(max10))
		return;

	remove_c827_pdev(max10->pv0_pdev);
	remove_c827_pdev(max10->pv1_pdev);
	sysfs_remove_group(&max10->dev->kobj, &pkvl_attr_group);
}

struct reg_info {
	const char *name;
	unsigned int regoff;
	size_t size;
};

/**
 * struct max10_subdev_info - describe a subdevice of MAX10
 *
 * @dev_name: subdev name used to distinguish different devs with same driver.
 * @modalias: matched module/driver name.
 * @regs: reg_info array for subdev.
 * @num_regs: array szie of the reg_info.
 * @attr_group: attribute group to create under MAX10.
 * @gen_pdata: callback to create platform data for subdev.
 */
struct max10_subdev_info {
	const char *dev_name;
	const char *modalias;
	struct reg_info *regs;
	unsigned int num_regs;
	struct attribute_group *attr_group;

	void *(*gen_pdata)(struct intel_max10 *max10,
			   struct max10_subdev_info *info, size_t *pdata_size);
};

static struct reg_info *get_reg_by_name(struct max10_subdev_info *info,
					const char *name)
{
	struct reg_info *reg;

	for (reg = info->regs; reg < info->regs + info->num_regs; reg++)
		if (reg->name && !strcmp(reg->name, name))
			return reg;

	return NULL;
}

#define FPGA_FLASH_CTLR		"fpga_flash"
#define BMCFW_FLASH_CTLR	"bmcfw_flash"
#define I2C_MASTER2_CTLR	"i2c_master2"
#define BMCIMG_FLASH_CTLR	"bmcimg_flash"
#define TMP_FLASH_CTLR		"fpga_flash2"

static void *intel_igsf_gen_pdata(struct intel_max10 *max10,
				  struct max10_subdev_info *info,
				  size_t *pdata_size)
{
	struct intel_generic_qspi_platform_data *pdata;
	struct spi_avmm_bridge *br = max10->bridge;
	struct reg_info *reg;

	pdata = devm_kzalloc(max10->dev, sizeof(*pdata), GFP_KERNEL);
	if (!pdata)
		return ERR_PTR(-ENOMEM);

	reg = get_reg_by_name(info, "csr");
	if (!reg)
		return ERR_PTR(-ENODEV);

	pdata->regmap_ctx = regmap_spi_avmm_gen_ctx(br, reg->regoff, reg->size);
	if (IS_ERR(pdata->regmap_ctx))
		return ERR_CAST(pdata->regmap_ctx);

	reg = get_reg_by_name(info, "mem");
	if (!reg)
		return ERR_PTR(-ENODEV);

	pdata->data_regmap_ctx = regmap_spi_avmm_gen_ctx(br, reg->regoff,
							 reg->size);
	if (IS_ERR(pdata->data_regmap_ctx))
		return ERR_CAST(pdata->data_regmap_ctx);

	pdata->regmap_bus = regmap_spi_avmm_bus;

	*pdata_size = sizeof(*pdata);

	return pdata;
}

static void *intel_i2c_gen_pdata(struct intel_max10 *max10,
				 struct max10_subdev_info *info,
				 size_t *pdata_size)
{
	struct spi_avmm_bridge *br = max10->bridge;
	struct altr_i2c_platform_data *pdata;
	struct i2c_board_info *i2c_info;
	struct reg_info *reg;
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,6,0)
	const char *clk_name;
#endif

	pdata = devm_kzalloc(max10->dev, sizeof(*pdata), GFP_KERNEL);
	if (!pdata)
		return ERR_PTR(-ENOMEM);

	reg = get_reg_by_name(info, "csr");
	if (!reg)
		return ERR_PTR(-ENODEV);

	pdata->regmap_ctx = regmap_spi_avmm_gen_ctx(br, reg->regoff, reg->size);
	if (IS_ERR(pdata->regmap_ctx))
		return ERR_CAST(pdata->regmap_ctx);

	i2c_info = devm_kzalloc(max10->dev, sizeof(*i2c_info), GFP_KERNEL);
	if (!i2c_info)
		return ERR_PTR(-ENOMEM);

#if LINUX_VERSION_CODE > KERNEL_VERSION(4,6,0)
	/*
	 * TBD: Should we get below information from device tree table instead.
	 */
	clk_name = devm_kasprintf(max10->dev, GFP_KERNEL, "%s-i2c2-clk",
				  dev_name(max10->dev));
	if (!clk_name)
		return ERR_PTR(-ENOMEM);
#endif

	snprintf(i2c_info->type, sizeof(i2c_info->type), "24c128");
	i2c_info->addr = 0x57;

	pdata->num_devices = 1;
	pdata->devices = i2c_info;
	pdata->bus_clk_rate = 100000;
	pdata->fifo_size = 64;
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,6,0)
	pdata->i2c_clk = clk_register_fixed_rate(max10->dev, clk_name, NULL, 0,
						 50000000);
	if (IS_ERR(pdata->i2c_clk)) {
		dev_err(max10->dev, "%s clk_register_fixed_rate failed\n",
			__func__);
		return pdata->i2c_clk;
	}
#endif
	pdata->regbus = regmap_spi_avmm_bus;

	*pdata_size = sizeof(*pdata);

	return pdata;
}

static void *intel_iocf_gen_pdata(struct intel_max10 *max10,
				  struct max10_subdev_info *info,
				  size_t *pdata_size)
{
	struct intel_on_chip_flash_platform_data *pdata;
	struct spi_avmm_bridge *br = max10->bridge;
	struct reg_info *reg;

	pdata = devm_kzalloc(max10->dev, sizeof(*pdata), GFP_KERNEL);
	if (!pdata)
		return ERR_PTR(-ENOMEM);

	reg = get_reg_by_name(info, "csr");
	if (!reg)
		return ERR_PTR(-ENODEV);

	pdata->regmap_ctx = regmap_spi_avmm_gen_ctx(br, reg->regoff, reg->size);
	if (IS_ERR(pdata->regmap_ctx))
		return ERR_CAST(pdata->regmap_ctx);

	reg = get_reg_by_name(info, "mem");
	if (!reg)
		return ERR_PTR(-ENODEV);

	pdata->data_regmap_ctx = regmap_spi_avmm_gen_ctx(br, reg->regoff,
							 reg->size);
	if (IS_ERR(pdata->data_regmap_ctx))
		return ERR_CAST(pdata->data_regmap_ctx);

	pdata->regmap_bus = regmap_spi_avmm_bus;

	*pdata_size = sizeof(*pdata);

	return pdata;
}

#define REG_REGION(_name, _regoff, _size) \
	.name = (_name), .regoff = (_regoff), .size = (_size)

static struct reg_info fpga_igsf_reginfo[] = {
	{ REG_REGION("mem", 0x10000000, 0x10000000),},
	{ REG_REGION("csr", 0x20000000, 0x100),},
};

static struct reg_info bmcfw_igsf_reginfo[] = {
	{ REG_REGION("mem", 0x800000, 0x800000),},
	{ REG_REGION("csr", 0x700000, 0x100),},
};

static struct reg_info i2c_master2_reginfo[] = {
	{ REG_REGION("csr", 0x300080, 0x40),},
};

static struct reg_info bmcimg_iocf_reginfo[] = {
	{ REG_REGION("mem", 0x0, 0x00160000),},
	{ REG_REGION("csr", 0x00200000, 0x8),},
};

static ssize_t fpga_flash_mode_show(struct device *dev,
				    struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);

	return sprintf(buf, "%d\n", max10->fpga_flash_enabled);
}

static int fpga_flash_mode_set(struct intel_max10 *max10, bool enable)
{
	int ret;

	if (enable == max10->fpga_flash_enabled)
		return 0;

	if (!enable)
		platform_device_del(max10->fpga_flash);

	ret = max10_flash_switch_set(max10, enable);
	if (ret)
		return ret;

	if (enable) {
		ret = platform_device_add(max10->fpga_flash);
		if (ret) {
			dev_err(max10->dev, "fail to add fpga igsf %d\n", ret);
			return ret;
		}
	}

	max10->fpga_flash_enabled = enable;

	return ret;
}

static ssize_t fpga_flash_mode_store(struct device *dev,
				     struct device_attribute *attr,
				     const char *buf, size_t count)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	bool enable;
	int ret;

	ret = kstrtobool(buf, &enable);
	if (ret)
		return ret;

	mutex_lock(&max10->fpga_flash_lock);
	ret = fpga_flash_mode_set(max10, enable);
	mutex_unlock(&max10->fpga_flash_lock);

	return ret ? : count;
}
static DEVICE_ATTR_RW(fpga_flash_mode);

static int i2c_master2_disable(struct intel_max10 *max10)
{
	int ret;

	if (!max10->i2c_master2_enabled)
		return 0;

	if (bmcfw_handshake_enter(max10))
		return -EBUSY;

	platform_device_del(max10->i2c_master2);

	ret = max10_sys_update_bits(max10, RSU_REG, NIOS2_I2C2_POLL_STOP, 0);
	if (ret)
		return ret;

	max10->i2c_master2_enabled = false;

	bmcfw_handshake_exit(max10);

	dev_dbg(max10->dev, "%s done\n", __func__);

	return ret;
}

static int i2c_master2_enable(struct intel_max10 *max10)
{
	int ret;

	if (max10->i2c_master2_enabled)
		return 0;

	if (bmcfw_handshake_enter(max10))
		return -EBUSY;

	ret = max10_sys_update_bits(max10, RSU_REG, NIOS2_I2C2_POLL_STOP,
				    NIOS2_I2C2_POLL_STOP);
	if (ret)
		goto out;

	msleep(NIOS2_I2C2_POLL_STOP_DELAY_MS);

	ret = platform_device_add(max10->i2c_master2);
	if (ret) {
		dev_err(max10->dev, "%s i2c_master2_add failed %d\n",
			__func__, ret);
		max10_sys_update_bits(max10, RSU_REG, NIOS2_I2C2_POLL_STOP, 0);
		goto out;
	}

	max10->i2c_master2_enabled = true;

	dev_dbg(max10->dev, "%s done\n", __func__);

out:
	bmcfw_handshake_exit(max10);
	return ret;
}

static ssize_t i2c_master2_mode_store(struct device *dev,
				      struct device_attribute *attr,
				      const char *buf, size_t count)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	bool enable;
	int ret;

	ret = kstrtobool(buf, &enable);
	if (ret)
		return ret;

	mutex_lock(&max10->i2c_master2_lock);
	ret = enable ? i2c_master2_enable(max10) : i2c_master2_disable(max10);
	mutex_unlock(&max10->i2c_master2_lock);

	return ret ? : count;
}

static ssize_t i2c_master2_mode_show(struct device *dev,
				     struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);

	return sprintf(buf, "%d\n", max10->i2c_master2_enabled);
}
static DEVICE_ATTR_RW(i2c_master2_mode);

static ssize_t bmcfw_flash_mode_show(struct device *dev,
				     struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);

	return sprintf(buf, "%d\n", max10->bmcfw_flash_enabled);
}

static int bmcfw_flash_disable(struct intel_max10 *max10)
{
	int ret;

	if (!max10->bmcfw_flash_enabled)
		return 0;

	platform_device_del(max10->bmcfw_flash);

	ret = max10_sys_update_bits(max10, RSU_REG,
				    NIOS2_PRERESET | NIOS2_HANG, 0);
	if (ret)
		return ret;

	ret = max10_sys_update_bits(max10, RSU_REG, NIOS2_RESET, 0);
	if (ret)
		return ret;

	max10->bmcfw_flash_enabled = false;

	dev_dbg(max10->dev, "%s done\n", __func__);
	return ret;
}

static int bmcfw_flash_enable(struct intel_max10 *max10)
{
	unsigned int val;
	int ret;

	if (max10->bmcfw_flash_enabled)
		return 0;

	ret = max10_sys_update_bits(max10, RSU_REG, NIOS2_PRERESET,
				    NIOS2_PRERESET);
	if (ret)
		return ret;

	ret = regmap_read_poll_timeout(max10->regmap, max10->base + RSU_REG,
				       val, (val & NIOS2_HANG),
				       NIOS2_HANG_POLL_INTERVAL_US,
				       NIOS2_HANG_POLL_TIMEOUT_US);
	if (ret == -ETIMEDOUT) {
		dev_dbg(max10->dev, "poll NIOS2_HANG timedout\n");
	} else if (ret) {
		dev_err(max10->dev, "fail to poll NIOS2_HANG %d\n", ret);
		return ret;
	}

	ret = max10_sys_update_bits(max10, RSU_REG, NIOS2_RESET, NIOS2_RESET);
	if (ret)
		return ret;

	ret = platform_device_add(max10->bmcfw_flash);
	if (ret) {
		dev_err(max10->dev, "fail to add bmcfw igsf %d\n", ret);
		return ret;
	}

	max10->bmcfw_flash_enabled = true;

	dev_dbg(max10->dev, "%s done\n", __func__);
	return 0;
}

static ssize_t bmcfw_flash_mode_store(struct device *dev,
				      struct device_attribute *attr,
				      const char *buf, size_t count)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	bool enable;
	int ret;

	ret = kstrtobool(buf, &enable);
	if (ret)
		return ret;

	down_write(&max10->bmcfw_lock);
	ret = enable ? bmcfw_flash_enable(max10) : bmcfw_flash_disable(max10);
	up_write(&max10->bmcfw_lock);

	return ret ? : count;
}
static DEVICE_ATTR_RW(bmcfw_flash_mode);

static ssize_t bmcfw_version_show(struct device *dev,
				  struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	unsigned int val;
	int ret;

	ret = max10_sys_read(max10, NIOS2_FW_VERSION, &val);
	if (ret)
		return ret;

	return sprintf(buf, "%x\n", val);
}
static DEVICE_ATTR_RO(bmcfw_version);

static int fpga_image_load_trigger(struct intel_max10 *max10, u8 val)
{
	int ret;

	/* rise edge of RP_LOAD to update FPGA_PAGE_INFO reg */
	ret = max10_sys_update_bits(max10, RSU_REG, FPGA_RP_LOAD, 0);
	if (ret)
		return ret;

	/* allow selecting different pages */
	ret = max10_sys_update_bits(max10, RSU_REG,
				    FPGA_RP_LOAD | FPGA_RECONF_PAGE,
				    FPGA_RP_LOAD |
				    FIELD_PREP(FPGA_RECONF_PAGE, val));
	if (ret)
		return ret;

	ret = max10_sys_update_bits(max10, FPGA_RECONF_REG, COUNTDOWN_START, 0);
	if (ret)
		return ret;

	ret = max10_sys_update_bits(max10, FPGA_RECONF_REG,
				    COUNTDOWN_START, COUNTDOWN_START);
	if (ret)
		return ret;

	return ret;
}

static int fpga_image_load_trigger_sec(struct intel_max10 *max10, u8 val)
{
	int ret;

	/* rise edge of RP_LOAD to update FPGA_PAGE_INFO reg */
	ret = max10_sys_update_bits(max10, FPGA_RECONF_REG, SFPGA_RP_LOAD, 0);
	if (ret)
		return ret;

	/* allow selecting different pages */
	ret = max10_sys_update_bits(max10, FPGA_RECONF_REG,
				    SFPGA_RP_LOAD | SFPGA_RECONF_PAGE,
				    SFPGA_RP_LOAD |
				    FIELD_PREP(SFPGA_RECONF_PAGE, val));
	if (ret)
		return ret;

	ret = max10_sys_update_bits(max10, FPGA_RECONF_REG, COUNTDOWN_START, 0);
	if (ret)
		return ret;

	ret = max10_sys_update_bits(max10, FPGA_RECONF_REG,
				    COUNTDOWN_START, COUNTDOWN_START);
	if (ret)
		return ret;

	return ret;
}

static ssize_t fpga_image_load_store(struct device *dev,
				     struct device_attribute *attr,
				     const char *buf, size_t count)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	u8 val;
	int ret;

	ret = kstrtou8(buf, 0, &val);
	if (ret)
		return ret;

	if ((val != 1) && (val != 0)) {
		dev_err(dev, "%s invalid reload val = %d\n", __func__, val);
		return -EINVAL;
	}

	mutex_lock(&max10->fpga_flash_lock);
	if (max10->fpga_flash_enabled) {
		dev_err(dev, "%s disable fpga_flash_mode first!!\n", __func__);
		ret = -EBUSY;
	} else {
		if (max10->flags | MAX10_FLAGS_SECURE)
			ret = fpga_image_load_trigger_sec(max10, val);
		else
			ret = fpga_image_load_trigger(max10, val);
	}
	mutex_unlock(&max10->fpga_flash_lock);

	return ret ? : count;
}

static ssize_t fpga_image_load_show(struct device *dev,
				    struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	unsigned int val;
	int ret;

	ret = max10_sys_read(max10, FPGA_PAGE_INFO, &val);
	if (ret)
		return ret;

	return sprintf(buf, "%u\n", val);
}
static DEVICE_ATTR_RW(fpga_image_load);

static int bmcimg_flash_mode_set(struct intel_max10 *max10, bool enable)
{
	int ret = 0;

	if (enable == max10->bmcimg_flash_enabled)
		return 0;

	if (!enable) {
		platform_device_del(max10->bmcimg_flash);
	} else {
		ret = platform_device_add(max10->bmcimg_flash);
		if (ret) {
			dev_err(max10->dev, "fail to add fpga igsf %d\n", ret);
			return ret;
		}
	}

	max10->bmcimg_flash_enabled = enable;

	return ret;
}

static ssize_t bmcimg_flash_mode_store(struct device *dev,
				       struct device_attribute *attr,
				       const char *buf, size_t count)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	bool enable;
	int ret;

	ret = kstrtobool(buf, &enable);
	if (ret)
		return ret;

	mutex_lock(&max10->bmcimg_flash_lock);
	ret = bmcimg_flash_mode_set(max10, enable);
	mutex_unlock(&max10->bmcimg_flash_lock);

	return ret ? : count;
}

static ssize_t bmcimg_flash_mode_show(struct device *dev,
				      struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);

	return sprintf(buf, "%d\n", max10->bmcimg_flash_enabled);
}
static DEVICE_ATTR_RW(bmcimg_flash_mode);

static int bmcimg_image_load_trigger(struct intel_max10 *max10, u8 val)
{
	struct device *dev = max10->dev;
	int ret;

	/*
	 * Use Altera Dual Configuration IP Core to reload MAX10 image.
	 *
	 * 1) Select User Page or Factory Page per input parameter.
	 * 2) Trigger reconfiguration.
	 */
	val = (val << 1) | 0x1;

	ret = max10_raw_write(max10, DUAL_CFG_CTRL1, val);
	if (ret)
		return ret;

	ret = max10_raw_write(max10, DUAL_CFG_CTRL0, 0x1);
	if (ret && ret != -EIO) {
		dev_err(dev, "%s fail to trigger reconfig %d\n", __func__, ret);
		return ret;
	}

	return 0;
}

static int bmcimg_image_load_trigger_sec(struct intel_max10 *max10, u8 val)
{
	/*
	 * With security update, DUAL_BOOT IP is not accessible, we have to
	 * use doorbell register instead.
	 */
	return max10_sys_update_bits(max10, MAX10_DOORBELL,
				     CONFIG_SEL | REBOOT_REQ,
				     FIELD_PREP(CONFIG_SEL, val) |
				     REBOOT_REQ);
}

static ssize_t bmcimg_image_load_store(struct device *dev,
				       struct device_attribute *attr,
				       const char *buf, size_t count)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	int ret;
	u8 val;

	ret = kstrtou8(buf, 0, &val);
	if (ret)
		return ret;

	if ((val != 1) && (val != 0)) {
		dev_err(dev, "%s invalid reload val = %d\n", __func__, val);
		return -EINVAL;
	}

	mutex_lock(&max10->bmcimg_flash_lock);
	if (max10->bmcimg_flash_enabled) {
		dev_err(dev, "%s disable bmcimg_flash_mode first!\n", __func__);
		ret = -EBUSY;
	} else {
		if (max10->flags & MAX10_FLAGS_SECURE)
			ret = bmcimg_image_load_trigger_sec(max10, val);
		else
			ret = bmcimg_image_load_trigger(max10, val);
	}
	mutex_unlock(&max10->bmcimg_flash_lock);

	return ret ? : count;
}
static DEVICE_ATTR_WO(bmcimg_image_load);

static struct attribute *fpga_flash_attrs[] = {
	&dev_attr_fpga_flash_mode.attr,
	&dev_attr_fpga_image_load.attr,
	NULL,
};

static struct attribute *fpga_flash_sec_attrs[] = {
	&dev_attr_fpga_image_load.attr,
	NULL,
};

static struct attribute *bmcfw_flash_attrs[] = {
	&dev_attr_bmcfw_version.attr,
	&dev_attr_bmcfw_flash_mode.attr,
	NULL,
};

static struct attribute *bmcfw_flash_sec_attrs[] = {
	&dev_attr_bmcfw_version.attr,
	NULL,
};

static struct attribute *i2c_master2_attrs[] = {
	&dev_attr_i2c_master2_mode.attr,
	NULL,
};

static struct attribute *bmcimg_flash_attrs[] = {
	&dev_attr_bmcimg_flash_mode.attr,
	&dev_attr_bmcimg_image_load.attr,
	NULL,
};

static struct attribute *bmcimg_flash_sec_attrs[] = {
	&dev_attr_bmcimg_image_load.attr,
	NULL,
};

static struct attribute_group fpga_flash_attr_group = {
	.name	= "fpga_flash_ctrl",
	.attrs	= fpga_flash_attrs,
};

static struct attribute_group fpga_flash_sec_attr_group = {
	.name	= "fpga_flash_ctrl",
	.attrs	= fpga_flash_sec_attrs,
};

static struct attribute_group bmcfw_flash_attr_group = {
	.name	= "bmcfw_flash_ctrl",
	.attrs	= bmcfw_flash_attrs,
};

static struct attribute_group bmcfw_flash_sec_attr_group = {
	.name	= "bmcfw_flash_ctrl",
	.attrs	= bmcfw_flash_sec_attrs,
};

static struct attribute_group i2c_master2_attr_group = {
	.name	= "i2c_master2_ctrl",
	.attrs	= i2c_master2_attrs,
};

static struct  attribute_group bmcimg_flash_attr_group = {
	.name	= "bmcimg_flash_ctrl",
	.attrs	= bmcimg_flash_attrs,
};

static struct  attribute_group bmcimg_flash_sec_attr_group = {
	.name	= "bmcimg_flash_ctrl",
	.attrs	= bmcimg_flash_sec_attrs,
};

static struct max10_subdev_info subdev_info[] = {
	{
		.dev_name = FPGA_FLASH_CTLR,
		.modalias = INTEL_GENERIC_QSPI_DRV_NAME,
		.regs = fpga_igsf_reginfo,
		.num_regs = ARRAY_SIZE(fpga_igsf_reginfo),
		.gen_pdata = intel_igsf_gen_pdata,
		.attr_group = &fpga_flash_attr_group,
	},
	{
		.dev_name = BMCFW_FLASH_CTLR,
		.modalias = INTEL_GENERIC_QSPI_DRV_NAME,
		.regs = bmcfw_igsf_reginfo,
		.num_regs = ARRAY_SIZE(bmcfw_igsf_reginfo),
		.gen_pdata = intel_igsf_gen_pdata,
		.attr_group = &bmcfw_flash_attr_group,
	},
	{
		.dev_name = I2C_MASTER2_CTLR,
		.modalias = ALTERA_I2C_DRV_NAME,
		.regs = i2c_master2_reginfo,
		.num_regs = ARRAY_SIZE(i2c_master2_reginfo),
		.gen_pdata = intel_i2c_gen_pdata,
		.attr_group = &i2c_master2_attr_group,
	},
	{
		.dev_name = BMCIMG_FLASH_CTLR,
		.modalias = INTEL_ON_CHIP_FLASH_DRV_NAME,
		.regs = bmcimg_iocf_reginfo,
		.num_regs = ARRAY_SIZE(bmcimg_iocf_reginfo),
		.gen_pdata = intel_iocf_gen_pdata,
		.attr_group = &bmcimg_flash_attr_group,
	},
};

static struct reg_info tmp_igsf_reginfo[] = {
	{ REG_REGION("mem", 0x18000000, 0x8000000),},
	{ REG_REGION("csr", 0x20000100, 0x100),},
};

static struct max10_subdev_info tmp_igsf_info = {
	.dev_name = TMP_FLASH_CTLR,
	.modalias = INTEL_GENERIC_QSPI_DRV_NAME,
	.regs = tmp_igsf_reginfo,
	.num_regs = ARRAY_SIZE(tmp_igsf_reginfo),
	.gen_pdata = intel_igsf_gen_pdata,
};

#define for_each_subdev_info(infos, info) \
	for (info = infos; info < infos + ARRAY_SIZE(infos); info++)

static void max10_subdev_uinit(struct intel_max10 *max10,
			       struct max10_subdev_info *info)
{
	const char *dev_name = info->dev_name;
	struct altr_i2c_platform_data *pdata;

	if (!strcmp(dev_name, FPGA_FLASH_CTLR) && max10->fpga_flash) {
		sysfs_remove_group(&max10->dev->kobj, info->attr_group);
		fpga_flash_mode_set(max10, false);
		platform_device_put(max10->fpga_flash);
		mutex_destroy(&max10->fpga_flash_lock);
	} else if (!strcmp(dev_name, BMCFW_FLASH_CTLR) && max10->bmcfw_flash) {
		sysfs_remove_group(&max10->dev->kobj, info->attr_group);
		bmcfw_flash_disable(max10);
		platform_device_put(max10->bmcfw_flash);
	} else if (!strcmp(dev_name, I2C_MASTER2_CTLR) && max10->i2c_master2) {
		sysfs_remove_group(&max10->dev->kobj, info->attr_group);
		i2c_master2_disable(max10);
		pdata = dev_get_platdata(&max10->i2c_master2->dev);
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,6,0)
		clk_unregister_fixed_rate(pdata->i2c_clk);
#endif
		platform_device_put(max10->i2c_master2);
		mutex_destroy(&max10->i2c_master2_lock);
	} else if (!strcmp(dev_name, BMCIMG_FLASH_CTLR) &&
		   max10->bmcimg_flash) {
		sysfs_remove_group(&max10->dev->kobj, info->attr_group);
		bmcimg_flash_mode_set(max10, false);
		platform_device_put(max10->bmcimg_flash);
		mutex_destroy(&max10->bmcimg_flash_lock);
	} else if (!strcmp(dev_name, TMP_FLASH_CTLR)) {
		platform_device_unregister(max10->tmp_igsf);
	}
}

static void max10_subdevs_uinit(struct intel_max10 *max10)
{
	struct max10_subdev_info *info;

	for_each_subdev_info(subdev_info, info)
		max10_subdev_uinit(max10, info);

	if (max10->flags & MAX10_FLAGS_TMP_IGSF)
		max10_subdev_uinit(max10, &tmp_igsf_info);
}

static int max10_subdev_init(struct intel_max10 *max10,
			     struct max10_subdev_info *info)
{
	struct platform_device **flash_pdev, *pdev;
	bool pdev_add = false;
	size_t pdata_size;
	void *pdata;
	int ret;

	/* Initialize Flash Controller for Flash chips which stores images */
	if (!strcmp(info->dev_name, FPGA_FLASH_CTLR)) {
		flash_pdev = &max10->fpga_flash;
		mutex_init(&max10->fpga_flash_lock);
	} else if (!strcmp(info->dev_name, BMCFW_FLASH_CTLR)) {
		flash_pdev = &max10->bmcfw_flash;
	} else if (!strcmp(info->dev_name, I2C_MASTER2_CTLR)) {
		/* If flags indicate I2C2 is not supported, then skip it */
		if (max10->flags & MAX10_FLAGS_NO_I2C2)
			return 0;

		flash_pdev = &max10->i2c_master2;
		mutex_init(&max10->i2c_master2_lock);
	} else if (!strcmp(info->dev_name, BMCIMG_FLASH_CTLR)) {
		flash_pdev = &max10->bmcimg_flash;
		mutex_init(&max10->bmcimg_flash_lock);
	} else if (!strcmp(info->dev_name, TMP_FLASH_CTLR)) {
		flash_pdev = &max10->tmp_igsf;
		pdev_add = true;
	} else {
		return -EINVAL;
	}

	pdev = platform_device_alloc(info->modalias, PLATFORM_DEVID_AUTO);
	if (!pdev)
		return -ENOMEM;

	pdev->dev.parent = max10->dev;

	if (info->gen_pdata) {
		pdata = info->gen_pdata(max10, info, &pdata_size);
		if (IS_ERR(pdata)) {
			ret = PTR_ERR(pdata);
			goto err;
		}

		ret = platform_device_add_data(pdev, pdata, pdata_size);
		if (ret)
			goto err;
	}

	if (info->attr_group) {
		ret = sysfs_create_group(&max10->dev->kobj, info->attr_group);
		if (ret) {
			dev_err(max10->dev, "failed to create %s attr_group\n",
				info->attr_group->name);
			goto err;
		}
	}

	if (pdev_add) {
		ret = platform_device_add(pdev);
		if (ret)
			goto err;
	}

	*flash_pdev = pdev;
	return 0;

err:
	platform_device_put(pdev);
	return ret;
}

static int max10_subdevs_init(struct intel_max10 *max10)
{
	struct max10_subdev_info *info;
	unsigned int val;
	int ret;

	init_rwsem(&max10->bmcfw_lock);

	ret = max10_sys_read(max10, MAX10_BUILD_VER, &val);
	if (ret)
		return ret;

	if (FIELD_GET(MAX10_VERSION_MAJOR, val) == 111U) {
		/*
		 * These code is only used in non-secure solution, to check
		 * if this is a specific version for ROT upgrade.
		 */
		dev_info(max10->dev, "Temp MAX10 for ROT upgrade is detected\n");
		max10->flags |= MAX10_FLAGS_TMP_IGSF;

		ret = max10_subdev_init(max10, &tmp_igsf_info);
		if (ret)
			goto err;
	}

	for_each_subdev_info(subdev_info, info) {
		ret = max10_subdev_init(max10, info);
		if (ret)
			goto err;
	}

	return 0;

err:
	max10_subdevs_uinit(max10);
	return ret;
}

struct subdev_fdt_parser {
	const char *modalias;
	void *(*gen_pdata)(struct intel_max10 *max10, const char *fdt_root,
			   int node_offset, size_t *pdata_size);
};

static struct subdev_fdt_parser subdev_fdt_parser_tb[] = {
	{ NULL,},
};

#define for_each_parser(parsers, parser) \
	for (parser = parsers; parser < parsers + ARRAY_SIZE(parsers); parser++)

static struct subdev_fdt_parser *find_subdev_fdt_parser(const char *modalias)
{
	struct subdev_fdt_parser *parser;

	for_each_parser(subdev_fdt_parser_tb, parser)
		if (parser->modalias && !strcmp(modalias, parser->modalias))
			return parser;

	return NULL;
}

static int max10_cell_init(struct intel_max10 *max10, const char *modalias,
			   void *pdata, size_t pdata_size,
			   struct mfd_cell *cell)
{
	if (pdata) {
		cell->platform_data = pdata;
		cell->pdata_size = pdata_size;
	}

	cell->name = modalias;
	return 0;
}

static int max10_generic_subdevs_init(struct intel_max10 *max10,
				      const char *fdt_root)
{
	int ret, offset, max_cells = 0, n_cells = 0;
	struct subdev_fdt_parser *parser;
	struct device *dev = max10->dev;
	struct mfd_cell *cells, *cell;
	const char *compat;
	size_t pdata_size;
	void *pdata;

	if (!fdt_root) {
		dev_dbg(dev, "skip generic subdev init as not DT\n");
		return 0;
	}

	fdt_for_each_subnode(offset, fdt_root, 0)
		max_cells++;
	if (!max_cells) {
		dev_dbg(dev, "no max10 subnode found\n");
		return 0;
	}

	cells = devm_kcalloc(dev, max_cells, sizeof(*cell), GFP_KERNEL);
	if (!cells)
		return -ENOMEM;

	cell = cells;
	fdt_for_each_subnode(offset, fdt_root, 0) {
		compat = devm_fdt_strdup(dev, fdt_root, offset, "compatible",
					 NULL);
		if (!compat) {
			dev_dbg(dev, "no compatible for node %s\n",
				fdt_get_name(fdt_root, offset, NULL));
			continue;
		}

		dev_dbg(dev, "%s: %s subdev found\n", __func__, compat);

		parser = find_subdev_fdt_parser(compat);
		if (!parser)
			continue;

		pdata = parser->gen_pdata(max10, fdt_root, offset, &pdata_size);
		if (IS_ERR(pdata))
			return PTR_ERR(pdata);

		ret = max10_cell_init(max10, compat, pdata, pdata_size, cell);
		if (ret)
			return ret;

		n_cells++;
		cell++;
	}

	if (n_cells) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,7,0)
		ret = devm_mfd_add_devices(dev, PLATFORM_DEVID_AUTO, cells,
					   n_cells, NULL, 0, NULL);
#else
		ret = mfd_add_devices(dev, PLATFORM_DEVID_AUTO, cells,
					   n_cells, NULL, 0, NULL);
#endif
		if (ret) {
			dev_err(dev, "fail to add mfd devices: %d\n", ret);
			return ret;
		}
	}

	return 0;
}

static ssize_t compatible_show(struct device *dev,
			       struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);

	if (!max10->id)
		return -ENODEV;

	return sprintf(buf, "%s\n", max10->id->compatible);
}
static DEVICE_ATTR_RO(compatible);

static ssize_t max10_version_show(struct device *dev,
				  struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	unsigned int val;
	int ret;

	ret = max10_sys_read(max10, MAX10_BUILD_VER, &val);
	if (ret)
		return ret;

	return sprintf(buf, "0x%x\n", val);
}
static DEVICE_ATTR_RO(max10_version);

static ssize_t mac_address_show(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	unsigned int macaddr1, macaddr2;
	int ret;

	ret = max10_sys_read(max10, MAX10_MACADDR1, &macaddr1);
	if (ret)
		return ret;

	ret = max10_sys_read(max10, MAX10_MACADDR2, &macaddr2);
	if (ret)
		return ret;

	return sprintf(buf, "%02x:%02x:%02x:%02x:%02x:%02x\n",
		       (u8)FIELD_GET(MAX10_MAC_BYTE1, macaddr1),
		       (u8)FIELD_GET(MAX10_MAC_BYTE2, macaddr1),
		       (u8)FIELD_GET(MAX10_MAC_BYTE3, macaddr1),
		       (u8)FIELD_GET(MAX10_MAC_BYTE4, macaddr1),
		       (u8)FIELD_GET(MAX10_MAC_BYTE5, macaddr2),
		       (u8)FIELD_GET(MAX10_MAC_BYTE6, macaddr2));
}
static DEVICE_ATTR_RO(mac_address);

static ssize_t mac_count_show(struct device *dev,
			      struct device_attribute *attr, char *buf)
{
	struct intel_max10 *max10 = dev_get_drvdata(dev);
	unsigned int macaddr2;
	int ret;

	ret = max10_sys_read(max10, MAX10_MACADDR2, &macaddr2);
	if (ret)
		return ret;

	return sprintf(buf, "%u\n",
		       (u8)FIELD_GET(MAX10_MAC_COUNT, macaddr2));
}
static DEVICE_ATTR_RO(mac_count);

static struct attribute *max10_mac_attrs[] = {
	&dev_attr_mac_address.attr,
	&dev_attr_mac_count.attr,
	NULL,
};

static struct attribute_group max10_mac_attr_group = {
	.attrs = max10_mac_attrs,
};

static struct attribute *max10_attrs[] = {
	&dev_attr_compatible.attr,
	&dev_attr_max10_version.attr,
	NULL,
};

static struct attribute_group max10_attr_group = {
	.attrs = max10_attrs,
};

static const struct attribute_group *max10_sec_attr_groups[] = {
	&max10_attr_group,
	&max10_mac_attr_group,
	&bmcfw_flash_sec_attr_group,
	&fpga_flash_sec_attr_group,
	&bmcimg_flash_sec_attr_group,
	NULL,
};

static inline bool sec_start_done(u32 doorbell)
{
	return (sec_stat(doorbell) == RSU_STAT_ERASE_FAIL ||
		sec_stat(doorbell) == RSU_STAT_WEAROUT ||
		(sec_prog(doorbell) != RSU_PROG_IDLE &&
		 sec_prog(doorbell) != RSU_PROG_RSU_DONE));
}

static void log_error_regs(struct intel_max10 *max10, u32 doorbell)
{
	u32 auth_result;

	dev_err(max10->dev, "RSU error status: 0x%08x\n", doorbell);

	if (!max10_sys_read(max10, MAX10_AUTH_RESULT, &auth_result))
		dev_err(max10->dev, "RSU auth result: 0x%08x\n", auth_result);
}

static int max10_sec_start(struct ifpga_sec_mgr *imgr)
{
	struct intel_max10 *max10 = imgr->priv;
	u32 doorbell;
	int ret;

	ret = max10_sys_read(max10, MAX10_DOORBELL, &doorbell);
	if (ret)
		return ret;

	switch (sec_prog(doorbell)) {
	case RSU_PROG_IDLE:
	case RSU_PROG_RSU_DONE:
		break;
	default:
		log_error_regs(max10, doorbell);
		return -EBUSY;
	}

	down_write(&max10->bmcfw_lock);
	ret = max10_sys_update_bits(max10, MAX10_DOORBELL,
				    RSU_REQUEST | HOST_STATUS, RSU_REQUEST);
	if (ret)
		goto up_write;

	ret = regmap_read_poll_timeout(max10->regmap,
				       max10->base + MAX10_DOORBELL, doorbell,
				       sec_start_done(doorbell),
				       SEC_START_INTERVAL_US,
				       SEC_START_TIMEOUT_US);
	if (ret) {
		log_error_regs(max10, doorbell);
		goto up_write;
	}

	switch (sec_stat(doorbell)) {
	case RSU_STAT_WEAROUT:
		ret = -EAGAIN;
		goto up_write;
	case RSU_STAT_ERASE_FAIL:
		log_error_regs(max10, doorbell);
		ret = -EIO;
		goto up_write;
	}

	max10->bmcfw_flash_enabled = true;
up_write:
	up_write(&max10->bmcfw_lock);
	return ret;
}

static int max10_sec_check_complete(struct ifpga_sec_mgr *imgr)
{
	struct intel_max10 *max10 = imgr->priv;
	u32 doorbell;
	int ret;

	ret = max10_sys_read(max10, MAX10_DOORBELL, &doorbell);
	if (ret)
		return ret;

	switch (sec_stat(doorbell)) {
	case RSU_STAT_NORMAL:
	case RSU_STAT_NIOS_OK:
	case RSU_STAT_USER_OK:
	case RSU_STAT_FACTORY_OK:
	case RSU_STAT_WEAROUT:
		break;
	default:
		log_error_regs(max10, doorbell);
		return -EIO;
	}

	switch (sec_prog(doorbell)) {
	case RSU_PROG_IDLE:
	case RSU_PROG_RSU_DONE:
		return 0;
	case RSU_PROG_AUTHENTICATING:
	case RSU_PROG_COPYING:
	case RSU_PROG_UPDATE_CANCEL:
	case RSU_PROG_PROGRAM_KEY_HASH:
		return -EAGAIN;
	case RSU_PROG_PREPARE:
	case RSU_PROG_READY:
		log_error_regs(max10, doorbell);
		return -EBUSY;
	default:
		log_error_regs(max10, doorbell);
		return -EIO;
	}
}

static int max10_sec_write_blk(struct ifpga_sec_mgr *imgr,
			       struct ifpga_secure_write *wb, char *buf)
{
	struct intel_max10 *max10 = imgr->priv;
	u32 staging_base = max10->staging_area.base;
	u32 staging_size = max10->staging_area.size;
	int stride = REGMAP_STRIDE;
	u32 doorbell;
	int ret;

	if (wb->offset + wb->size > staging_size)
		return -ENOMEM;

	ret = max10_sys_read(max10, MAX10_DOORBELL, &doorbell);
	if (ret)
		return ret;

	if (sec_prog(doorbell) == RSU_PROG_PREPARE) {
		return -EAGAIN;
	} else if (sec_prog(doorbell) != RSU_PROG_READY) {
		log_error_regs(max10, doorbell);
		return -EBUSY;
	}

	return max10_raw_bulk_write(max10, staging_base + wb->offset,
				    buf, wb->size / stride);
}

static int max10_sec_data_sent(struct ifpga_sec_mgr *imgr)
{
	struct intel_max10 *max10 = imgr->priv;
	u32 doorbell;
	int ret;

	ret = max10_sys_read(max10, MAX10_DOORBELL, &doorbell);
	if (ret)
		return ret;

	if (sec_prog(doorbell) != RSU_PROG_READY) {
		log_error_regs(max10, doorbell);
		return -EBUSY;
	}

	down_write(&max10->bmcfw_lock);
	ret = max10_sys_update_bits(max10, MAX10_DOORBELL, HOST_STATUS,
				    FIELD_PREP(HOST_STATUS,
					       HOST_STATUS_WRITE_DONE));
	if (ret)
		goto up_write;

	ret = regmap_read_poll_timeout(max10->regmap,
				       max10->base + MAX10_DOORBELL, doorbell,
				       sec_prog(doorbell) != RSU_PROG_READY,
				       NIOS_HANDSHAKE_INTERVAL_US,
				       NIOS_HANDSHAKE_TIMEOUT_US);
	if (ret) {
		log_error_regs(max10, doorbell);
		goto up_write;
	}

	switch (FIELD_GET(SEC_STATUS, doorbell)) {
	case RSU_STAT_NORMAL:
	case RSU_STAT_NIOS_OK:
	case RSU_STAT_USER_OK:
	case RSU_STAT_FACTORY_OK:
		break;
	default:
		log_error_regs(max10, doorbell);
		ret = -EIO;
		goto up_write;
	}

	max10->bmcfw_flash_enabled = true;
up_write:
	up_write(&max10->bmcfw_lock);
	return ret;
}

static int max10_sec_cancel(struct ifpga_sec_mgr *imgr)
{
	struct intel_max10 *max10 = imgr->priv;
	u32 doorbell;
	int ret;

	ret = max10_sys_read(max10, MAX10_DOORBELL, &doorbell);
	if (ret)
		return ret;

	if (sec_prog(doorbell) != RSU_PROG_READY)
		return -EBUSY;

	return max10_sys_update_bits(max10, MAX10_DOORBELL, HOST_STATUS,
				     FIELD_PREP(HOST_STATUS,
						HOST_STATUS_ABORT_RSU));
}

static const struct ifpga_sec_mgr_ops max10_iops = {
	.start = max10_sec_start,
	.write_blk = max10_sec_write_blk,
	.data_sent = max10_sec_data_sent,
	.check_complete = max10_sec_check_complete,
	.cancel = max10_sec_cancel
};

static void ifpga_sec_mgr_uinit(struct intel_max10 *max10)
{
	if (!max10->imgr)
		return;

	ifpga_sec_mgr_unregister(max10->imgr);
}

static int ifpga_sec_mgr_init(struct intel_max10 *max10, const char *fdt_root)
{
	int ret;

	/* get staging area information for device tree table */
	ret = max10_staging_area_init(max10, fdt_root);
	if (ret)
		return ret;

	max10->imgr = ifpga_sec_mgr_create(max10->dev, dev_name(max10->dev),
					   &max10_iops, max10);
	if (!max10->imgr)
		return -ENOMEM;

	ret = ifpga_sec_mgr_register(max10->imgr, THIS_MODULE);
	if (ret) {
		ifpga_sec_mgr_free(max10->imgr);
		max10->imgr = NULL;
	}
	return ret;
}

static void max10_check_capability(struct intel_max10 *max10,
				   const char *fdt_root)
{
	if (!fdt_root)
		return;

	if (is_max10_pac_n3000(max10))
		max10->flags |= MAX10_FLAGS_NO_I2C2;
	else
		max10->flags |= MAX10_FLAGS_MAC_CACHE;
}

/* cleanup Non-secure MAX10 hardware */
static void max10_hardware_uinit(struct intel_max10 *max10)
{
	max10_pkvl_uinit(max10);
	max10_sensor_uinit(max10);
	max10_subdevs_uinit(max10);
	sysfs_remove_group(&max10->dev->kobj, &max10_attr_group);

	if (max10->flags & MAX10_FLAGS_MAC_CACHE)
		sysfs_remove_group(&max10->dev->kobj, &max10_mac_attr_group);
}

/* Initialize Non-secure MAX10 hardware */
static int max10_hardware_init(struct intel_max10 *max10)
{
	const char *fdt_root = fetch_subdev_table(max10);
	int ret;

	max10_check_capability(max10, fdt_root);

	/*
	 * MAX10 initialization needs below steps
	 *
	 * a) initialize sensors.
	 * b) initialize MAX10 specific function blocks, e.g. flash controller
	 * c) initialize generic subdevice.
	 */
	ret = sysfs_create_group(&max10->dev->kobj, &max10_attr_group);
	if (ret)
		goto done;

	/* MAC address is cached in MAX10 */
	if (max10->flags & MAX10_FLAGS_MAC_CACHE) {
		ret = sysfs_create_group(&max10->dev->kobj,
					 &max10_mac_attr_group);
		if (ret)
			goto exit_remove_group;
	}

	ret = max10_subdevs_init(max10);
	if (ret)
		goto exit_remove_mac_group;

	max10_sensor_init(max10, fdt_root, 0);

	ret = max10_pkvl_init(max10);
	if (ret)
		goto exit_sensor_uinit;

	ret = max10_generic_subdevs_init(max10, fdt_root);
	if (!ret)
		goto done;

	max10_pkvl_uinit(max10);
exit_sensor_uinit:
	max10_sensor_uinit(max10);
	max10_subdevs_uinit(max10);
exit_remove_mac_group:
	if (max10->flags & MAX10_FLAGS_MAC_CACHE)
		sysfs_remove_group(&max10->dev->kobj, &max10_mac_attr_group);
exit_remove_group:
	sysfs_remove_group(&max10->dev->kobj, &max10_attr_group);
done:
	kfree(fdt_root);
	return ret;
}

/* cleanup Secure MAX10 hardware */
static void max10_secure_hw_uinit(struct intel_max10 *max10)
{
	ifpga_sec_mgr_uinit(max10);
	max10_system_manager_uinit(max10);
	max10_pkvl_uinit(max10);
	sysfs_remove_groups(&max10->dev->kobj, max10_sec_attr_groups);

	mutex_destroy(&max10->fpga_flash_lock);
	mutex_destroy(&max10->bmcimg_flash_lock);
}

/* Initialize Secure MAX10 hardware */
static int max10_secure_hw_init(struct intel_max10 *max10)
{
	const char *fdt_root = fetch_subdev_table(max10);
	int ret;

	max10->fdt_exist = !!fdt_root;

	max10_check_capability(max10, fdt_root);

	init_rwsem(&max10->bmcfw_lock);
	mutex_init(&max10->fpga_flash_lock);
	mutex_init(&max10->bmcimg_flash_lock);

	/*
	 * secure MAX10 initialization needs below steps
	 *
	 * a) initialize sysfs entries.
	 * b) initialize sensors per device tree table.
	 * c) initialize pkvl related functions.
	 */
	ret = sysfs_create_groups(&max10->dev->kobj, max10_sec_attr_groups);
	if (ret)
		goto done;

	max10_system_manager_init(max10, fdt_root);
	ret = ifpga_sec_mgr_init(max10, fdt_root);
	if (ret)
		dev_err(max10->dev,
			"Security manager failed to start: %d\n", ret);

	ret = max10_pkvl_init(max10);
	if (!ret)
		goto done;

	ifpga_sec_mgr_uinit(max10);
	max10_system_manager_uinit(max10);
	sysfs_remove_groups(&max10->dev->kobj, max10_sec_attr_groups);
done:
	kfree(fdt_root);
	return ret;
}

static int check_max10_version(struct intel_max10 *max10)
{
	unsigned int v;

	if (!max10_sys_read(max10, MAX10_BASE_ADDR + MAX10_BUILD_VER, &v)) {
		if (v != 0xffffffff) {
			dev_info(max10->dev, "non-secure MAX10 detected\n");
			max10->base = MAX10_BASE_ADDR;
		} else {
			dev_info(max10->dev, "secure MAX10 detected\n");
			max10->base = MAX10_SEC_BASE_ADDR;
			max10->flags |= MAX10_FLAGS_SECURE;
		}
		return 0;
	}

	return -ENODEV;
}

static int intel_max10_spi_probe(struct spi_device *spi)
{
	struct regmap_spi_avmm_ctx *ctx;
	struct spi_avmm_bridge *bridge;
	struct device *dev = &spi->dev;
	struct intel_max10 *max10;
	int ret;

	/* try 32 bits bpw first then fall back to 8 bits bpw */
	if (max10_spi_setup(spi, 32) && max10_spi_setup(spi, 8))
		return -ENODEV;

	max10 = devm_kzalloc(dev, sizeof(*max10), GFP_KERNEL);
	if (!max10)
		return -ENOMEM;

	bridge = spi_avmm_bridge_alloc_and_init(spi);
	if (!bridge)
		return -ENOMEM;

	max10->dev = dev;
	max10->bridge = bridge;

	ctx = regmap_spi_avmm_gen_ctx_full(bridge);
	if (IS_ERR(ctx)) {
		ret = PTR_ERR(ctx);
		goto rm_br;
	}

	max10->regmap = devm_regmap_init(dev, &regmap_spi_avmm_bus, ctx,
					 &intel_max10_regmap_config);
	if (IS_ERR(max10->regmap)) {
		ret = PTR_ERR(max10->regmap);
		dev_err(dev, "Failed to allocate global regmap: %d\n", ret);
		goto rm_br;
	}

	spi_set_drvdata(spi, max10);

	ret = check_max10_version(max10);
	if (ret) {
		dev_err(dev, "Failed to identify max10 hardware\n");
		goto rm_br;
	}

	if (max10->flags & MAX10_FLAGS_SECURE)
		ret = max10_secure_hw_init(max10);
	else
		ret = max10_hardware_init(max10);

	if (ret) {
		dev_err(dev, "Failed to init max10 hardware: %d\n", ret);
		goto rm_br;
	}

	max10->debug = intel_max10_debug_init(dev, max10->regmap);

	return 0;

rm_br:
	spi_avmm_bridge_destroy(max10->bridge);
	return ret;
}

static int intel_max10_spi_remove(struct spi_device *spi)
{
	struct intel_max10 *max10 = spi_get_drvdata(spi);

	intel_max10_debug_exit(max10->debug);

	if (max10->flags & MAX10_FLAGS_SECURE)
		max10_secure_hw_uinit(max10);
	else
		max10_hardware_uinit(max10);

	spi_avmm_bridge_destroy(max10->bridge);

	return 0;
}

static struct spi_driver intel_max10_spi_driver = {
	.probe = intel_max10_spi_probe,
	.remove = intel_max10_spi_remove,
	.driver = {
		.name = "intel-max10",
	},
};

module_spi_driver(intel_max10_spi_driver);

MODULE_DESCRIPTION("Intel Max10 Device Driver");
MODULE_AUTHOR("Intel Corporation");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS("spi:intel-max10");
