// SPDX-License-Identifier: GPL-2.0
/*
 * Driver for Intel SPI Slave to AVMM Bus Bridge
 *
 * Copyright (C) 2018 Intel Corporation. All rights reserved.
 *
 */

#include "intel-spi-avmm.h"
#include <linux/version.h>

/*
 * Read/write protocol for SPI slave to AVMM bus master
 *
 * The SPI interface carries mmio register read/write requirements to its
 * internal AVMM bus. The transfer protocol contains 3 layers: transaction
 * layer, packet layer and physical layer.
 *
 * Reference Documents could be found at:
 * https://www.intel.com/content/www/us/en/programmable/documentation/
 * sfo1400787952932.html
 *
 * Chapter 36 does general introduction about the protocol.
 * Chapter 4 describes Physical layer.
 * Chapter 40 describes Packet layer.
 * Chapter 38 describes Transaction layer.
 */

#define TRANS_CODE_WRITE	0x0
#define TRANS_CODE_SEQ_WRITE	0x4
#define TRANS_CODE_READ		0x10
#define TRANS_CODE_SEQ_READ	0x14
#define TRANS_CODE_NO_TRANS	0x7f

struct trans_header {
	u8 trans_code;
	u8 rsvd;
	__be16 size;
	__be32 addr;
};

struct trans_response {
	u8 r_trans_code;
	u8 rsvd;
	__be16 size;
};

/* temp value before we confirm HW */
#define MAX10_SPI_XFER_TIMEOUT	(msecs_to_jiffies(200))

/* max10 register addr is 32 bits */
#define REG_SIZE		4UL

/* max10 register value is 32 bits */
#define VAL_SIZE		4UL

/*
 * max rx size could be larger. But considering the buffer consuming,
 * it is proper that we limit 1KB xfer at max.
 */
#define MAX_RX_CNT   256UL
#define MAX_TX_CNT   1UL

#define TRANS_HEAD_SIZE		(sizeof(struct trans_header))
#define TRANS_RESP_SIZE		(sizeof(struct trans_response))

#define WR_TRANS_TX_SIZE(n)	(TRANS_HEAD_SIZE + VAL_SIZE * (n))
#define RD_TRANS_TX_SIZE	TRANS_HEAD_SIZE
#define WR_TRANS_RX_SIZE	TRANS_RESP_SIZE

#define TRANS_SIZE_TO_PACKET(n)	(2 * (n) + 4)
#define PACKET_SIZE_TO_PHY(n)	(2 * (n))
#define TRANS_SIZE_TO_PHY(n)	PACKET_SIZE_TO_PHY(TRANS_SIZE_TO_PACKET(n))

#define TRANS_TX_SIZE          WR_TRANS_TX_SIZE(MAX_TX_CNT)
#define PACKET_TX_SIZE         TRANS_SIZE_TO_PACKET(TRANS_TX_SIZE)
#define PHY_TX_SIZE            ALIGN(PACKET_SIZE_TO_PHY(PACKET_TX_SIZE), 4)

/*
 * Unlike tx, rx is bothered by possible PHY_IDLE bytes from slave,
 * See phy_rx_length() on how to calculate the max rx buffer.
 */
#define TRANS_RX_SIZE  (VAL_SIZE * MAX_RX_CNT)
#define PACKET_RX_SIZE TRANS_SIZE_TO_PACKET(TRANS_RX_SIZE)
#define PHY_RX_SIZE    (ALIGN(TRANS_SIZE_TO_PHY(VAL_SIZE) + 3, 4) * MAX_RX_CNT)

/* FIXME: temporary value, need furthe confirmation */
#define SPI_AVMM_XFER_TIMEOUT	(msecs_to_jiffies(200))

/*
 * Transaction Layer
 *
 * Transaction layer capsules transaction code (read/write), size (16 bits),
 * addr (32 bits) and value (32 bits) into transaction format.
 * SEQ_READ/WRITE will read/write number of bytes(specified by head->size)
 * on incrementing addr start from header->addr field.
 * Transaction header will be followed by register data for write operations.
 * Please be noted that size and addr are sent in big endian but value is
 * sent in little endian according to transaction layer protocol.
 */

/* count parameter must be n * VAL_SIZE */
static void trans_tx_prepare(u32 reg, bool is_read, u32 *wr_val, u16 size,
			     char *tx_buf, unsigned int *tx_len)
{
	u16 count = size / VAL_SIZE;
	struct trans_header *header;
	u8 trans_code;
	__le32 *data;
	int i;

	if (is_read)
		if (count == 1)
			trans_code = TRANS_CODE_READ;
		else
			trans_code = TRANS_CODE_SEQ_READ;
	else
		if (count == 1)
			trans_code = TRANS_CODE_WRITE;
		else
			trans_code = TRANS_CODE_SEQ_WRITE;

	header = (struct trans_header *)tx_buf;
	header->trans_code = trans_code;
	header->rsvd = 0;
	header->size = cpu_to_be16(size);
	header->addr = cpu_to_be32(reg);

	if (is_read) {
		*tx_len = RD_TRANS_TX_SIZE;
	} else {
		data = (__le32 *)(tx_buf + TRANS_HEAD_SIZE);

		for (i = 0; i < count; i++)
			*data++ = cpu_to_le32(*wr_val++);

		*tx_len = WR_TRANS_TX_SIZE(count);
	}
}

/*
 * For read transaction, avmm bus will directly response register values
 * without transaction response header.
 */
static int rd_trans_rx_parse(char *rx_buf, unsigned int rx_len, u32 *val)
{
	unsigned int count;
	__le32 *data;
	int i;

	if (!rx_len || !IS_ALIGNED(rx_len, VAL_SIZE))
		return -EINVAL;

	count = rx_len / VAL_SIZE;

	data = (__le32 *)rx_buf;
	for (i = 0; i < count; i++)
		*val++ = le32_to_cpu(*data++);

	return 0;
}

/* For write transaction, max10 will response a transaction response header. */
static int wr_trans_rx_parse(char *rx_buf, unsigned int rx_len,
			     u16 expected_count)
{
	struct trans_response *resp;
	u16 val_len, count;
	u8 trans_code;

	if (rx_len != TRANS_RESP_SIZE)
		return -EINVAL;

	resp = (struct trans_response *)rx_buf;

	trans_code = resp->r_trans_code ^ 0x80;
	val_len = be16_to_cpu(resp->size);
	if (!val_len || !IS_ALIGNED(val_len, VAL_SIZE))
		return -EINVAL;

	count = val_len / VAL_SIZE;

	if (count != expected_count)
		return -EIO;

	if ((count == 1 && trans_code != TRANS_CODE_WRITE) ||
	    (count > 1 && trans_code != TRANS_CODE_SEQ_WRITE))
		return -EFAULT;

	return 0;
}

static int trans_rx_parse(char *rx_buf, unsigned int rx_len, bool is_read,
			  u16 expected_count, u32 *rd_val)
{
	if (is_read) {
		if ((expected_count * VAL_SIZE) != rx_len)
			return -EINVAL;

		return rd_trans_rx_parse(rx_buf, rx_len, rd_val);
	}

	return wr_trans_rx_parse(rx_buf, rx_len, expected_count);
}

/* Packet Layer */

/*
 * Packet layer appends SOP/EOP chars to mark the scope of one packet.
 * It also appends CHANNEL char and the following channel number for
 * internal bus. In order to distinguish special chars and data chars, Packet
 * layer introduce ESCAPE char (0x7d).
 */
#define PKT_SOP		0x7a
#define PKT_EOP		0x7b
#define PKT_CHANNEL	0x7c
#define PKT_ESC		0x7d

static void packet_tx_prepare(char *trans_tx_buf, unsigned int trans_tx_len,
			      char *pkt_tx_buf, unsigned int *pkt_tx_len)
{
	char *b, *p;
	int i;

	b = trans_tx_buf;
	p = pkt_tx_buf;

	*p++ = PKT_SOP;

	/* channel number is always 0 since max10 only support channel 0 */
	*p++ = PKT_CHANNEL;
	*p++ = 0x0;

	for (i = 0; i < trans_tx_len; i++) {
		/* EOP should be inserted before last valid char*/
		if (i == trans_tx_len - 1)
			*p++ = PKT_EOP;

		/* insert ESCAPE char if data value equals any special char */
		if (*b == PKT_SOP || *b == PKT_EOP ||
		    *b == PKT_CHANNEL || *b == PKT_ESC) {
			*p++ = PKT_ESC;
			*p++ = *b ^ 0x20;
		} else {
			*p++ = *b;
		}

		b++;
	}

	*pkt_tx_len = p - pkt_tx_buf;
}

static int packet_rx_parse(char *pkt_rx_buf, unsigned int pkt_rx_len,
			   char *trans_rx_buf, unsigned int *trans_rx_len)
{
	char *b, *p, *sop = NULL;

	b = pkt_rx_buf;
	p = trans_rx_buf;

	/* Find the last SOP */
	while (b < pkt_rx_buf + pkt_rx_len) {
		if (*b == PKT_SOP)
			sop = b;
		b++;
	}

	if (!sop)
		return -EINVAL;

	b = sop + 1;

	while (b < pkt_rx_buf + pkt_rx_len) {
		switch (*b) {
		case PKT_CHANNEL:
			/*
			 * FIXME: in theory, the char after CHANNEL could be
			 * a special char too
			 */
			b += 2;
			break;
		case PKT_ESC:
			b++;
			*p++ = *b++ ^ 0x20;
			break;
		case PKT_EOP:
			/* the char after EOP is the last valid char*/
			b++;

			switch (*b) {
			case PKT_ESC:
			/* the last char may also be escaped */
				b++;
				*p++ = *b++ ^ 0x20;
				break;
			case PKT_SOP:
			case PKT_CHANNEL:
			case PKT_EOP:
				return -EINVAL;
			default:
				*p++ = *b++;
				break;
			}

			*trans_rx_len = p - trans_rx_buf;

			return 0;
		case PKT_SOP:
			return -EINVAL;
		default:
			*p++ = *b++;
			break;
		}
	}

	/* We parsed all the bytes but didn't find EOP */
	return -EINVAL;
}

/* physical layer */
/*
 * Physical layer will insert IDLE char to indicate there is no data in
 * spi send buffer. Similarly Physical layer ESCAPE char is 0x4d.
 */
#define PHY_IDLE	0x4a
#define PHY_ESC		0x4d

static void phy_tx_prepare(char *pkt_tx_buf, unsigned int pkt_tx_len,
			   char *phy_tx_buf, unsigned int *phy_tx_len)
{
	char *b, *p;
	int i;

	b = pkt_tx_buf;
	p = phy_tx_buf;

	for (i = 0; i < pkt_tx_len; i++) {
		if (*b == PHY_IDLE || *b == PHY_ESC) {
			*p++ = PHY_ESC;
			*p++ = *b ^ 0x20;
		} else {
			*p++ = *b;
		}

		b++;
	}

	*phy_tx_len = p - phy_tx_buf;
}

static void phy_rx_parse(char *phy_rx_buf, unsigned int phy_rx_len,
			 char *pkt_rx_buf, unsigned int *pkt_rx_len)
{
	char *b, *p;

	b = phy_rx_buf;
	p = pkt_rx_buf;

	while (b < phy_rx_buf + phy_rx_len) {
		switch (*b) {
		case PHY_IDLE:
			b++;
			break;
		case PHY_ESC:
			b++;
			*p++ = *b++ ^ 0x20;
			break;
		default:
			*p++ = *b++;
			break;
		}
	}

	*pkt_rx_len = p - pkt_rx_buf;
}

/* Pad spare bytes of tx buf with PHY_IDLE, so that max10 just drop them */
static void phy_tx_pad(u8 bits_per_word, char *phy_buf,
		       unsigned int phy_buf_len, unsigned int *aligned_len)
{
	char *p = &phy_buf[phy_buf_len - 1], *dst_p;

	*aligned_len = ALIGN(phy_buf_len, bits_per_word / 8);

	if (*aligned_len == phy_buf_len)
		return;

	dst_p = &phy_buf[*aligned_len - 1];

	/* move EOP and bytes after EOP to the end of aligned size*/
	while (p > phy_buf) {
		*dst_p = *p;

		if (*p == PKT_EOP)
			break;

		p--;
		dst_p--;
	}

	/* fill the hole with PHY_IDLEs */
	while (p < dst_p)
		*p++ = PHY_IDLE;
}

static int phy_rx_length(u8 bits_per_word, bool is_read, unsigned int count)
{
	u32 BPW = bits_per_word / 8;
	u32 trans_batch_size;

	if (is_read) {
		trans_batch_size = VAL_SIZE;
	} else {
		trans_batch_size = WR_TRANS_RX_SIZE;
		count = 1;
	}

	/*
	 * We will abandon words filled with all PHY_IDLEs when rx, but
	 * PHY_IDLEs may still exsits in the first and last word of one
	 * response batch. The worst case is that BPW - 1 PHY_IDLEs are
	 * in the first word, and from last valid byte to aligned size there
	 * are also all PHY_IDLEs.
	 */
	return ALIGN(TRANS_SIZE_TO_PHY(trans_batch_size) + (BPW - 1), BPW) *
	       count;
}

static void reorder_phy_data(u8 bits_per_word, char *aligned_buf,
			     unsigned int aligned_buf_len)
{
	unsigned int count = aligned_buf_len / (bits_per_word / 8);
	u32 *p;

	if (bits_per_word == 32) {
		p = (u32 *)aligned_buf;
		while (count--) {
			*p = cpu_to_be32(*p);
			p++;
		}
	}
}

#define RX_NOT_READY_1	PHY_IDLE
#define RX_NOT_READY_4	(PHY_IDLE << 24 | PHY_IDLE << 16 |	\
			 PHY_IDLE << 8 | PHY_IDLE)

static bool is_rx_not_ready(u8 *rxbuf, u32 BPW)
{
	return (BPW == 1 && *rxbuf == RX_NOT_READY_1) ||
	       (BPW == 4 && *(u32 *)rxbuf == RX_NOT_READY_4);
}

static void *find_eop(u8 *rxbuf, u32 BPW)
{
	return memchr(rxbuf, PKT_EOP, BPW);
}

/*
 * We are uncertain about how many PHY_IDLEs will be received, and we
 * don't prepare buffer for all of them. So we have to read word one by
 * one to filter them.
 * When EOP is detected, 2 more bytes should be received(may be a PKT_ESC
 * before last valid byte) then rx should finish.
 */
static int do_spi_txrx_timeout(struct spi_device *spi, const void *txbuf,
			       u32 n_tx, void *rxbuf, u32 n_rx,
			       u32 *actual_n_rx, unsigned long timeout)
{
	u32 BPW = spi->bits_per_word / 8;
	unsigned long poll_timeout;
	bool eop_found = false;
	void *eop;
	u32 rx_cnt = 0;
	int status, add_byte;
	u8 *ptr;

	status = spi_write(spi, txbuf, n_tx);
	if (status)
		return status;

	poll_timeout = jiffies + timeout;

	while (rx_cnt < n_rx) {
		/* read word by word */
		status = spi_read(spi, rxbuf, BPW);
		if (status)
			return status;

		/* abandon word having no valid byte.
		 * We timeout when rx keeps invalid for some time since last
		 * valid word.
		 */
		if (!eop_found && is_rx_not_ready(rxbuf, BPW)) {
			if (!time_after(jiffies, poll_timeout))
				continue;

			/* It is possible we scheduled out for long time after
			 * spi_read. And when we scheduled in, SW timeout
			 * happens. But actually HW may be ready now. We need
			 * to do an extra read, if we got valid word we can
			 * continue to xfer, otherwise real HW issue happens.
			 */
			status = spi_read(spi, rxbuf, BPW);
			if (status)
				return status;

			if (is_rx_not_ready(rxbuf, BPW))
				return -ETIME;
		}

		/* update poll timeout when we get valid word */
		poll_timeout = jiffies + timeout;

		/* reorder word back now */
		if (BPW == 4)
			*(u32 *)rxbuf = cpu_to_be32(*(u32 *)rxbuf);

		rx_cnt += BPW;

		if (!eop_found) {
			/*
			 * EOP is found, we read 2 more bytes and exit.
			 * if BPW==32, the last 2 bytes may exsits in
			 * current word, in that case we can just exit.
			 * Or we need to read 1 more word.
			 * If we reached the end of n_rx, the last 2 bytes
			 * must exist in current word, otherwise let protocol
			 * parsing code to handle the err.
			 */
			eop = find_eop(rxbuf, BPW);
			if (eop) {
				if ((BPW + rxbuf - eop) > 2) {
					/*
					 * check if the last 2 bytes are already
					 * received in current word.
					 */
					break;
				} else if ((BPW + rxbuf - eop) == 2) {
					/*
					 * skip if last byte is not PHY_ESC or
					 * PKT_ESC. this is the valid end of a
					 * response too.
					 */
					ptr = eop + 1;

					if (*ptr != PHY_ESC && *ptr != PKT_ESC)
						break;

					/* read 1 more byte */
					add_byte = 1;
				} else {
					/* read 2 more bytes */
					add_byte = 2;
				}

				n_rx = min(n_rx, ALIGN(rx_cnt + add_byte, BPW));
				eop_found = true;
			}
		}

		rxbuf += BPW;
	}

	*actual_n_rx = rx_cnt;

	return 0;
}

static int do_reg_access(void *context, bool is_read, unsigned int reg,
			 unsigned int *value, u16 count)
{
	unsigned int rx_len, tx_len, *wr_val, *rd_val;
	struct spi_avmm_bridge *br = context;
	struct spi_device *spi = br->spi;
	int ret;

	mutex_lock(&br->lock);

	wr_val = is_read ? NULL : value;
	rd_val = is_read ? value : NULL;

	trans_tx_prepare(reg, is_read, wr_val, VAL_SIZE * count,
			 br->trans_tx_buf, &tx_len);
	packet_tx_prepare(br->trans_tx_buf, tx_len,
			  br->packet_tx_buf, &tx_len);
	phy_tx_prepare(br->packet_tx_buf, tx_len, br->phy_tx_buf, &tx_len);
	phy_tx_pad(spi->bits_per_word, br->phy_tx_buf, tx_len, &tx_len);

	/*
	 * Till now, phy layer data is filled byte by byte. But if SPI data
	 * width is 32 bits, 4 bytes will be shifted out at one time and Most
	 * Significant Byte should be transferred first. So we should reorder
	 * phy data into big endian before send.
	 */
	reorder_phy_data(spi->bits_per_word, br->phy_tx_buf, tx_len);

	rx_len = phy_rx_length(spi->bits_per_word, is_read, count);

	ret = do_spi_txrx_timeout(spi, br->phy_tx_buf, tx_len, br->phy_rx_buf,
				  rx_len, &rx_len, SPI_AVMM_XFER_TIMEOUT);
	if (ret) {
		dev_err(&spi->dev, "%s invalid packet %d\n", __func__, ret);
		print_hex_dump(KERN_ERR, "Read SPI:", DUMP_PREFIX_OFFSET,
				16, 1, br->phy_rx_buf, rx_len, true);
		goto exit;
	}

	phy_rx_parse(br->phy_rx_buf, rx_len, br->packet_rx_buf, &rx_len);

	ret = packet_rx_parse(br->packet_rx_buf, rx_len,
			      br->trans_rx_buf, &rx_len);
	if (ret) {
		dev_err(&spi->dev, "%s packet_rx_parse failed %d\n",
			__func__, ret);
		print_hex_dump(KERN_ERR, "packet rx:", DUMP_PREFIX_OFFSET,
				16, 1, br->trans_rx_buf, rx_len, true);
		goto exit;
	}

	ret = trans_rx_parse(br->trans_rx_buf, rx_len, is_read, count, rd_val);
	if (ret)
		dev_err(&spi->dev, "%s trans_rx_parse failed %d\n",
			__func__, ret);

exit:
	mutex_unlock(&br->lock);
	return ret;
}

#define do_reg_read(_ctx, _reg, _value, _count) \
	do_reg_access(_ctx, true, _reg, _value, _count)
#define do_reg_write(_ctx, _reg, _value, _count) \
	do_reg_access(_ctx, false, _reg, _value, _count)

/* check if reg is valid for register region defined by ctx */
static bool regmap_spi_avmm_check_region(struct regmap_spi_avmm_ctx *ctx,
					 unsigned int reg, u16 count)
{
	struct device *dev = &ctx->bridge->spi->dev;

	if ((reg + count * VAL_SIZE < reg) ||
	    (ctx->size && reg + count * VAL_SIZE > ctx->size)) {
		dev_dbg(dev, "%s out of register region: %x\n", __func__, reg);
		return false;
	}

	return true;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,16,0)
static int regmap_spi_avmm_reg_read(void *context, unsigned int reg,
				    unsigned int *val)
{
	struct regmap_spi_avmm_ctx *ctx = context;

	if (!regmap_spi_avmm_check_region(ctx, reg, 1))
		return -EINVAL;

	return do_reg_read(ctx->bridge, ctx->regoff + reg, val, 1);
}

static int regmap_spi_avmm_reg_write(void *context, unsigned int reg,
				     unsigned int val)
{
	struct regmap_spi_avmm_ctx *ctx = context;

	if (!regmap_spi_avmm_check_region(ctx, reg, 1))
		return -EINVAL;

	return do_reg_write(ctx->bridge, ctx->regoff + reg, &val, 1);
}
#endif

int regmap_spi_avmm_gather_write(void *context,
				 const void *reg_buf, size_t reg_len,
				 const void *val_buf, size_t val_len)
{
	struct regmap_spi_avmm_ctx *ctx = context;
	u32 reg = *(u32 *)reg_buf;
	u16 count = val_len / VAL_SIZE;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,3,0)
	int ret = 0;
	int i;
#endif

	if (reg_len != REG_SIZE || !val_len || !IS_ALIGNED(val_len, VAL_SIZE))
		return -EINVAL;

	if (!regmap_spi_avmm_check_region(ctx, reg, count))
		return -EINVAL;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,3,0)
	return do_reg_write(ctx->bridge, ctx->regoff + reg, (u32 *)val_buf,
			    count);
#else
	for (i = 0; i < count; i++) {
		ret = do_reg_write(ctx->bridge, ctx->regoff + reg,
				   (u32 *)val_buf, MAX_TX_CNT);
		if (ret)
			break;
		
		reg += MAX_TX_CNT * REG_SIZE;
		val_buf += MAX_TX_CNT * VAL_SIZE;
	}

	return ret;
#endif
}

int regmap_spi_avmm_write(void *context, const void *data, size_t count)
{
	if (count < REG_SIZE + VAL_SIZE)
		return -EINVAL;

	return regmap_spi_avmm_gather_write(context, data, REG_SIZE,
					    data + REG_SIZE, count - REG_SIZE);
}

int regmap_spi_avmm_read(void *context,
			 const void *reg_buf, size_t reg_len,
			 void *val_buf, size_t val_len)
{
	struct regmap_spi_avmm_ctx *ctx = context;
	u32 reg = *(u32 *)reg_buf;
	u16 count = val_len / VAL_SIZE;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,3,0)
	u16 rd_cnt = 0;
	int ret = 0;
#endif

	if (reg_len != REG_SIZE || !val_len || !IS_ALIGNED(val_len, VAL_SIZE))
		return -EINVAL;

	if (!regmap_spi_avmm_check_region(ctx, reg, count))
		return -EINVAL;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,3,0)
	return do_reg_read(ctx->bridge, ctx->regoff + reg, val_buf, count);
#else
	while (count > 0) {
		if (count > MAX_RX_CNT)
			rd_cnt = MAX_RX_CNT;
		else
			rd_cnt = count;

		ret = do_reg_read(ctx->bridge, ctx->regoff + reg,
				  val_buf, rd_cnt);

		if (ret)
			break;

		count -= rd_cnt;
		reg += rd_cnt * REG_SIZE;
		val_buf += rd_cnt * VAL_SIZE;
	}

	return ret;
#endif
}

struct regmap_bus regmap_spi_avmm_bus = {
	.write = regmap_spi_avmm_write,
	.gather_write = regmap_spi_avmm_gather_write,
	.read = regmap_spi_avmm_read,
	.reg_format_endian_default = REGMAP_ENDIAN_NATIVE,
	.val_format_endian_default = REGMAP_ENDIAN_NATIVE,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,3,0)
	.max_raw_read = VAL_SIZE * MAX_RX_CNT,
	.max_raw_write = VAL_SIZE * MAX_TX_CNT,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,16,0)
	.reg_write = regmap_spi_avmm_reg_write,
	.reg_read = regmap_spi_avmm_reg_read,
#endif
};

struct spi_avmm_bridge *spi_avmm_bridge_alloc_and_init(struct spi_device *spi)
{
	struct spi_avmm_bridge *br;

	br = kzalloc(sizeof(*br), GFP_KERNEL);
	if (!br)
		return NULL;

	br->spi = spi;
	mutex_init(&br->lock);

	/* pre-allocated buffers used in translation */
	br->trans_tx_buf = kzalloc(TRANS_TX_SIZE, GFP_KERNEL);
	br->trans_rx_buf = kzalloc(TRANS_RX_SIZE, GFP_KERNEL);
	br->packet_tx_buf = kzalloc(PACKET_TX_SIZE, GFP_KERNEL);
	br->packet_rx_buf = kzalloc(PACKET_RX_SIZE, GFP_KERNEL);
	br->phy_tx_buf = kzalloc(PHY_TX_SIZE, GFP_KERNEL);
	br->phy_rx_buf = kzalloc(PHY_RX_SIZE, GFP_KERNEL);

	if (br->trans_tx_buf && br->trans_rx_buf && br->packet_tx_buf &&
	    br->packet_rx_buf && br->phy_tx_buf && br->phy_rx_buf)
		return br;

	spi_avmm_bridge_destroy(br);
	return NULL;
}

void spi_avmm_bridge_destroy(struct spi_avmm_bridge *bridge)
{
	kfree(bridge->trans_tx_buf);
	kfree(bridge->trans_rx_buf);
	kfree(bridge->packet_tx_buf);
	kfree(bridge->packet_rx_buf);
	kfree(bridge->phy_tx_buf);
	kfree(bridge->phy_rx_buf);
	mutex_destroy(&bridge->lock);
	kfree(bridge);
}

struct regmap_spi_avmm_ctx *
regmap_spi_avmm_gen_ctx(struct spi_avmm_bridge *bridge, unsigned int regoff,
			size_t size)
{
	struct device *dev = &bridge->spi->dev;
	struct regmap_spi_avmm_ctx *ctx;

	/*
	 * size must be valid, unless it is a global regmap ctx (size = 0 &&
	 * regoff = 0)
	 */
	if (regoff && !size)
		return ERR_PTR(-EINVAL);

	/* check overflow */
	if (size + regoff < size)
		return ERR_PTR(-EINVAL);

	ctx = devm_kzalloc(dev, sizeof(*ctx),  GFP_KERNEL);
	if (ctx) {
		ctx->bridge = bridge;
		ctx->regoff = regoff;
		ctx->size = size;
	}

	return ctx;
}
