// SPDX-License-Identifier: GPL-2.0
/*
 * Driver for Intel On Chip Flash IP
 *
 * https://www.intel.com/content/dam/www/programmable/us/en/pdfs/
 * literature/hb/max-10/ug_m10_ufm.pdf
 *
 *
 * Copyright (C) 2019 Intel Corporation. All rights reserved.
 *
 */

#include <linux/bitfield-intel-pac.h>
#include <linux/iopoll-4-12.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/intel-on-chip-flash.h>
#include <linux/of_device.h>
#include <linux/version.h>

#define IOCF_VAL_SIZE			4
#define IOCF_REG_SIZE			4

#define IOCF_STAT_OFF			(0x00 << 2)
#define IOCF_CTRL_OFF			(0x01 << 2)

#define IOCF_STAT_BUSY			GENMASK(1, 0)
#define IOCF_STAT_BUSY_IDLE		0x00
#define IOCF_STAT_BUSY_BUSY_ERASE	0x01
#define IOCF_STAT_BUSY_BUSY_WRITE	0x02
#define IOCF_STAT_BUSY_BUSY_READ	0x03

#define IOCF_STAT_RS			BIT(2)
#define IOCF_STAT_WS			BIT(3)
#define IOCF_STAT_ES			BIT(4)
#define IOCF_STAT_SP1			BIT(5)
#define IOCF_STAT_SP2			BIT(6)
#define IOCF_STAT_SP3			BIT(7)
#define IOCF_STAT_SP4			BIT(8)
#define IOCF_STAT_SP5			BIT(9)

#define IOCF_CTRL_PAGE_ER_ADDR		GENMASK(19, 0)
#define IOCF_CTRL_SECT_ER_ADDR		GENMASK(22, 20)
#define IOCF_CTRL_WP1			BIT(23)
#define IOCF_CTRL_WP2			BIT(24)
#define IOCF_CTRL_WP3			BIT(25)
#define IOCF_CTRL_WP4			BIT(26)
#define IOCF_CTRL_WP5			BIT(27)

#define IOCF_POLL_INTERVAL_US		1
#define IOCF_POLL_TIMEOUT_US		350000

static const struct regmap_config iocf_csr_regmap_config = {
	.name = "csr",
	.reg_bits = IOCF_REG_SIZE * 8,
	.reg_stride = 4,
	.val_bits = IOCF_VAL_SIZE * 8,
	.cache_type = REGCACHE_NONE,
};

static const struct regmap_config iocf_data_regmap_config = {
	.name = "data",
	.reg_bits = IOCF_REG_SIZE * 8,
	.reg_stride = 4,
	.val_bits = IOCF_VAL_SIZE * 8,
	.cache_type = REGCACHE_NONE,
};

static struct mtd_erase_region_info eraseregions[] = {
	{
		.offset = 0,
		.erasesize = 0x8000,
		.numblocks = 1,
	},
	{
		.offset = 0x8000,
		.erasesize = 0x8000,
		.numblocks = 1,
	},
	{
		.offset = 0x10000,
		.erasesize = 0x60000,
		.numblocks = 1,
	},
	{
		.offset = 0x70000,
		.erasesize = 0x48000,
		.numblocks = 1,
	},
	{
		.offset = 0xb8000,
		.erasesize = 0xa8000,
		.numblocks = 1,
	},
};

struct iocf_ctrl {
	struct regmap *csr_regmap;
	struct regmap *data_regmap;
	struct device *dev;
	struct mtd_info mtd;
};

static int iocf_wr_ctrl_idle(struct iocf_ctrl *ctrl, u32 ctrl_reg)
{
	u32 val;
	int ret;

	dev_dbg(ctrl->dev, "%s ctrl 0x%x\n", __func__, ctrl_reg);

	ret = regmap_write(ctrl->csr_regmap, IOCF_CTRL_OFF, ctrl_reg);

	if (ret) {
		dev_err(ctrl->dev, "%s failed to set CTRL %d\n",
			__func__, ret);
		return ret;
	}

	ret = regmap_read_poll_timeout(ctrl->csr_regmap, IOCF_STAT_OFF, val,
				       (IOCF_STAT_BUSY_IDLE ==
				       FIELD_GET(IOCF_STAT_BUSY, val)),
				       IOCF_POLL_INTERVAL_US,
				       IOCF_POLL_TIMEOUT_US);

	if (ret)
		dev_err(ctrl->dev, "%s failed to see IDLE after ctrl 0x%x\n",
			__func__, val);

	return ret;
}

static int iocf_disable_wp(struct iocf_ctrl *ctrl, int index,
			   int disable, u32 *preg)
{
	u32 val, ctrl_reg;
	int ret;

	ret = regmap_read_poll_timeout(ctrl->csr_regmap, IOCF_STAT_OFF, val,
				       (IOCF_STAT_BUSY_IDLE ==
				       FIELD_GET(IOCF_STAT_BUSY, val)),
				       IOCF_POLL_INTERVAL_US,
				       IOCF_POLL_TIMEOUT_US);

	if (ret) {
		dev_err(ctrl->dev, "%s failed to see IDLE before ctrl 0x%x\n",
			__func__, val);
		return ret;
	}

	ret = regmap_read(ctrl->csr_regmap, IOCF_CTRL_OFF, &val);

	if (ret) {
		dev_err(ctrl->dev, "%s failed to read CTRL %d\n",
			__func__, ret);
		return ret;
	}

	if (disable)
		ctrl_reg = val & ~BIT(index + 23);
	else
		ctrl_reg = val | BIT(index + 23) | IOCF_CTRL_SECT_ER_ADDR;

	ret = iocf_wr_ctrl_idle(ctrl, ctrl_reg);

	*preg = ctrl_reg;

	return ret;
}

static int iocf_erase(struct mtd_info *mtd, struct erase_info *instr)
{
	struct mtd_erase_region_info *er = eraseregions;
	struct iocf_ctrl *ctrl = mtd->priv;
	u32 ctrl_reg;
	int i, ret;

	dev_dbg(ctrl->dev, "%s erase 0x%llx bytes from 0x%llx\n",
		__func__, instr->len, instr->addr);

	ret = -ENODEV;

	for (i = 0; i < ARRAY_SIZE(eraseregions); i++, er++) {
		if ((er->offset == instr->addr) &&
		    (er->erasesize == instr->len))
			break;
	}

	if (i >= ARRAY_SIZE(eraseregions)) {
		dev_err(ctrl->dev, "%s no sector found\n", __func__);
		ret = -EINVAL;
		goto exit;
	}

	ret = iocf_disable_wp(ctrl, i, 1, &ctrl_reg);

	if (ret) {
		dev_err(ctrl->dev, "%s failed to disable write protect on %d",
			__func__, i);
		goto exit;
	}

	ctrl_reg &= ~IOCF_CTRL_SECT_ER_ADDR;
	ctrl_reg |= FIELD_PREP(IOCF_CTRL_SECT_ER_ADDR, (i + 1));

	ret = iocf_wr_ctrl_idle(ctrl, ctrl_reg);

	if (ret)
		goto exit;

	ret = iocf_disable_wp(ctrl, i, 0, &ctrl_reg);

	if (ret) {
		dev_err(ctrl->dev, "%s failed to disable write protect on %d",
			__func__, i);
		goto exit;
	}
exit:
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,17,0)
	instr->state = MTD_ERASE_DONE;
	mtd_erase_callback(instr);
#endif
	return ret;
}

static int iocf_read(struct mtd_info *mtd, loff_t from, size_t len,
		     size_t *retlen, u_char *buf)
{
	struct iocf_ctrl *ctrl = mtd->priv;
	int ret;
	u32 val;

	if (!IS_ALIGNED(from, IOCF_VAL_SIZE) ||
	    !IS_ALIGNED(len, IOCF_VAL_SIZE)) {
		dev_err(ctrl->dev,
			"%s non aligned read from 0x%llx len 0x%zx\n",
			__func__, from, len);
		return -EINVAL;
	}

	dev_dbg(ctrl->dev, "%s read 0x%zx bytes from 0x%llx\n",
		__func__, len, from);

	ret = regmap_read_poll_timeout(ctrl->csr_regmap, IOCF_STAT_OFF, val,
				       (IOCF_STAT_BUSY_IDLE ==
				       FIELD_GET(IOCF_STAT_BUSY, val)),
				       IOCF_POLL_INTERVAL_US,
				       IOCF_POLL_TIMEOUT_US);

	if (ret) {
		dev_err(ctrl->dev, "%s regmap_read_poll_timeout failed %d\n",
			__func__, ret);
		return ret;
	}

	ret = regmap_bulk_read(ctrl->data_regmap, from, buf,
			       len / IOCF_VAL_SIZE);
	if (ret) {
		dev_err(ctrl->dev, "%s regmap_read failed %d\n", __func__, ret);
		return ret;
	}

	return *retlen = len;
}

static int iocf_write(struct mtd_info *mtd, loff_t to, size_t len,
		      size_t *retlen, const u_char *buf)
{
	struct mtd_erase_region_info *er = eraseregions;
	struct iocf_ctrl *ctrl = mtd->priv;
	u32 ctrl_reg;
	int i, ret;
	size_t j;

	if (!IS_ALIGNED(to, IOCF_VAL_SIZE) ||
	    !IS_ALIGNED(len, IOCF_VAL_SIZE)) {
		dev_err(ctrl->dev,
			"%s non aligned write to 0x%llx len 0x%zx\n",
			__func__, to, len);
		return -EINVAL;
	}

	ret = -ENODEV;

	for (i = 0; i < ARRAY_SIZE(eraseregions); i++, er++) {
		dev_dbg(ctrl->dev,
			"%s write 0x%zx bytes to 0x%llx 0x%llx 0x%x\n",
			__func__, len, to, er->offset, er->erasesize);
		if ((to >= er->offset) && (to < (er->offset + er->erasesize)))
			break;
	}

	if (i >= ARRAY_SIZE(eraseregions)) {
		dev_err(ctrl->dev, "%s no sector found\n", __func__);
		return -EINVAL;
	}

	dev_dbg(ctrl->dev, "%s write 0x%zx bytes to 0x%llx\n",
		__func__, len, to);

	for (j = 0; j < len; j += IOCF_VAL_SIZE) {
		ret = iocf_disable_wp(ctrl, i, 1, &ctrl_reg);

		if (ret) {
			dev_err(ctrl->dev,
				"%s failed to disable write protect on %d",
				__func__, i);
			return ret;
		}

		ret = regmap_bulk_write(ctrl->data_regmap, to, buf, 1);

		if (ret) {
			dev_err(ctrl->dev, "%s regmap_write failed %d\n",
				__func__, ret);
			return ret;
		}

		ret = iocf_disable_wp(ctrl, i, 0, &ctrl_reg);

		if (ret) {
			dev_err(ctrl->dev,
				"%s failed to disable write protect on %d",
				__func__, i);
			return ret;
		}

		to += IOCF_VAL_SIZE;
		buf += IOCF_VAL_SIZE;
	}

	*retlen = len;

	return 0;
}

static int iocf_probe(struct platform_device *pdev)
{
	struct intel_on_chip_flash_platform_data *pdata;
	struct device *dev = &pdev->dev;
	struct iocf_ctrl *ctrl;
	struct mtd_info *mtd;
	int ret = 0;
	u32 val;

	dev_dbg(&pdev->dev, "%s\n", __func__);

	pdata = dev_get_platdata(dev);

	ctrl = devm_kzalloc(dev, sizeof(*ctrl), GFP_KERNEL);

	if (!ctrl)
		return -ENOMEM;

	ctrl->csr_regmap = devm_regmap_init(dev, &pdata->regmap_bus,
					    pdata->regmap_ctx,
					    &iocf_csr_regmap_config);
	if (IS_ERR(ctrl->csr_regmap)) {
		dev_err(dev, "failed to init iocf regmap: %ld\n",
			PTR_ERR(ctrl->csr_regmap));
		return PTR_ERR(ctrl->csr_regmap);
	}

	ctrl->data_regmap = devm_regmap_init(dev, &pdata->regmap_bus,
					    pdata->data_regmap_ctx,
					    &iocf_data_regmap_config);

	if (IS_ERR(ctrl->data_regmap)) {
		dev_err(dev, "failed to init iocf data_regmap: %ld\n",
			PTR_ERR(ctrl->data_regmap));
		return PTR_ERR(ctrl->data_regmap);
	}

	ctrl->dev = dev;

	dev_set_drvdata(dev, ctrl);

	ret = regmap_read(ctrl->csr_regmap, IOCF_STAT_OFF, &val);
	if (ret)
		dev_err(dev, "regmap read failed %d\n", ret);
	else
		dev_dbg(dev, "IOCF_STAT_OFF 0x%08x\n", val);

	ret = regmap_read(ctrl->csr_regmap, IOCF_CTRL_OFF, &val);
	if (ret)
		dev_err(dev, "regmap read failed %d\n", ret);
	else
		dev_dbg(dev, "IOCF_CTRL_OFF 0x%08x\n", val);

	mtd = &ctrl->mtd;

	if (!mtd->name)
		mtd->name = dev_name(dev);

	mtd->priv = ctrl;
	mtd->type = MTD_NORFLASH;
	mtd->writesize = IOCF_VAL_SIZE;
	mtd->flags = MTD_CAP_NORFLASH;
	mtd->size = mtd->erasesize = 0x160000;
	mtd->_erase = iocf_erase;
	mtd->_read = iocf_read;
	mtd->_write = iocf_write;
	mtd->eraseregions = eraseregions;
	mtd->numeraseregions = ARRAY_SIZE(eraseregions);
	mtd->dev.parent = dev;

	ret = mtd_device_register(mtd, NULL, 0);

	if (ret)
		dev_err(dev, "%s mtd_device_register failed %d\n",
			__func__, ret);

	return ret;
}

static int iocf_remove(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct iocf_ctrl *ctrl;
	int ret;

	dev_dbg(&pdev->dev, "%s\n", __func__);

	ctrl = dev_get_drvdata(dev);

	ret = mtd_device_unregister(&ctrl->mtd);

	if (ret)
		dev_err(dev, "%s mtd_device_unregister failed %d\n",
			__func__, ret);

	return ret;
}

static struct platform_driver iocf_driver = {
	.driver = {
		.name = INTEL_ON_CHIP_FLASH_DRV_NAME,
	},
	.probe = iocf_probe,
	.remove = iocf_remove,
};
module_platform_driver(iocf_driver);

MODULE_AUTHOR("Matthew Gerlach <matthew.gerlach@linux.intel.com>");
MODULE_DESCRIPTION("Intel On Chip Flash IP");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS("platform:" INTEL_ON_CHIP_FLASH_DRV_NAME);
