// SPDX-License-Identifier: GPL-2.0
/*
 * Driver for Generic Serial Flash Interface Intel FPGA IP
 *
 * https://www.intel.com/content/www/us/en/programmable/documentation/
 * tkt1520669802999.html
 *
 * Copyright (C) 2018 Intel Corporation. All rights reserved.
 *
 */

#include <linux/bitfield-intel-pac.h>
#include <linux/iopoll-4-12.h>
#include <linux/module.h>
#include <linux/mtd/intel-generic-qspi.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/spi-nor-4-12.h>
#include <linux/of_device.h>

#define IGSF_MAX_FLASH			3
#define IGSF_VAL_SIZE			4
#define IGSF_REG_SIZE			4

#define IGSF_CTRL_REG_OFF		(0x0 << 2)
#define IGSF_BAUD_RATE_REG_OFF		(0x1 << 2)
#define IGSF_CS_DELAY_REG_OFF		(0x2 << 2)
#define IGSF_READ_CAP_REG_OFF		(0x3 << 2)
#define IGSF_OP_SET_REG_OFF		(0x4 << 2)
#define IGSF_RD_INSTR_REG_OFF		(0x5 << 2)
#define IGSF_WR_INSTR_REG_OFF		(0x6 << 2)
#define IGSF_CMD_SETTING_REG_OFF	(0x7 << 2)
#define IGSF_CMD_CTRL_REG_OFF		(0x8 << 2)
#define IGSF_CMD_ADDR_REG_OFF		(0x9 << 2)
#define IGSF_CMD_WR_LO_REG_OFF		(0xa << 2)
#define IGSF_CMD_WR_HI_REG_OFF		(0xb << 2)
#define IGSF_CMD_RD_LO_REG_OFF		(0xc << 2)
#define IGSF_CMD_RD_HI_REG_OFF		(0xd << 2)

#define IGSF_CTRL_ENABLE		BIT(0)
#define IGSF_CTRL_4_BYTE_ADDR		BIT(8)

#define IGSF_CMD_SET_DUMMY_CYCLES	GENMASK(20, 16)
#define IGSF_CMD_SET_DATA_BYTES		GENMASK(15, 12)
#define IGSF_CMD_SET_TYPE		BIT(11)
#define IGSF_CMD_SET_TYPE_RD		1
#define IGSF_CMD_SET_ADDR_BYTES		GENMASK(10, 8)
#define IGSF_CMD_SET_OPCODE		GENMASK(7, 0)

#define IGSF_OP_SET_RD_DATA_MODE       GENMASK(17, 16)
#define IGSF_OP_SET_RD_ADDR_MODE       GENMASK(13, 12)
#define IGSF_OP_SET_WR_DATA_MODE       GENMASK(9, 8)
#define IGSF_OP_SET_WR_ADDR_MODE       GENMASK(5, 4)
#define IGSF_OP_SET_INS_MODE           GENMASK(1, 0)

#define IGSF_RD_INSTR_OPCODE		GENMASK(7, 0)
#define IGSF_RD_INSTR_DUMMY_CYCLES	GENMASK(12, 8)

#define IGSF_WR_INSTR_POLL_OPCODE	GENMASK(15, 8)
#define IGSF_WR_INSTR_WR_OPCODE		GENMASK(7, 0)

#define IGSF_CMD_START			BIT(0)

#define IGSF_MAX_DATA_BYTES		(2 * IGSF_VAL_SIZE)

#define IGSF_POLL_INTERVAL_US		10
#define IGSF_POLL_TIMEOUT_US		1000

static const struct regmap_config igsf_csr_regmap_config = {
	.name = "csr",
	.reg_bits = IGSF_REG_SIZE * 8,
	.reg_stride = 4,
	.val_bits = IGSF_VAL_SIZE * 8,
	.cache_type = REGCACHE_NONE,
};

static const struct regmap_config igsf_data_regmap_config = {
	.name = "data",
	.reg_bits = IGSF_REG_SIZE * 8,
	.reg_stride = 4,
	.val_bits = IGSF_VAL_SIZE * 8,
	.cache_type = REGCACHE_NONE,
};

struct igsf_flash {
	struct spi_nor nor;
	struct igsf_ctrl *ctrl;
};

struct igsf_ctrl {
	struct regmap *csr_regmap;
	struct regmap *data_regmap;
	struct device *dev;
	struct igsf_flash flash;
	/* only one flash at a time can be accessed */
	struct mutex bus_mutex;
};

static int igsf_prep(struct spi_nor *nor, enum spi_nor_ops ops)
{
	struct igsf_flash *flash = nor->priv;
	struct igsf_ctrl *ctrl = flash->ctrl;

	mutex_lock(&ctrl->bus_mutex);

	return 0;
}

static void igsf_unprep(struct spi_nor *nor, enum spi_nor_ops ops)
{
	struct igsf_flash *flash = nor->priv;
	struct igsf_ctrl *ctrl = flash->ctrl;

	mutex_unlock(&ctrl->bus_mutex);
}

static int igsf_read_data(struct igsf_flash *flash, u32 cmd, u8 *buf, int len)
{
	struct igsf_ctrl *ctrl = flash->ctrl;
	u32 reg;
	int ret;

	if (len > IGSF_MAX_DATA_BYTES) {
		dev_err(ctrl->dev, "%s bad len %d > %d\n",
			__func__, len, IGSF_MAX_DATA_BYTES);
		return -EINVAL;
	}

	dev_dbg(ctrl->dev, "%s op 0x%08x len %d\n", __func__, cmd, len);

	ret = regmap_write(ctrl->csr_regmap, IGSF_CMD_SETTING_REG_OFF, cmd);

	if (ret) {
		dev_err(ctrl->dev, "%s regmap_write to 0x%x failed %d\n",
			__func__, IGSF_CMD_SETTING_REG_OFF, ret);
		return ret;
	}

	ret = regmap_write(ctrl->csr_regmap, IGSF_CMD_CTRL_REG_OFF,
			   IGSF_CMD_START);

	if (ret) {
		dev_err(ctrl->dev, "%s regmap_write to 0x%x failed %d\n",
			__func__, IGSF_CMD_CTRL_REG_OFF, ret);
		return ret;
	}

	ret = regmap_read_poll_timeout(ctrl->csr_regmap, IGSF_CMD_CTRL_REG_OFF,
				       reg, (0 == (reg & IGSF_CMD_START)),
				       IGSF_POLL_INTERVAL_US,
				       IGSF_POLL_TIMEOUT_US);

	if (ret) {
		dev_err(ctrl->dev, "%s regmap_read_poll_timeout failed %d\n",
			__func__, ret);
		return ret;
	}

	ret = regmap_read(ctrl->csr_regmap, IGSF_CMD_RD_LO_REG_OFF, &reg);

	if (ret) {
		dev_err(ctrl->dev, "%s regmap_read to 0x%x failed %d\n",
			__func__, IGSF_CMD_RD_HI_REG_OFF, ret);
		return ret;
	}

	memcpy(buf, &reg, min_t(int, len, IGSF_VAL_SIZE));

	if (len > IGSF_VAL_SIZE) {
		ret = regmap_read(ctrl->csr_regmap,
				  IGSF_CMD_RD_HI_REG_OFF, &reg);

		if (ret) {
			dev_err(ctrl->dev, "%s regmap_read to 0x%x failed %d\n",
				__func__, IGSF_CMD_RD_HI_REG_OFF, ret);
			return ret;
		}
		memcpy(buf + IGSF_VAL_SIZE, &reg, len - IGSF_VAL_SIZE);
	}

	return 0;
}

static int igsf_read_reg(struct spi_nor *nor, u8 opcode, u8 *val, int len)
{
	struct igsf_flash *flash = nor->priv;
	struct igsf_ctrl *ctrl = flash->ctrl;
	u32 reg;

	dev_dbg(ctrl->dev, "%s op 0x%02x len %d\n", __func__, opcode, len);

	reg = FIELD_PREP(IGSF_CMD_SET_DATA_BYTES, len) |
		FIELD_PREP(IGSF_CMD_SET_TYPE, IGSF_CMD_SET_TYPE_RD) |
		FIELD_PREP(IGSF_CMD_SET_OPCODE, opcode);

	return igsf_read_data(flash, reg, val, len);
}

static int igsf_write_data(struct igsf_flash *flash, u32 cmd,
			   const u8 *buf, int len)
{
	struct igsf_ctrl *ctrl = flash->ctrl;
	u32 reg;
	int ret;

	if (len > IGSF_MAX_DATA_BYTES) {
		dev_err(ctrl->dev, "%s bad len %d > %d\n",
			__func__, len, IGSF_MAX_DATA_BYTES);
		return -EINVAL;
	}

	dev_dbg(ctrl->dev, "%s cmd 0x%x len %d\n", __func__, cmd, len);

	ret = regmap_write(ctrl->csr_regmap, IGSF_CMD_SETTING_REG_OFF, cmd);

	if (ret) {
		dev_err(ctrl->dev, "%s regmap_write to 0x%x failed %d\n",
			__func__, IGSF_CMD_SETTING_REG_OFF, ret);
		return ret;
	}

	reg = 0;

	memcpy(&reg, buf, min_t(int, len, IGSF_VAL_SIZE));

	ret = regmap_write(ctrl->csr_regmap, IGSF_CMD_WR_LO_REG_OFF, reg);

	if (ret) {
		dev_err(ctrl->dev, "%s regmap_write to 0x%x failed %d\n",
			__func__, IGSF_CMD_WR_LO_REG_OFF, ret);
		return ret;
	}

	if (len > IGSF_VAL_SIZE) {
		reg = 0;
		memcpy(&reg, buf + IGSF_VAL_SIZE, len - IGSF_VAL_SIZE);
		ret = regmap_write(ctrl->csr_regmap,
				   IGSF_CMD_WR_HI_REG_OFF, reg);

		if (ret) {
			dev_err(ctrl->dev,
				"%s regmap_write to 0x%x failed %d\n",
				__func__, IGSF_CMD_WR_HI_REG_OFF, ret);
			return ret;
		}
	}
	ret = regmap_write(ctrl->csr_regmap, IGSF_CMD_CTRL_REG_OFF,
			   IGSF_CMD_START);

	if (ret) {
		dev_err(ctrl->dev, "%s regmap_write to 0x%x failed %d\n",
			__func__, IGSF_CMD_CTRL_REG_OFF, ret);
		return ret;
	}

	ret = regmap_read_poll_timeout(ctrl->csr_regmap, IGSF_CMD_CTRL_REG_OFF,
				       reg, (0 == (reg & IGSF_CMD_START)),
				       IGSF_POLL_INTERVAL_US,
				       IGSF_POLL_TIMEOUT_US);

	if (ret) {
		dev_err(ctrl->dev, "%s regmap_read_poll_timeout failed %d\n",
			__func__, ret);
		return ret;
	}
	return ret;
}

static int igsf_write_reg(struct spi_nor *nor, u8 opcode, u8 *val, int len)
{
	struct igsf_flash *flash = nor->priv;
	struct igsf_ctrl *ctrl = flash->ctrl;
	u32 reg;

	dev_dbg(ctrl->dev, "%s op 0x%02x len %d\n", __func__, opcode, len);

	reg = FIELD_PREP(IGSF_CMD_SET_DATA_BYTES, len) |
		FIELD_PREP(IGSF_CMD_SET_OPCODE, opcode);

	return igsf_write_data(flash, reg, (const u8 *)val, len);
}

static ssize_t igsf_read(struct spi_nor *nor, loff_t from, size_t len,
			 u_char *buf)
{
	struct igsf_flash *flash = nor->priv;
	struct igsf_ctrl *ctrl = flash->ctrl;
	int ret;

	if (!IS_ALIGNED(from, IGSF_VAL_SIZE) ||
	    !IS_ALIGNED(len, IGSF_VAL_SIZE)) {
		dev_err(ctrl->dev, "%s non aligned read from 0x%llx len 0x%zx\n",
			__func__, from, len);
		return -EINVAL;
	}

	dev_dbg(ctrl->dev, "%s len %zu from 0x%llx\n", __func__,
		len, from);

	ret = regmap_bulk_read(ctrl->data_regmap, from, buf,
			       len / IGSF_VAL_SIZE);

	if (ret) {
		dev_err(ctrl->dev, "%s regmap_read failed %d\n", __func__, ret);
		return ret;
	}

	return len;
}

static ssize_t igsf_write(struct spi_nor *nor, loff_t to, size_t len,
			  const u_char *buf)
{
	struct igsf_flash *flash = nor->priv;
	struct igsf_ctrl *ctrl = flash->ctrl;
	int ret;

	if (!IS_ALIGNED(to, IGSF_VAL_SIZE) ||
	    !IS_ALIGNED(len, IGSF_VAL_SIZE)) {
		dev_err(ctrl->dev, "%s non aligned write to 0x%llx len 0x%zx\n",
			__func__, to, len);
		return -EINVAL;
	}

	dev_dbg(ctrl->dev, "%s len %zu to 0x%llx\n", __func__,
		len, to);

	ret = regmap_bulk_write(ctrl->data_regmap, to, buf,
				len / IGSF_VAL_SIZE);

	if (ret) {
		dev_err(ctrl->dev, "%s regmap_write failed %d\n",
			__func__, ret);
		return ret;
	}

	return len;
}

struct xfer_mode_info {
	u8 opcode;
	u8 addr_mode;
	u8 data_mode;
};

static struct xfer_mode_info rd_mode[] = {
	{SPINOR_OP_READ,		0, 0},
	{SPINOR_OP_READ_4B,		0, 0},
	{SPINOR_OP_READ_FAST,		0, 0},
	{SPINOR_OP_READ_FAST_4B,	0, 0},
	{SPINOR_OP_READ_1_1_2,		0, 1},
	{SPINOR_OP_READ_1_1_2_4B,	0, 1},
	{SPINOR_OP_READ_1_2_2,		1, 1},
	{SPINOR_OP_READ_1_2_2_4B,	1, 1},
	{SPINOR_OP_READ_1_1_4,		0, 2},
	{SPINOR_OP_READ_1_1_4_4B,	0, 2},
	{SPINOR_OP_READ_1_4_4,		2, 2},
	{SPINOR_OP_READ_1_4_4_4B,	2, 2},
};

static struct xfer_mode_info wr_mode[] = {
	{SPINOR_OP_PP,		0, 0},
	{SPINOR_OP_PP_4B,	0, 0},
	{SPINOR_OP_PP_1_1_4,	0, 2},
	{SPINOR_OP_PP_1_1_4_4B,	0, 2},
	{SPINOR_OP_PP_1_4_4,	2, 2},
	{SPINOR_OP_PP_1_4_4_4B,	2, 2},
};

static int igsf_get_read_mode(u8 opcode, u8 *addr_mode, u8 *data_mode)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(rd_mode); i++) {
		if (rd_mode[i].opcode == opcode) {
			*addr_mode = rd_mode[i].addr_mode;
			*data_mode = rd_mode[i].data_mode;

			return 0;
		}
	}

	return -EINVAL;
}

static int igsf_get_write_mode(u8 opcode, u8 *addr_mode, u8 *data_mode)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(wr_mode); i++) {
		if (wr_mode[i].opcode == opcode) {
			*addr_mode = wr_mode[i].addr_mode;
			*data_mode = wr_mode[i].data_mode;

			return 0;
		}
	}

	return -EINVAL;
}

static int igsf_setup_flash(struct igsf_ctrl *ctrl)
{
	u8 rd_addr_mode, rd_data_mode, wr_addr_mode, wr_data_mode, poll_opcode;
	struct igsf_flash *flash = &ctrl->flash;
	struct spi_nor *nor;
	int ret;
	u32 val;

	flash->ctrl = ctrl;

	ret = regmap_write(ctrl->csr_regmap, IGSF_BAUD_RATE_REG_OFF, 0x1);
	if (ret) {
		dev_err(ctrl->dev,
			"%s failed write to IGSF_BAUD_RATE_REG_OFF %d\n",
			__func__, ret);
		return ret;
	}

	nor = &flash->nor;
	nor->dev = ctrl->dev;
	nor->priv = flash;
	nor->mtd.priv = nor;

	nor->prepare = igsf_prep;
	nor->unprepare = igsf_unprep;

	nor->read_reg = igsf_read_reg;
	nor->write_reg = igsf_write_reg;

	nor->read = igsf_read;
	nor->write = igsf_write;

	ret = spi_nor_scan(nor, NULL, SPI_NOR_QUAD);

	if (ret) {
		dev_err(nor->dev, "flash not found\n");
		return ret;
	}

	dev_dbg(nor->dev,
		"%s rd_op 0x%02x wr_op 0x%02x dummy %d addr %d flags 0x%02x\n",
		__func__, nor->read_opcode, nor->program_opcode,
		nor->read_dummy, nor->addr_width, nor->flags);

	val = IGSF_CTRL_ENABLE;
	if (nor->addr_width == 4)
		val |= IGSF_CTRL_4_BYTE_ADDR;
	ret = regmap_write(ctrl->csr_regmap, IGSF_CTRL_REG_OFF, val);
	if (ret) {
		dev_err(nor->dev,
			"%s failed write to IGSF_CTRL_REG_OFF %d\n",
			__func__, ret);
		return ret;
	}
	dev_dbg(nor->dev, "set CTRL[0x%x] = 0x%x\n", IGSF_CTRL_REG_OFF, val);

	ret = igsf_get_read_mode(nor->read_opcode, &rd_addr_mode,
				 &rd_data_mode);
	if (ret) {
		dev_err(nor->dev, "read opcode 0x%x not supported\n",
			nor->read_opcode);
		return ret;
	}

	ret = igsf_get_write_mode(nor->program_opcode, &wr_addr_mode,
				  &wr_data_mode);
	if (ret) {
		dev_err(nor->dev, "program opcode 0x%x not supported\n",
			nor->program_opcode);
		return ret;
	}

	val = FIELD_PREP(IGSF_OP_SET_RD_DATA_MODE, rd_data_mode) |
		FIELD_PREP(IGSF_OP_SET_RD_ADDR_MODE, rd_addr_mode) |
		FIELD_PREP(IGSF_OP_SET_WR_DATA_MODE, wr_data_mode) |
		FIELD_PREP(IGSF_OP_SET_WR_ADDR_MODE, wr_addr_mode) |
		FIELD_PREP(IGSF_OP_SET_INS_MODE, 0);
	ret = regmap_write(ctrl->csr_regmap, IGSF_OP_SET_REG_OFF, val);
	if (ret) {
		dev_err(nor->dev,
			"%s failed write to IGSF_OP_SET_REG_OFF %d\n",
			__func__, ret);
		return ret;
	}
	dev_dbg(nor->dev, "set OP_SET[0x%x] = 0x%x\n", IGSF_OP_SET_REG_OFF,
		val);

	val = FIELD_PREP(IGSF_RD_INSTR_DUMMY_CYCLES, nor->read_dummy) |
		FIELD_PREP(IGSF_RD_INSTR_OPCODE, nor->read_opcode);
	ret = regmap_write(ctrl->csr_regmap, IGSF_RD_INSTR_REG_OFF, val);
	if (ret) {
		dev_err(nor->dev,
			"%s failed write to IGSF_RD_INSTR_REG_OFF %d\n",
			__func__, ret);
		return ret;
	}
	dev_dbg(nor->dev, "set RD_INSTR[0x%x] = 0x%x\n", IGSF_RD_INSTR_REG_OFF,
		val);

	poll_opcode = (nor->flags & SNOR_F_USE_FSR) ? SPINOR_OP_RDFSR :
						      SPINOR_OP_RDSR;
	val = FIELD_PREP(IGSF_WR_INSTR_POLL_OPCODE, poll_opcode) |
		FIELD_PREP(IGSF_WR_INSTR_WR_OPCODE, nor->program_opcode);
	ret = regmap_write(ctrl->csr_regmap, IGSF_WR_INSTR_REG_OFF, val);
	if (ret) {
		dev_err(nor->dev,
			"%s failed write to IGSF_WR_INSTR_REG_OFF %d\n",
			__func__, ret);
		return ret;
	}
	dev_dbg(nor->dev, "set WR_INSTR[0x%x] = 0x%x\n", IGSF_WR_INSTR_REG_OFF,
		val);

	ret = mtd_device_register(&nor->mtd, NULL, 0);
	if (ret)
		dev_err(nor->dev, "%s mtd_device_register failed %d\n",
			__func__, ret);

	return ret;
}

static int igsf_probe(struct platform_device *pdev)
{
	struct intel_generic_qspi_platform_data *pdata;
	struct device *dev = &pdev->dev;
	struct igsf_ctrl *ctrl;
	int ret;

	dev_dbg(dev, "%s\n", __func__);

	pdata = dev_get_platdata(dev);

	ctrl = devm_kzalloc(dev, sizeof(*ctrl), GFP_KERNEL);

	if (!ctrl)
		return -ENOMEM;

	ctrl->csr_regmap = devm_regmap_init(dev, &pdata->regmap_bus,
					    pdata->regmap_ctx,
					    &igsf_csr_regmap_config);
	if (IS_ERR(ctrl->csr_regmap)) {
		dev_err(&pdev->dev, "failed to init igsf regmap: %ld\n",
			PTR_ERR(ctrl->csr_regmap));
		return PTR_ERR(ctrl->csr_regmap);
	}

	ctrl->data_regmap = devm_regmap_init(dev, &pdata->regmap_bus,
					    pdata->data_regmap_ctx,
					    &igsf_data_regmap_config);

	if (IS_ERR(ctrl->data_regmap)) {
		dev_err(&pdev->dev, "failed to init igsf data_regmap: %ld\n",
			PTR_ERR(ctrl->data_regmap));
		return PTR_ERR(ctrl->data_regmap);
	}

	ctrl->dev = dev;

	mutex_init(&ctrl->bus_mutex);

	dev_set_drvdata(dev, ctrl);

	ret = igsf_setup_flash(ctrl);

	if (ret)
		dev_err(dev, "%s igsf_add_flash failed %d\n",
			__func__, ret);

	return ret;
}

static int igsf_remove(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct igsf_flash *flash;
	struct igsf_ctrl *ctrl;
	int ret;

	dev_dbg(dev, "%s\n", __func__);

	ctrl = dev_get_drvdata(dev);

	flash = &ctrl->flash;

	ret = mtd_device_unregister(&flash->nor.mtd);
	if (ret)
		dev_err(dev, "%s mtd_device_unregister failed %d\n",
			__func__, ret);

	mutex_destroy(&ctrl->bus_mutex);

	return ret;
}

static struct platform_driver igsf_driver = {
	.driver = {
		.name = INTEL_GENERIC_QSPI_DRV_NAME,
	},
	.probe = igsf_probe,
	.remove = igsf_remove,
};
module_platform_driver(igsf_driver);

MODULE_AUTHOR("Matthew Gerlach <matthew.gerlach@linux.intel.com>");
MODULE_DESCRIPTION("Generic Serial Flash Intel FPGA IP");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS("platform:" INTEL_GENERIC_QSPI_DRV_NAME);
