// SPDX-License-Identifier: GPL-2.0

/* Device Driver of Intel(R) C827 Retimer for Parkvale
 *
 * Copyright (C) 2019 Intel Corporation. All rights reserved.
 */

#include <linux/c827_retimer.h>
#include <linux/module.h>
#include <linux/pac_n3000_net.h>
#include <linux/platform_device.h>

#define MAX_LINK 4

struct c827_retimer {
	struct platform_device *pdev;
	struct pkvl_dev pvdev[MAX_LINK];
};

#define PKVL_LINK_STATUS	0x0

#define PKVL_LINK_STAT_BIT(pkvl_id, link_id) \
	BIT(((pkvl_id) << 2) + (link_id))

static u32 c827_retimer_get_link(struct pkvl_dev *pvdev)
{
	struct c827_retimer *retimer = pvdev->priv;
	struct device *dev = &retimer->pdev->dev;
	struct c827_retimer_platdata *pdata = dev_get_platdata(dev);
	struct regmap *regmap = pdata->regmap;
	unsigned int val;

	dev_dbg(dev, "%s\n", __func__);

	if (regmap_read(regmap, pdata->addr + PKVL_LINK_STATUS, &val)) {
		dev_err(dev, "fail to read PKVL_LINK_STATUS\n");
		return 0;
	}

	if (val & PKVL_LINK_STAT_BIT(pvdev->id, pvdev->link_id))
		return 1;

	return 0;
}

static int c827_retimer_probe(struct platform_device *pdev)
{
	struct c827_retimer_platdata *pdata = dev_get_platdata(&pdev->dev);
	struct c827_retimer *retimer;
	unsigned int i;
	int ret;

	dev_dbg(&pdev->dev, "%s parent_id = %u id = %u\n", __func__,
		pdata->parent_id, pdata->id);

	retimer = devm_kzalloc(&pdev->dev, sizeof(*retimer), GFP_KERNEL);
	if (!retimer)
		return -ENOMEM;

	retimer->pdev = pdev;

	for (i = 0; i < MAX_LINK; i++) {
		retimer->pvdev[i].parent_id = pdata->parent_id;
		retimer->pvdev[i].id = pdata->id;
		retimer->pvdev[i].link_id = i;
		retimer->pvdev[i].priv = retimer;
		retimer->pvdev[i].get_link = c827_retimer_get_link;

		ret = register_pkvl_dev(&retimer->pvdev[i]);
		if (ret)
			goto unregister_pvdev_exit;
	}

	dev_set_drvdata(&pdev->dev, retimer);
	return 0;

unregister_pvdev_exit:
	while (i--)
		unregister_pkvl_dev(&retimer->pvdev[i]);
	return ret;
}

static int c827_retimer_remove(struct platform_device *pdev)
{
	struct c827_retimer *retimer = dev_get_drvdata(&pdev->dev);
	unsigned int i;

	dev_dbg(&pdev->dev, "%s\n", __func__);

	for (i = 0; i < MAX_LINK; i++)
		unregister_pkvl_dev(&retimer->pvdev[i]);

	return 0;
}

static struct platform_driver c827_retimer_driver = {
	.driver = {
		.name = C827_RETIMER_DRV,
	},
	.probe = c827_retimer_probe,
	.remove = c827_retimer_remove,
};

module_platform_driver(c827_retimer_driver);

MODULE_DESCRIPTION("Device Driver for Intel(R) C827 Retimer for Parkvale");
MODULE_AUTHOR("Intel Corporation");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS("platform:" C827_RETIMER_DRV);
