// SPDX-License-Identifier: GPL-2.0

/* Intel(R) PAC N3000 Network Driver
 *
 * Copyright (C) 2019 Intel Corporation. All rights reserved.
 */

#include <linux/version.h>
#include <linux/bitfield-intel-pac.h>
#include <linux/ethtool.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,4,0)
#include <linux/io-64-nonatomic-lo-hi.h>
#endif
#include <linux/iopoll-4-12.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/pac_n3000_net.h>
#include <linux/platform_device.h>
#include <linux/uaccess.h>
#include <uapi/linux/pac_n3000_net.h>

#define ETH_GROUP_INFO		0x8
#define LIGHT_WEIGHT_MAC	BIT_ULL(25)
#define INFO_DIRECTION		BIT_ULL(24)
#define INFO_SPEED		GENMASK_ULL(23, 16)
#define INFO_PHY_NUM		GENMASK_ULL(15, 8)
#define INFO_GROUP_ID		GENMASK_ULL(7, 0)

#define ETH_GROUP_CTRL		0x10
#define CTRL_CMD		GENMASK_ULL(63, 62)
#define CMD_NOP			0
#define CMD_RD			1
#define CMD_WR			2
#define CTRL_DEV_SELECT		GENMASK_ULL(53, 49)
#define CTRL_FEAT_SELECT	BIT_ULL(48)
#define SELECT_IP		0
#define SELECT_FEAT		1
#define CTRL_ADDR		GENMASK_ULL(47, 32)
#define CTRL_WR_DATA		GENMASK_ULL(31, 0)

#define ETH_GROUP_STAT		0x18
#define STAT_RW_VAL		BIT_ULL(32)
#define STAT_RD_DATA		GENMASK_ULL(31, 0)

/* Additional Feature Register */
#define ADD_PHY_CTRL		0x0
#define PHY_RESET		BIT(0)

#define eth_group_for_each_mac(mac, egp) \
	for ((mac) = (egp)->macs; (mac) < (egp)->macs + (egp)->phy_num; (mac)++)

#define eth_group_for_each_phy(phy, egp) \
	for ((phy) = (egp)->phys; (phy) < (egp)->phys + (egp)->phy_num; (phy)++)

#define eth_com_speed(com)	((com)->parent ? (com)->parent->speed : 0)
#define eth_com_base(com)	((com)->parent ? (com)->parent->base : NULL)
#define eth_com_dev(com)	((com)->parent ? (com)->parent->dev : NULL)
#define eth_dev_speed(edev)	((edev)->egroup ? (edev)->egroup->speed : 0)

struct eth_com {
	struct eth_group *parent;
	unsigned int type;
	unsigned int index;
	u8 select;
};

struct eth_group {
	char name[32];
	struct device *dev;
	struct miscdevice miscdev;

	void __iomem *base;
	struct mutex lock;
	unsigned int parent_id;
	unsigned int direction;
	unsigned int group_id;
	unsigned int speed;
	unsigned int phy_num;
	unsigned int config;
	unsigned int lw_mac;
	struct eth_com *phys;
	struct eth_com *macs;
	struct eth_com *wrapper;
};

struct eth_dev {
	struct eth_group *egroup;
	unsigned int index;
	struct list_head node;
	struct net_device *netdev;
};

struct n3000_netdev {
	char name[32];
	struct eth_dev *edev;
	struct pkvl_dev *pvdev;
	struct net_device *netdev;
};

static DEFINE_MUTEX(list_lock);
static LIST_HEAD(edev_list);
static LIST_HEAD(pvdev_list);

static struct eth_group *file_to_eth_group(struct file *file)
{
	return container_of(file->private_data, struct eth_group, miscdev);
}

static const char *eth_com_type_string(unsigned int type)
{
	switch (type) {
	case ETH_GROUP_PHY:
		return "phy";
	case ETH_GROUP_MAC:
		return "mac";
	case ETH_GROUP_ETHER:
		return "ethernet wrapper";
	default:
		return "unknown";
	}
}

static const char *eth_group_config_string(unsigned int config)
{
	switch (config) {
	case ETH_CONFIG_8x10G:
		return "8x10G";
	case ETH_CONFIG_4x25G:
		return "4x25G";
	case ETH_CONFIG_2x1x25G:
		return "2x1x25G";
	case ETH_CONFIG_4x25G_2x25G:
		return "4x25G_2x25G";
	case ETH_CONFIG_2x2x25G:
		return "2x2x25G";
	default:
		return "unknown";
	}
}

#define RW_VAL_INVL		1 /* us */
#define RW_VAL_POLL_TIMEOUT	10 /* us */

static int __eth_com_write_reg(struct eth_com *ecom,
			       u32 flags, u16 addr, u32 data)
{
	struct device *dev = eth_com_dev(ecom);
	void __iomem *base = eth_com_base(ecom);
	u64 v = 0;

	dev_dbg(dev, "%s [%s] index %u select %x flags %x addr %x data %x\n",
		__func__, eth_com_type_string(ecom->type), ecom->index,
		ecom->select, flags, addr, data);

	/* only PHY has additional feature registers */
	if (flags & ETH_GROUP_SELECT_FEAT && ecom->type != ETH_GROUP_PHY)
		return -EINVAL;

	v |= FIELD_PREP(CTRL_CMD, CMD_WR);
	v |= FIELD_PREP(CTRL_DEV_SELECT, ecom->select);
	v |= FIELD_PREP(CTRL_ADDR, addr);
	v |= FIELD_PREP(CTRL_WR_DATA, data);
	v |= FIELD_PREP(CTRL_FEAT_SELECT, flags & ETH_GROUP_SELECT_FEAT);

	writeq(v, base + ETH_GROUP_CTRL);

	if (readq_poll_timeout(base + ETH_GROUP_STAT, v, v & STAT_RW_VAL,
			       RW_VAL_INVL, RW_VAL_POLL_TIMEOUT))
		return -ETIMEDOUT;

	return 0;
}

static int __eth_com_read_reg(struct eth_com *ecom,
			      u32 flags, u16 addr, u32 *data)
{
	struct device *dev = eth_com_dev(ecom);
	void __iomem *base = eth_com_base(ecom);
	u64 v = 0;

	dev_dbg(dev, "%s [%s] index %u select %x flags %x addr %x\n",
		__func__, eth_com_type_string(ecom->type), ecom->index,
		ecom->select, flags, addr);

	/* only PHY has additional feature registers */
	if (flags & ETH_GROUP_SELECT_FEAT && ecom->type != ETH_GROUP_PHY)
		return -EINVAL;

	v |= FIELD_PREP(CTRL_CMD, CMD_RD);
	v |= FIELD_PREP(CTRL_DEV_SELECT, ecom->select);
	v |= FIELD_PREP(CTRL_ADDR, addr);
	v |= FIELD_PREP(CTRL_FEAT_SELECT, flags & ETH_GROUP_SELECT_FEAT);

	writeq(v, base + ETH_GROUP_CTRL);

	if (readq_poll_timeout(base + ETH_GROUP_STAT, v, v & STAT_RW_VAL,
			       RW_VAL_INVL, RW_VAL_POLL_TIMEOUT))
		return -ETIMEDOUT;

	*data = FIELD_GET(STAT_RD_DATA, v);

	dev_dbg(dev, "%s data 0x%x\n", __func__, *data);

	return 0;
}

static int eth_com_write_reg(struct eth_com *ecom,
			     u32 flags, u16 addr, u32 data)
{
	int ret;

	mutex_lock(&ecom->parent->lock);
	ret = __eth_com_write_reg(ecom, flags, addr, data);
	mutex_unlock(&ecom->parent->lock);
	return ret;
}

static int eth_com_read_reg(struct eth_com *ecom,
			    u32 flags, u16 addr, u32 *data)
{
	int ret;

	mutex_lock(&ecom->parent->lock);
	ret = __eth_com_read_reg(ecom, flags, addr, data);
	mutex_unlock(&ecom->parent->lock);
	return ret;
}

static bool eth_pkvl_match(struct eth_dev *edev, struct pkvl_dev *pvdev)
{
	if (edev->netdev || pvdev->netdev)
		return false;

	if (edev->egroup->direction == 0)
		return false;

	/* check parent_id to make sure we bind devices on the same board. */
	if (pvdev->parent_id != edev->egroup->parent_id)
		return false;

	switch (edev->egroup->config) {
	case ETH_CONFIG_8x10G:
		/* In 8x10G configuration, egroup has 8 eth devs (PHY + MAC).
		 *
		 *    [pkvl_dev]   <------->   [eth_dev]
		 *        0.0                      0
		 *        0.1                      1
		 *        0.2                      2
		 *        0.3                      3
		 *        1.0                      4
		 *        1.1                      5
		 *        1.2                      6
		 *        1.3                      7
		 */
		if (edev->index == (pvdev->id * 4 + pvdev->link_id))
			return true;
		break;
	case ETH_CONFIG_4x25G:
		/* In 4x25G configuration, egroup has 4 eth devs (PHY + MAC).
		 *
		 *    [pkvl_dev]   <------->   [eth_dev]
		 *        0.0                      0
		 *        0.1                      1
		 *        0.2                      2
		 *        0.3                      3
		 *        1.0
		 *        1.1
		 *        1.2
		 *        1.3
		 */
		if (pvdev->id == 0 && edev->index == pvdev->link_id)
			return true;
		break;
	case ETH_CONFIG_2x1x25G:
		/* In 2x1x25G configuration, egroup has 2 eth devs (PHY + MAC).
		 *
		 *    [pkvl_dev]   <------->   [eth_dev]
		 *        0.0                      0
		 *        0.1
		 *        0.2
		 *        0.3
		 *        1.0                      1
		 *        1.1
		 *        1.2
		 *        1.3
		 */
		if (edev->index == pvdev->id && pvdev->link_id == 0)
			return true;
		break;
	case ETH_CONFIG_2x2x25G:
		/* In 2x2x25G configuration, egroup has 4 eth devs (PHY + MAC).
		 *
		 *    [pkvl_dev]   <------->   [eth_dev]
		 *        0.0                      0
		 *        0.1                      1
		 *        0.2
		 *        0.3
		 *        1.0                      2
		 *        1.1                      3
		 *        1.2
		 *        1.3
		 */
		if (edev->index == 0 && pvdev->id == 0 && pvdev->link_id == 0)
			return true;
		if (edev->index == 1 && pvdev->id == 0 && pvdev->link_id == 1)
			return true;
		if (edev->index == 2 && pvdev->id == 1 && pvdev->link_id == 0)
			return true;
		if (edev->index == 3 && pvdev->id == 1 && pvdev->link_id == 1)
			return true;
		break;
	}

	return false;
}

static struct eth_dev *net_device_to_eth_dev(struct net_device *netdev)
{
	struct n3000_netdev *ndev = netdev_priv(netdev);

	return ndev->edev;
}

static struct eth_com *eth_dev_to_phy(struct eth_dev *edev)
{
	struct eth_group *egroup = edev->egroup;

	return &egroup->phys[edev->index];
}

static struct eth_com *eth_dev_to_mac(struct eth_dev *edev)
{
	struct eth_group *egroup = edev->egroup;

	return &egroup->macs[edev->index];
}

#define MAX_MTU			0xffff
/* 25G PHY/MAC Register */
#define PHY_PMA_SLOOP		0x313
#define MAX_TX_SIZE_CONFIG	0x407
#define MAX_RX_SIZE_CONFIG	0x506
#define TX_FLOW_CTRL_EN		0x605
#define TX_FLOW_CTRL_QUANTA	0x620
#define TX_FLOW_CTRL_HOLDOFF	0x628
#define TX_FLOW_CTRL_SEL	0x640
/* 10G PHY/MAC Register */
#define TX_FRAME_MAXLENGTH	0x2c
#define TX_PAUSE_FRAME_QUANTA	0x42
#define TX_PAUSE_FRAME_HOLDOFF	0x43
#define TX_PAUSE_FRAME_EN	0x44
#define RX_FRAME_MAXLENGTH	0xae

static void n3000_netdev_setup_mtu(struct net_device *netdev)
{
	struct eth_dev *edev = net_device_to_eth_dev(netdev);
	struct eth_com *ecom = eth_dev_to_mac(edev);
	u32 tx = 0, rx = 0;

	mutex_lock(&ecom->parent->lock);
	if (eth_dev_speed(edev) == 10) {
		__eth_com_read_reg(ecom, 0, TX_FRAME_MAXLENGTH, &tx);
		__eth_com_read_reg(ecom, 0, RX_FRAME_MAXLENGTH, &rx);

		if (tx > rx)
			__eth_com_write_reg(ecom, 0, TX_FRAME_MAXLENGTH, rx);
		else if (tx < rx)
			__eth_com_write_reg(ecom, 0, RX_FRAME_MAXLENGTH, tx);

	} else if (eth_dev_speed(edev) == 25 || eth_dev_speed(edev) == 40) {
		__eth_com_read_reg(ecom, 0, MAX_TX_SIZE_CONFIG, &tx);
		__eth_com_read_reg(ecom, 0, MAX_RX_SIZE_CONFIG, &rx);

		if (tx > rx)
			__eth_com_write_reg(ecom, 0, MAX_TX_SIZE_CONFIG, rx);
		else if (tx < rx)
			__eth_com_write_reg(ecom, 0, MAX_RX_SIZE_CONFIG, tx);
	}
	mutex_unlock(&ecom->parent->lock);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,10,0)
	netdev->max_mtu = MAX_MTU;
#elif LINUX_VERSION_CODE == KERNEL_VERSION(3,10,0)
	netdev->extended->max_mtu = MAX_MTU;
#endif
	netdev->mtu = (tx > rx) ? rx : tx;
}

static int n3000_netdev_change_mtu(struct net_device *netdev, int new_mtu)
{
	struct eth_dev *edev = net_device_to_eth_dev(netdev);
	struct eth_com *ecom = eth_dev_to_mac(edev);

	mutex_lock(&ecom->parent->lock);
	if (eth_dev_speed(edev) == 10) {
		__eth_com_write_reg(ecom, 0, TX_FRAME_MAXLENGTH, (u32)new_mtu);
		__eth_com_write_reg(ecom, 0, RX_FRAME_MAXLENGTH, (u32)new_mtu);

	} else if (eth_dev_speed(edev) == 25 || eth_dev_speed(edev) == 40) {
		__eth_com_write_reg(ecom, 0, MAX_TX_SIZE_CONFIG, (u32)new_mtu);
		__eth_com_write_reg(ecom, 0, MAX_RX_SIZE_CONFIG, (u32)new_mtu);
	}
	mutex_unlock(&ecom->parent->lock);

	netdev->mtu = new_mtu;
	return 0;
}

static int n3000_set_loopback(struct net_device *netdev, bool en)
{
	struct eth_dev *edev = net_device_to_eth_dev(netdev);
	struct eth_com *ecom;
	u32 data;

	mutex_lock(&edev->egroup->lock);
	if (eth_dev_speed(edev) == 10) {
		ecom = eth_dev_to_phy(edev);
		__eth_com_read_reg(ecom, 0, 0x2e1, &data);
		data &= ~0x1;
		data |= en;
		__eth_com_write_reg(ecom, 0, 0x2e1, data);
	} else if (eth_dev_speed(edev) == 25) {
		ecom = eth_dev_to_mac(edev);
		__eth_com_write_reg(ecom, 0, PHY_PMA_SLOOP, en);
	} else if (eth_dev_speed(edev) == 40) {
		ecom = eth_dev_to_mac(edev);
		__eth_com_write_reg(ecom, 0, PHY_PMA_SLOOP, en ? 0xf : 0);
	}
	mutex_unlock(&edev->egroup->lock);

	return 0;
}

static int n3000_netdev_set_features(struct net_device *dev,
				     netdev_features_t features)
{
	netdev_features_t changed = dev->features ^ features;

	if ((changed & NETIF_F_LOOPBACK) && netif_running(dev))
		n3000_set_loopback(dev, !!(features & NETIF_F_LOOPBACK));

	return 0;
}

static const struct net_device_ops n3000_netdev_ops = {
#if LINUX_VERSION_CODE == KERNEL_VERSION(3,10,0)
	.ndo_change_mtu_rh74 = n3000_netdev_change_mtu,
#else
	.ndo_change_mtu = n3000_netdev_change_mtu,
#endif
	.ndo_set_features = n3000_netdev_set_features,
};

static u32 n3000_ethtool_get_link(struct net_device *netdev)
{
	struct n3000_netdev *ndev = netdev_priv(netdev);
	struct pkvl_dev *pvdev = ndev->pvdev;

	return pvdev->get_link(pvdev);
}

struct stat_info {
	unsigned int addr;
	char string[ETH_GSTRING_LEN];
};

#define STAT_INFO(_addr, _string) \
	.addr = _addr, .string = _string,

static struct stat_info stats_10g[] = {
	/* TX Statistics */
	{STAT_INFO(0x142, "tx_frame_ok")},
	{STAT_INFO(0x144, "tx_frame_err")},
	{STAT_INFO(0x146, "tx_frame_crc_err")},
	{STAT_INFO(0x148, "tx_octets_ok")},
	{STAT_INFO(0x14a, "tx_pause_mac_ctrl_frames")},
	{STAT_INFO(0x14c, "tx_if_err")},
	{STAT_INFO(0x14e, "tx_unicast_frame_ok")},
	{STAT_INFO(0x150, "tx_unicast_frame_err")},
	{STAT_INFO(0x152, "tx_multicast_frame_ok")},
	{STAT_INFO(0x154, "tx_multicast_frame_err")},
	{STAT_INFO(0x156, "tx_broadcast_frame_ok")},
	{STAT_INFO(0x158, "tx_broadcast_frame_err")},
	{STAT_INFO(0x15a, "tx_ether_octets")},
	{STAT_INFO(0x15c, "tx_ether_pkts")},
	{STAT_INFO(0x15e, "tx_ether_undersize_pkts")},
	{STAT_INFO(0x160, "tx_ether_oversize_pkts")},
	{STAT_INFO(0x162, "tx_ether_pkts_64_octets")},
	{STAT_INFO(0x164, "tx_ether_pkts_65_127_octets")},
	{STAT_INFO(0x166, "tx_ether_pkts_128_255_octets")},
	{STAT_INFO(0x168, "tx_ether_pkts_256_511_octets")},
	{STAT_INFO(0x16a, "tx_ether_pkts_512_1023_octets")},
	{STAT_INFO(0x16c, "tx_ether_pkts_1024_1518_octets")},
	{STAT_INFO(0x16e, "tx_ether_pkts_1519_x_octets")},
	/* {STAT_INFO(0x170, "tx_ether_fragments")}, */
	/* {STAT_INFO(0x172, "tx_ether_jabbers")}, */
	/* {STAT_INFO(0x174, "tx_ether_crc_err")}, */
	{STAT_INFO(0x176, "tx_unicast_mac_ctrl_frames")},
	{STAT_INFO(0x178, "tx_multicast_mac_ctrl_frames")},
	{STAT_INFO(0x17a, "tx_broadcast_mac_ctrl_frames")},
	{STAT_INFO(0x17c, "tx_pfc_mac_ctrl_frames")},

	/* RX Statistics */
	{STAT_INFO(0x1c2, "rx_frame_ok")},
	{STAT_INFO(0x1c4, "rx_frame_err")},
	{STAT_INFO(0x1c6, "rx_frame_crc_err")},
	{STAT_INFO(0x1c8, "rx_octets_ok")},
	{STAT_INFO(0x1ca, "rx_pause_mac_ctrl_frames")},
	{STAT_INFO(0x1cc, "rx_if_err")},
	{STAT_INFO(0x1ce, "rx_unicast_frame_ok")},
	{STAT_INFO(0x1d0, "rx_unicast_frame_err")},
	{STAT_INFO(0x1d2, "rx_multicast_frame_ok")},
	{STAT_INFO(0x1d4, "rx_multicast_frame_err")},
	{STAT_INFO(0x1d6, "rx_broadcast_frame_ok")},
	{STAT_INFO(0x1d8, "rx_broadcast_frame_err")},
	{STAT_INFO(0x1da, "rx_ether_octets")},
	{STAT_INFO(0x1dc, "rx_ether_pkts")},
	{STAT_INFO(0x1de, "rx_ether_undersize_pkts")},
	{STAT_INFO(0x1e0, "rx_ether_oversize_pkts")},
	{STAT_INFO(0x1e2, "rx_ether_pkts_64_octets")},
	{STAT_INFO(0x1e4, "rx_ether_pkts_65_127_octets")},
	{STAT_INFO(0x1e6, "rx_ether_pkts_128_255_octets")},
	{STAT_INFO(0x1e8, "rx_ether_pkts_256_511_octets")},
	{STAT_INFO(0x1ea, "rx_ether_pkts_512_1023_octets")},
	{STAT_INFO(0x1ec, "rx_ether_pkts_1024_1518_octets")},
	{STAT_INFO(0x1ee, "rx_ether_pkts_1519_x_octets")},
	{STAT_INFO(0x1f0, "rx_ether_fragments")},
	{STAT_INFO(0x1f2, "rx_ether_jabbers")},
	{STAT_INFO(0x1f4, "rx_ether_crc_err")},
	{STAT_INFO(0x1f6, "rx_unicast_mac_ctrl_frames")},
	{STAT_INFO(0x1f8, "rx_multicast_mac_ctrl_frames")},
	{STAT_INFO(0x1fa, "rx_broadcast_mac_ctrl_frames")},
	{STAT_INFO(0x1fc, "rx_pfc_mac_ctrl_frames")},
};

static struct stat_info stats_25g[] = {
	/* TX Statistics */
	{STAT_INFO(0x800, "tx_fragments")},
	{STAT_INFO(0x802, "tx_jabbers")},
	{STAT_INFO(0x804, "tx_fcs")},
	{STAT_INFO(0x806, "tx_crc_err")},
	{STAT_INFO(0x808, "tx_mcast_data_err")},
	{STAT_INFO(0x80a, "tx_bcast_data_err")},
	{STAT_INFO(0x80c, "tx_ucast_data_err")},
	{STAT_INFO(0x80e, "tx_mcast_ctrl_err")},
	{STAT_INFO(0x810, "tx_bcast_ctrl_err")},
	{STAT_INFO(0x812, "tx_ucast_ctrl_err")},
	{STAT_INFO(0x814, "tx_pause_err")},
	{STAT_INFO(0x816, "tx_64_byte")},
	{STAT_INFO(0x818, "tx_65_127_byte")},
	{STAT_INFO(0x81a, "tx_128_255_byte")},
	{STAT_INFO(0x81c, "tx_256_511_byte")},
	{STAT_INFO(0x81e, "tx_512_1023_byte")},
	{STAT_INFO(0x820, "tx_1024_1518_byte")},
	{STAT_INFO(0x822, "tx_1519_max_byte")},
	{STAT_INFO(0x824, "tx_oversize")},
	{STAT_INFO(0x826, "tx_mcast_data_ok")},
	{STAT_INFO(0x828, "tx_bcast_data_ok")},
	{STAT_INFO(0x82a, "tx_ucast_data_ok")},
	{STAT_INFO(0x82c, "tx_mcast_ctrl_ok")},
	{STAT_INFO(0x82e, "tx_bcast_ctrl_ok")},
	{STAT_INFO(0x830, "tx_ucast_ctrl_ok")},
	{STAT_INFO(0x832, "tx_pause")},
	{STAT_INFO(0x834, "tx_runt")},
	{STAT_INFO(0x860, "tx_payload_octets_ok")},
	{STAT_INFO(0x862, "tx_frame_octets_ok")},

	/* RX Statistics */
	{STAT_INFO(0x900, "rx_fragments")},
	{STAT_INFO(0x902, "rx_jabbers")},
	{STAT_INFO(0x904, "rx_fcs")},
	{STAT_INFO(0x906, "rx_crc_err")},
	{STAT_INFO(0x908, "rx_mcast_data_err")},
	{STAT_INFO(0x90a, "rx_bcast_data_err")},
	{STAT_INFO(0x90c, "rx_ucast_data_err")},
	{STAT_INFO(0x90e, "rx_mcast_ctrl_err")},
	{STAT_INFO(0x910, "rx_bcast_ctrl_err")},
	{STAT_INFO(0x912, "rx_ucast_ctrl_err")},
	{STAT_INFO(0x914, "rx_pause_err")},
	{STAT_INFO(0x916, "rx_64_byte")},
	{STAT_INFO(0x918, "rx_65_127_byte")},
	{STAT_INFO(0x91a, "rx_128_255_byte")},
	{STAT_INFO(0x91c, "rx_256_511_byte")},
	{STAT_INFO(0x91e, "rx_512_1023_byte")},
	{STAT_INFO(0x920, "rx_1024_1518_byte")},
	{STAT_INFO(0x922, "rx_1519_max_byte")},
	{STAT_INFO(0x924, "rx_oversize")},
	{STAT_INFO(0x926, "rx_mcast_data_ok")},
	{STAT_INFO(0x928, "rx_bcast_data_ok")},
	{STAT_INFO(0x92a, "rx_ucast_data_ok")},
	{STAT_INFO(0x92c, "rx_mcast_ctrl_ok")},
	{STAT_INFO(0x92e, "rx_bcast_ctrl_ok")},
	{STAT_INFO(0x930, "rx_ucast_ctrl_ok")},
	{STAT_INFO(0x932, "rx_pause")},
	{STAT_INFO(0x934, "rx_runt")},
	{STAT_INFO(0x960, "rx_payload_octets_ok")},
	{STAT_INFO(0x962, "rx_frame_octets_ok")},
};

static void n3000_ethtool_get_strings(struct net_device *netdev, u32 stringset,
				      u8 *s)
{
	struct eth_dev *edev = net_device_to_eth_dev(netdev);
	unsigned int i, stats_num = 0;
	struct stat_info *stat;

	switch (stringset) {
	case ETH_SS_STATS:
		if (edev->egroup->lw_mac)
			return;

		if (eth_dev_speed(edev) == 25 || eth_dev_speed(edev) == 40) {
			stat = stats_25g;
			stats_num = ARRAY_SIZE(stats_25g);
		} else if (eth_dev_speed(edev) == 10) {
			stat = stats_10g;
			stats_num = ARRAY_SIZE(stats_10g);
		}
		break;
	default:
		return;
	}

	for (i = 0; i < stats_num; i++, s += ETH_GSTRING_LEN)
		memcpy(s, stat[i].string, ETH_GSTRING_LEN);
}

static int n3000_ethtool_get_sset_count(struct net_device *netdev,
					int stringset)
{
	struct eth_dev *edev = net_device_to_eth_dev(netdev);

	switch (stringset) {
	case ETH_SS_STATS:
		if (edev->egroup->lw_mac)
			return 0;

		if (eth_dev_speed(edev) == 25 || eth_dev_speed(edev) == 40)
			return ARRAY_SIZE(stats_25g);
		else if (eth_dev_speed(edev) == 10)
			return ARRAY_SIZE(stats_10g);
		break;
	}

	return 0;
}

static u64 __read_stats(struct eth_com *ecom, unsigned int addr)
{
	u32 data_l, data_h;

	if (__eth_com_read_reg(ecom, 0, addr, &data_l) ||
	    __eth_com_read_reg(ecom, 0, addr + 1, &data_h))
		return 0xffffffffffffffffULL;

	return data_l + ((u64)data_h << 32);
}

static void n3000_ethtool_get_stats(struct net_device *netdev,
				    struct ethtool_stats *stats, u64 *data)
{
	struct eth_dev *edev = net_device_to_eth_dev(netdev);
	unsigned int i, stats_num = 0;
	struct stat_info *stat;
	struct eth_com *ecom;

	if (edev->egroup->lw_mac)
		return;

	if (eth_dev_speed(edev) == 10) {
		ecom = eth_dev_to_mac(edev);
		stat = stats_10g;
		stats_num = ARRAY_SIZE(stats_10g);
	} else if (eth_dev_speed(edev) == 25 || eth_dev_speed(edev) == 40) {
		ecom = eth_dev_to_mac(edev);
		stat = stats_25g;
		stats_num = ARRAY_SIZE(stats_25g);
	}

	mutex_lock(&edev->egroup->lock);
	for (i = 0; i < stats_num; i++)
		data[i] = __read_stats(ecom, stat[i].addr);
	mutex_unlock(&edev->egroup->lock);
}

static void n3000_ethtool_get_pauseparam(struct net_device *netdev,
					 struct ethtool_pauseparam *param)
{
	struct eth_dev *edev = net_device_to_eth_dev(netdev);
	struct eth_com *ecom = eth_dev_to_mac(edev);
	u32 data;

	mutex_lock(&edev->egroup->lock);

	if (eth_dev_speed(edev) == 10) {
		__eth_com_read_reg(ecom, 0, TX_PAUSE_FRAME_EN, &data);
		param->tx_pause = (data & 0x1) ? 1 : 0;
	} else if (eth_dev_speed(edev) == 25) {
		__eth_com_read_reg(ecom, 0, TX_FLOW_CTRL_EN, &data);
		param->tx_pause = (data & 0x1) ? 1 : 0;
	}

	mutex_unlock(&edev->egroup->lock);

	param->autoneg = 0;
	param->rx_pause = 0;
}

static int n3000_ethtool_set_pauseparam(struct net_device *netdev,
					struct ethtool_pauseparam *param)
{
	struct eth_dev *edev = net_device_to_eth_dev(netdev);
	struct eth_com *ecom = eth_dev_to_mac(edev);
	bool enable = param->tx_pause;

	if (param->autoneg || param->rx_pause)
		return -EOPNOTSUPP;

	mutex_lock(&edev->egroup->lock);

	if (eth_dev_speed(edev) == 10) {
		__eth_com_write_reg(ecom, 0, TX_PAUSE_FRAME_EN,
				    enable ? 0x1 : 0);
	} else if (eth_dev_speed(edev) == 25) {
		__eth_com_write_reg(ecom, 0, TX_FLOW_CTRL_EN,
				    enable ? 0x1 : 0);
	}

	mutex_unlock(&edev->egroup->lock);

	return 0;
}

static const struct ethtool_ops n3000_ethtool_ops = {
	.get_link = n3000_ethtool_get_link,
	.get_strings = n3000_ethtool_get_strings,
	.get_sset_count = n3000_ethtool_get_sset_count,
	.get_ethtool_stats = n3000_ethtool_get_stats,
	.get_pauseparam = n3000_ethtool_get_pauseparam,
	.set_pauseparam = n3000_ethtool_set_pauseparam,
};

static void n3000_netdev_setup(struct net_device *netdev)
{
	netdev->netdev_ops = &n3000_netdev_ops;
	netdev->ethtool_ops = &n3000_ethtool_ops;

	netdev->features = 0;
	netdev->hw_features |= NETIF_F_LOOPBACK;
	netdev->hard_header_len = 0;
	netdev->priv_flags |= IFF_NO_QUEUE;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0)
	netdev->needs_free_netdev  = true;
#else
	netdev->destructor = free_netdev;
#endif
}

static ssize_t tx_pause_frame_quanta_show(struct device *d,
					  struct device_attribute *attr,
					  char *buf)
{
	struct eth_dev *edev = net_device_to_eth_dev(to_net_dev(d));
	struct eth_com *ecom = eth_dev_to_mac(edev);
	u32 data;
	int ret = 0;

	mutex_lock(&edev->egroup->lock);

	if (eth_dev_speed(edev) == 10)
		__eth_com_read_reg(ecom, 0, TX_PAUSE_FRAME_QUANTA, &data);
	else if (eth_dev_speed(edev) == 25)
		__eth_com_read_reg(ecom, 0, TX_FLOW_CTRL_QUANTA, &data);
	else
		ret = -ENODEV;

	mutex_unlock(&edev->egroup->lock);

	return ret ? : sprintf(buf, "0x%x\n", data);
}

static ssize_t tx_pause_frame_quanta_store(struct device *d,
					   struct device_attribute *attr,
					   const char *buf, size_t len)
{
	struct eth_dev *edev = net_device_to_eth_dev(to_net_dev(d));
	struct eth_com *ecom = eth_dev_to_mac(edev);
	u32 data;
	int ret = 0;

	if (kstrtou32(buf, 0, &data))
		return -EINVAL;

	mutex_lock(&edev->egroup->lock);

	if (eth_dev_speed(edev) == 10)
		__eth_com_write_reg(ecom, 0, TX_PAUSE_FRAME_QUANTA, data);
	else if (eth_dev_speed(edev) == 25)
		__eth_com_write_reg(ecom, 0, TX_FLOW_CTRL_QUANTA, data);
	else
		ret = -ENODEV;

	mutex_unlock(&edev->egroup->lock);

	return ret ? : len;
}
static DEVICE_ATTR_RW(tx_pause_frame_quanta);

static ssize_t tx_pause_frame_holdoff_show(struct device *d,
					   struct device_attribute *attr,
					   char *buf)
{
	struct eth_dev *edev = net_device_to_eth_dev(to_net_dev(d));
	struct eth_com *ecom = eth_dev_to_mac(edev);
	u32 data;
	int ret = 0;

	mutex_lock(&edev->egroup->lock);

	if (eth_dev_speed(edev) == 10)
		__eth_com_read_reg(ecom, 0, TX_PAUSE_FRAME_HOLDOFF, &data);
	else if (eth_dev_speed(edev) == 25)
		__eth_com_read_reg(ecom, 0, TX_FLOW_CTRL_HOLDOFF, &data);
	else
		ret = -ENODEV;

	mutex_unlock(&edev->egroup->lock);

	return ret ? : sprintf(buf, "0x%x\n", data);
}

static ssize_t tx_pause_frame_holdoff_store(struct device *d,
					    struct device_attribute *attr,
					    const char *buf, size_t len)
{
	struct eth_dev *edev = net_device_to_eth_dev(to_net_dev(d));
	struct eth_com *ecom = eth_dev_to_mac(edev);
	u32 data;
	int ret = 0;

	if (kstrtou32(buf, 0, &data))
		return -EINVAL;

	mutex_lock(&edev->egroup->lock);

	if (eth_dev_speed(edev) == 10)
		__eth_com_write_reg(ecom, 0, TX_PAUSE_FRAME_HOLDOFF, data);
	else if (eth_dev_speed(edev) == 25)
		__eth_com_write_reg(ecom, 0, TX_FLOW_CTRL_HOLDOFF, data);
	else
		ret = -ENODEV;

	mutex_unlock(&edev->egroup->lock);

	return ret ? : len;
}
static DEVICE_ATTR_RW(tx_pause_frame_holdoff);

static struct attribute *n3000_dev_attrs[] = {
	&dev_attr_tx_pause_frame_quanta.attr,
	&dev_attr_tx_pause_frame_holdoff.attr,
	NULL
};

static const struct attribute_group n3000_attr_group = {
	.attrs = n3000_dev_attrs
};

static void n3000_flow_control_init(struct net_device *netdev)
{
	struct eth_dev *edev = net_device_to_eth_dev(netdev);
	struct eth_com *ecom = eth_dev_to_mac(edev);

	mutex_lock(&edev->egroup->lock);
	if (eth_dev_speed(edev) == 10) {
		__eth_com_write_reg(ecom, 0, TX_PAUSE_FRAME_EN, 0x0);
	} else if (eth_dev_speed(edev) == 25) {
		__eth_com_write_reg(ecom, 0, TX_FLOW_CTRL_EN, 0x0);
		__eth_com_write_reg(ecom, 0, TX_FLOW_CTRL_SEL, 0x0);
	}

	mutex_unlock(&edev->egroup->lock);
}

static int eth_pkvl_probe(struct eth_dev *edev, struct pkvl_dev *pvdev)
{
	struct eth_group *egroup = edev->egroup;
	struct n3000_netdev *ndev;
	struct net_device *netdev;
	char name[IFNAMSIZ];
	int ret;

	/* The name of n3000 network device is using this format "npacfAgBlC"
	 *
	 * A is the unique FPGA id (e.g. for A10)
	 * B is the group id of this ETH Group.
	 * C is the PHY/MAC link index for Line side ethernet group.
	 */
	snprintf(name, IFNAMSIZ, "npacf%ug%ul%u", egroup->parent_id,
		 egroup->group_id, edev->index);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,17,0)
	netdev = alloc_netdev(sizeof(*ndev), name, NET_NAME_UNKNOWN,
			      n3000_netdev_setup);
#else
	netdev = alloc_netdev(sizeof(*ndev), name, n3000_netdev_setup);
#endif
	if (!netdev)
		return -ENOMEM;

	ndev = netdev_priv(netdev);
	ndev->edev = edev;
	ndev->pvdev = pvdev;
	SET_NETDEV_DEV(netdev, egroup->dev);
	n3000_netdev_setup_mtu(netdev);
	n3000_set_loopback(netdev, !!(netdev->features & NETIF_F_LOOPBACK));
	n3000_flow_control_init(netdev);

	netdev->sysfs_groups[0] = &n3000_attr_group;

	ret = register_netdev(netdev);
	if (ret) {
		dev_err(egroup->dev, "fail to register netdev %s\n", name);
		free_netdev(netdev);
		return ret;
	}

	edev->netdev = netdev;
	pvdev->netdev = netdev;

	set_bit(__LINK_STATE_START, &netdev->state);

	dev_info(egroup->dev, "register %s net device successfully\n", name);

	return 0;
}

static int eth_dev_remove(struct eth_dev *edev)
{
	struct net_device *netdev = edev->netdev;
	struct n3000_netdev *ndev;
	struct pkvl_dev *pvdev;

	if (!netdev)
		return 0;

	ndev = netdev_priv(netdev);
	pvdev = ndev->pvdev;

	unregister_netdev(netdev);

	pvdev->netdev = NULL;
	edev->netdev = NULL;
	return 0;
}

static int pkvl_dev_remove(struct pkvl_dev *pvdev)
{
	struct net_device *netdev = pvdev->netdev;
	struct n3000_netdev *ndev;
	struct eth_dev *edev;

	if (!netdev)
		return 0;

	ndev = netdev_priv(netdev);
	edev = ndev->edev;

	unregister_netdev(netdev);

	pvdev->netdev = NULL;
	edev->netdev = NULL;
	return 0;
}

static void dump_eth_com(struct eth_com *ecom)
{
	struct device *dev = eth_com_dev(ecom);

	dev_dbg(dev, "%s: index %u\n", __func__, ecom->index);
	dev_dbg(dev, "%s: select %u\n", __func__, ecom->select);
	dev_dbg(dev, "%s: type %s\n", __func__,
		eth_com_type_string(ecom->type));
}

static void dump_eth_dev(struct eth_dev *edev)
{
	struct device *dev = edev->egroup->dev;

	dev_dbg(dev, "%s: index %u\n", __func__, edev->index);
	dev_dbg(dev, "%s: egroup %p\n", __func__, edev->egroup);
	dev_dbg(dev, "%s: netdev %p\n", __func__, edev->netdev);
}

static void dump_eth_group(struct eth_group *egroup)
{
	struct device *dev = egroup->dev;
	int i;

	dev_dbg(dev, "%s: name %s\n", __func__, egroup->name);
	dev_dbg(dev, "%s: direction %u\n", __func__, egroup->direction);
	dev_dbg(dev, "%s: speed %u\n", __func__, egroup->speed);
	dev_dbg(dev, "%s: phy_num %u\n", __func__, egroup->phy_num);
	dev_dbg(dev, "%s: group_id %u\n", __func__, egroup->group_id);
	dev_dbg(dev, "%s: base %p\n", __func__, egroup->base);
	dev_dbg(dev, "%s: parent_id %u\n", __func__, egroup->parent_id);
	dev_dbg(dev, "%s: config %s\n", __func__,
		eth_group_config_string(egroup->config));

	for (i = 0; i < egroup->phy_num; i++)
		dump_eth_com(&egroup->phys[i]);
	for (i = 0; i < egroup->phy_num; i++)
		dump_eth_com(&egroup->macs[i]);
	dump_eth_com(egroup->wrapper);
}

static void dump_registered_eth_devs(void)
{
	struct eth_dev *edev;

	mutex_lock(&list_lock);
	if (list_empty(&edev_list))
		goto unlock_exit;

	list_for_each_entry(edev, &edev_list, node)
		dump_eth_dev(edev);

unlock_exit:
	mutex_unlock(&list_lock);
}

static int register_eth_group_devs(struct eth_group *egroup)
{
	struct pkvl_dev *pvdev;
	struct eth_dev *edev;
	int i, ret = 0;

	mutex_lock(&list_lock);

	if (!list_empty(&edev_list)) {
		list_for_each_entry(edev, &edev_list, node)
			if (edev->egroup == egroup) {
				ret = -EEXIST;
				goto unlock_exit;
			}
	}

	edev = devm_kcalloc(egroup->dev, egroup->phy_num, sizeof(*edev),
			    GFP_KERNEL);
	if (!edev) {
		ret = -ENOMEM;
		goto unlock_exit;
	}

	for (i = 0; i < egroup->phy_num; i++) {
		edev[i].egroup = egroup;
		edev[i].index = i;
		list_add_tail(&edev[i].node, &edev_list);

		if (list_empty(&pvdev_list))
			continue;

		list_for_each_entry(pvdev, &pvdev_list, node)
			if (eth_pkvl_match(&edev[i], pvdev))
				eth_pkvl_probe(&edev[i], pvdev);
	}

unlock_exit:
	mutex_unlock(&list_lock);
	return ret;
}

static int unregister_eth_group_devs(struct eth_group *egroup)
{
	struct eth_dev *edev, *tmp;

	mutex_lock(&list_lock);
	if (list_empty(&edev_list))
		goto unlock_exit;

	list_for_each_entry_safe(edev, tmp, &edev_list, node)
		if (edev->egroup == egroup) {
			list_del(&edev->node);
			eth_dev_remove(edev);
		}

unlock_exit:
	mutex_unlock(&list_lock);
	return 0;
}

int register_pkvl_dev(struct pkvl_dev *pvdev)
{
	struct eth_dev *edev;
	struct pkvl_dev *pkvl;
	int ret = 0;

	mutex_lock(&list_lock);

	if (!list_empty(&pvdev_list)) {
		list_for_each_entry(pkvl, &pvdev_list, node)
			if (pvdev == pkvl) {
				ret = -EEXIST;
				goto unlock_exit;
			}
	}

	list_add_tail(&pvdev->node, &pvdev_list);

	if (list_empty(&edev_list))
		goto unlock_exit;

	list_for_each_entry(edev, &edev_list, node)
		if (eth_pkvl_match(edev, pvdev))
			eth_pkvl_probe(edev, pvdev);

unlock_exit:
	mutex_unlock(&list_lock);
	return ret;
}
EXPORT_SYMBOL_GPL(register_pkvl_dev);

void unregister_pkvl_dev(struct pkvl_dev *pvdev)
{
	struct pkvl_dev *pkvl, *tmp;

	mutex_lock(&list_lock);
	if (list_empty(&pvdev_list))
		goto unlock_exit;

	list_for_each_entry_safe(pkvl, tmp, &pvdev_list, node)
		if (pkvl == pvdev) {
			list_del(&pvdev->node);
			pkvl_dev_remove(pvdev);
		}

unlock_exit:
	mutex_unlock(&list_lock);
}
EXPORT_SYMBOL_GPL(unregister_pkvl_dev);

static void init_eth_com(struct eth_com *com, struct eth_group *parent,
			 unsigned int type, unsigned int index)
{
	com->parent = parent;
	com->type = type;
	com->index = index;

	if (type == ETH_GROUP_PHY)
		com->select = index * 2 + 2;
	else if (type == ETH_GROUP_MAC)
		com->select = index * 2 + 3;
	else if (type == ETH_GROUP_ETHER)
		com->select = 0;
}

static int eth_group_ioctl_check_extension(struct eth_group *egroup,
					   unsigned long arg)
{
	/* No extension support for now */
	return 0;
}

static long eth_group_ioctl_get_info(struct eth_group *egroup, void __user *arg)
{
	struct eth_group_info info;
	unsigned long minsz;

	minsz = offsetofend(struct eth_group_info, group_id);

	if (copy_from_user(&info, arg, minsz))
		return -EFAULT;

	if (info.argsz < minsz)
		return -EINVAL;

	if (egroup->lw_mac)
		info.flags = ETH_GROUP_FLAGS_LW_MAC;

	info.speed = egroup->speed;
	info.phy_num = egroup->phy_num;
	info.mac_num = egroup->phy_num;
	info.group_id = egroup->group_id;

	if (copy_to_user(arg, &info, sizeof(info)))
		return -EFAULT;

	return 0;
}

static struct eth_com *eth_group_find_com(struct eth_group *egroup,
					  unsigned int type, unsigned int index)
{
	switch (type) {
	case ETH_GROUP_PHY:
		if (egroup->lw_mac || index >= egroup->phy_num)
			return NULL;
		return &egroup->phys[index];

	case ETH_GROUP_MAC:
		if (index >= egroup->phy_num)
			return NULL;
		return &egroup->macs[index];

	case ETH_GROUP_ETHER:
		if (egroup->lw_mac)
			return NULL;

		return index ? NULL : egroup->wrapper;

	default:
		return NULL;
	}
}

static int eth_group_ioctl_write_reg(struct eth_group *egroup, void __user *arg)
{
	struct eth_group_write_reg reg;
	struct eth_com *com;
	unsigned long minsz;

	minsz = offsetofend(struct eth_group_write_reg, data);

	if (copy_from_user(&reg, arg, minsz))
		return -EFAULT;

	if (reg.argsz < minsz)
		return -EINVAL;

	com = eth_group_find_com(egroup, reg.type, reg.index);
	if (!com)
		return -EINVAL;

	return eth_com_write_reg(com, reg.flags, reg.addr, reg.data);
}

static int eth_group_ioctl_read_reg(struct eth_group *egroup, void __user *arg)
{
	struct eth_group_read_reg reg;
	struct eth_com *com;
	unsigned long minsz;
	int ret;

	minsz = offsetofend(struct eth_group_read_reg, data);

	if (copy_from_user(&reg, arg, minsz))
		return -EFAULT;

	if (reg.argsz < minsz)
		return -EINVAL;

	com = eth_group_find_com(egroup, reg.type, reg.index);
	if (!com)
		return -EINVAL;

	ret = eth_com_read_reg(com, reg.flags, reg.addr, &reg.data);
	if (ret)
		return ret;

	if (copy_to_user(arg, &reg, sizeof(reg)))
		return -EFAULT;

	return 0;
}

static long eth_group_ioctl(struct file *filp, unsigned int cmd,
			    unsigned long arg)
{
	struct eth_group *egroup = file_to_eth_group(filp);

	switch (cmd) {
	case ETH_GROUP_GET_API_VERSION:
		return ETH_GROUP_API_VERSION;
	case ETH_GROUP_CHECK_EXTENSION:
		return eth_group_ioctl_check_extension(egroup, arg);
	case ETH_GROUP_GET_INFO:
		return eth_group_ioctl_get_info(egroup, (void __user *)arg);
	case ETH_GROUP_READ_REG:
		return eth_group_ioctl_read_reg(egroup, (void __user *)arg);
	case ETH_GROUP_WRITE_REG:
		return eth_group_ioctl_write_reg(egroup, (void __user *)arg);
	}

	return -EINVAL;
}

static const struct file_operations eth_group_fops = {
	.owner = THIS_MODULE,
	.open = nonseekable_open,
	.llseek = noop_llseek,
	.unlocked_ioctl = eth_group_ioctl,
};

static struct eth_group *eth_group_alloc_and_init(struct platform_device *pdev)
{
	struct pac_n3000_net_platdata *pdata = dev_get_platdata(&pdev->dev);
	struct device *dev = &pdev->dev;
	struct eth_group *egroup;
	struct eth_com *ecom;
	int i;
	u64 v;

	if (!pdata->base)
		return ERR_PTR(-EINVAL);

	egroup = devm_kzalloc(dev, sizeof(*egroup), GFP_KERNEL);
	if (!egroup)
		return ERR_PTR(-ENOMEM);

	/* read parameters of this ethernet components group */
	v = readq(pdata->base + ETH_GROUP_INFO);

	egroup->direction = FIELD_GET(INFO_DIRECTION, v);
	egroup->speed = FIELD_GET(INFO_SPEED, v);
	egroup->phy_num = FIELD_GET(INFO_PHY_NUM, v);
	egroup->group_id = FIELD_GET(INFO_GROUP_ID, v);
	egroup->base = pdata->base;
	egroup->parent_id = pdata->parent_id;
	egroup->config = pdata->config;
	egroup->lw_mac = FIELD_GET(LIGHT_WEIGHT_MAC, v);

	ecom = devm_kcalloc(dev, egroup->phy_num * 2 + 1, sizeof(*ecom),
			    GFP_KERNEL);
	if (!ecom)
		return ERR_PTR(-ENOMEM);

	egroup->phys = ecom;
	egroup->macs = ecom + egroup->phy_num;
	egroup->wrapper = ecom + egroup->phy_num * 2;

	for (i = 0; i < egroup->phy_num; i++) {
		init_eth_com(&egroup->phys[i], egroup, ETH_GROUP_PHY, i);
		init_eth_com(&egroup->macs[i], egroup, ETH_GROUP_MAC, i);
	}
	init_eth_com(egroup->wrapper, egroup, ETH_GROUP_ETHER, 0);

	snprintf(egroup->name, sizeof(egroup->name), "eth_group%d.%u",
		 egroup->parent_id, egroup->group_id);

	egroup->miscdev.minor = MISC_DYNAMIC_MINOR;
	egroup->miscdev.name = egroup->name;
	egroup->miscdev.parent = dev;
	egroup->miscdev.fops = &eth_group_fops;
	egroup->dev = &pdev->dev;

	mutex_init(&egroup->lock);

	dump_eth_group(egroup);
	return egroup;
}

static void eth_group_uinit(struct eth_group *egroup)
{
	mutex_destroy(&egroup->lock);
}

static int eth_group_phy_reset(struct eth_com *phy, bool enable)
{
	struct device *dev = eth_com_dev(phy);
	u32 val;
	int ret;

	dev_dbg(dev, "%s %s\n", __func__, enable ? "true" : "false");

	/* only support 10G PHY reset for now. It uses external reset. */
	if (eth_com_speed(phy) != 10)
		return 0;

	ret = eth_com_read_reg(phy, ETH_GROUP_SELECT_FEAT, ADD_PHY_CTRL, &val);
	if (ret) {
		dev_err(dev, "fail to read ADD_PHY_CTRL reg: %d\n", ret);
		return ret;
	}

	/* return if PHY is already in expected state */
	if ((val & PHY_RESET && enable) || (!(val & PHY_RESET) && !enable))
		return 0;

	if (enable)
		val |= PHY_RESET;
	else
		val &= ~PHY_RESET;

	ret = eth_com_write_reg(phy, ETH_GROUP_SELECT_FEAT, ADD_PHY_CTRL, val);
	if (ret)
		dev_err(dev, "fail to write ADD_PHY_CTRL reg: %d\n", ret);

	dev_dbg(dev, "%s done\n", __func__);

	return ret;
}

static void eth_group_phys_uinit(struct eth_group *egroup)
{
	struct eth_com *phy;

	dev_dbg(egroup->dev, "%s\n", __func__);

	if (egroup->lw_mac)
		return;

	eth_group_for_each_phy(phy, egroup)
		eth_group_phy_reset(phy, true);
}

static int eth_group_phys_init(struct eth_group *egroup)
{
	struct eth_com *phy;
	int ret;

	dev_dbg(egroup->dev, "%s\n", __func__);

	if (egroup->lw_mac)
		return 0;

	eth_group_for_each_phy(phy, egroup) {
		ret = eth_group_phy_reset(phy, false);
		if (ret) {
			dev_err(egroup->dev, "fail to reset phys\n");
			eth_group_phys_uinit(egroup);
			return ret;
		}
	}

	return 0;
}

/* registers for 25G/40G mac */
#define MAC_CONFIG	0x310
#define MAC_RESET_MASK	GENMASK(2, 0)

static int eth_group_mac_reset(struct eth_com *mac, bool enable)
{
	struct device *dev = eth_com_dev(mac);
	u32 val;
	int ret;

	dev_dbg(dev, "%s %s\n", __func__, enable ? "true" : "false");

	/* only support 25G & 40G mac reset for now. It uses internal reset.
	 * as PHY and MAC are integrated together, below action will trigger
	 * PHY reset too.
	 */
	if (eth_com_speed(mac) == 10)
		return 0;

	ret = eth_com_read_reg(mac, 0, MAC_CONFIG, &val);
	if (ret) {
		dev_err(dev, "fail to read PHY_CONFIG: %d\n", ret);
		return ret;
	}

	/* skip if mac is in expected state already */
	if ((((val & MAC_RESET_MASK) == MAC_RESET_MASK) && enable) ||
	    (((val & MAC_RESET_MASK) == 0) && !enable))
		return 0;

	if (enable)
		val |= MAC_RESET_MASK;
	else
		val &= ~MAC_RESET_MASK;

	ret = eth_com_write_reg(mac, 0, MAC_CONFIG, val);
	if (ret)
		dev_err(dev, "fail to write PHY_CONFIG: %d\n", ret);

	dev_dbg(dev, "%s done\n", __func__);

	return ret;
}

static void eth_group_macs_uinit(struct eth_group *egroup)
{
	struct eth_com *mac;

	eth_group_for_each_mac(mac, egroup)
		eth_group_mac_reset(mac, true);
}

static int eth_group_macs_init(struct eth_group *egroup)
{
	struct eth_com *mac;
	int ret;

	eth_group_for_each_mac(mac, egroup) {
		ret = eth_group_mac_reset(mac, false);
		if (ret) {
			dev_err(egroup->dev, "fail to reset macs\n");
			eth_group_macs_uinit(egroup);
			return ret;
		}
	}

	return 0;
}

static int eth_group_hw_init(struct eth_group *egroup)
{
	int ret;

	ret = eth_group_phys_init(egroup);
	if (ret) {
		dev_err(egroup->dev, "fail to init eth group phys\n");
		return ret;
	}

	ret = eth_group_macs_init(egroup);
	if (ret) {
		dev_err(egroup->dev, "fail to init eth group macs\n");
		goto phy_uinit_exit;
	}

	return 0;

phy_uinit_exit:
	eth_group_phys_uinit(egroup);
	return ret;
}

static int eth_group_hw_uinit(struct eth_group *egroup)
{
	eth_group_macs_uinit(egroup);
	eth_group_phys_uinit(egroup);
	return 0;
}

static int pac_n3000_net_probe(struct platform_device *pdev)
{
	struct eth_group *egroup;
	int ret;

	dev_dbg(&pdev->dev, "%s\n", __func__);

	egroup = eth_group_alloc_and_init(pdev);
	if (IS_ERR(egroup))
		return PTR_ERR(egroup);

	ret = eth_group_hw_init(egroup);
	if (ret) {
		dev_err(&pdev->dev, "fail to init eth group hardware\n");
		goto uinit_exit;
	}

	ret = misc_register(&egroup->miscdev);
	if (ret) {
		dev_err(&pdev->dev, "fail to register misc device\n");
		goto hw_uinit_exit;
	}

	ret = register_eth_group_devs(egroup);
	if (ret) {
		dev_err(&pdev->dev, "fail to register eth group devs\n");
		goto miscdev_exit;
	}

	dev_set_drvdata(&pdev->dev, egroup);
	return 0;

miscdev_exit:
	misc_deregister(&egroup->miscdev);
hw_uinit_exit:
	eth_group_hw_uinit(egroup);
uinit_exit:
	eth_group_uinit(egroup);
	return ret;
}

static int pac_n3000_net_remove(struct platform_device *pdev)
{
	struct eth_group *egroup = dev_get_drvdata(&pdev->dev);

	dev_dbg(&pdev->dev, "%s\n", __func__);

	unregister_eth_group_devs(egroup);
	dump_registered_eth_devs();
	misc_deregister(&egroup->miscdev);
	eth_group_hw_uinit(egroup);
	eth_group_uinit(egroup);
	return 0;
}

static struct platform_driver pac_n3000_net_driver = {
	.driver = {
		.name = PAC_N3000_NET_DRV,
	},
	.probe = pac_n3000_net_probe,
	.remove = pac_n3000_net_remove,
};

module_platform_driver(pac_n3000_net_driver);

MODULE_DESCRIPTION("Network Device Driver for Intel(R) PAC N3000");
MODULE_AUTHOR("Intel Corporation");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS("platform:" PAC_N3000_NET_DRV);
